@@ -12,46 +12,9 @@ std::map<wxString, int> CDirectoryListingParser::m_MonthNamesMap;
 //#define LISTDEBUG
 #ifdef LISTDEBUG
 static char data[][110]={
-
-	/* IBM MVS listings */
-	// Volume Unit    Referred Ext Used Recfm Lrecl BlkSz Dsorg Dsname
-	"  WYOSPT 3420   2003/05/21  1  200  FB      80  8053  PS  48-MVS.FILE",
-	"  WPTA01 3290   2004/03/04  1    3  FB      80  3125  PO  49-MVS.DATASET",
-	"  TSO004 3390   VSAM 50-mvs-file",
-	"  TSO005 3390   2005/06/06 213000 U 0 27998 PO 51-mvs-dir",
-	"  NRP004 3390   **NONE**    1   15  NONE     0     0  PO  52-MVS-NONEDATE.DATASET",
-
-	/* Dataset members */
-	// Name         VV.MM   Created      Changed       Size  Init  Mod Id
-	// ADATAB /* filenames without data, only check for those on MVS servers */
-	"  53-MVSPDSMEMBER 01.01 2004/06/22 2004/06/22 16:32   128   128    0 BOBY12",
-
-	"54-MVSPDSMEMBER2 00B308 000411  00 FO                31    ANY",
-	"55-MVSPDSMEMBER3 00B308 000411  00 FO        RU      ANY    24",
-
-	// Migrated MVS file
-	"Migrated				SOME.FILE",
-
-	// z/VM, another IBM abomination. Description by Alexandre Charbey
-	// 
-	// ACTAPI
-	//   is a filename
-	// TRACE
-	//   is a filetype (extension, like exe or com or jpg...)
-	// V
-	//   is the file format. Designates how records are arranged in a file. F=Fixed and V=Variable. I don't think you care
-	// 65
-	//   is the logical record length.
-	// 107
-	//   is Number of records in a file.
-	// 2
-	//   (seems wrong) is the block size ( iirc 1 is 127, 2 is 381, 3 is 1028 and 4 is 4072 - not sure - the numbers are not the usual binary numbers)
-	// there is the date/time
-	// 060191
-	//   I think it is some internal stuff saying who the file belongs to.  191 is the "handle" of the user's disk. I don't know what 060 is. This 060191 is what FZ shows in its file list.
-	"ACTAPI  TRACE   V        65      107        2 2005-10-04 15:28:42 060191",
-
-	""};
+	
+	"" // Has to be terminated with empty string
+};
 
 #endif
 
@@ -106,7 +69,7 @@ public:
 		if (type > 2)
 			return _T("");
 
-		if (m_str[type] != _T(""))
+		if (!m_str[type].IsEmpty())
 			return m_str[type];
 
 		if (!type)
@@ -345,7 +308,7 @@ protected:
 class CLine
 {
 public:
-	CLine(wxChar* p, int len = -1)
+	CLine(wxChar* p, int len = -1, int trailing_whitespace = 0)
 	{
 		m_pLine = p;
 		if (len != -1)
@@ -354,6 +317,10 @@ public:
 			m_len = wxStrlen(p);
 
 		m_parsePos = 0;
+
+		m_Tokens.reserve(10);
+		m_LineEndTokens.reserve(10);
+		m_trailing_whitespace = trailing_whitespace;
 	}
 
 	~CLine()
@@ -367,7 +334,7 @@ public:
 			delete *iter;
 	}
 
-	bool GetToken(unsigned int n, CToken &token, bool toEnd = false)
+	bool GetToken(unsigned int n, CToken &token, bool toEnd = false, bool include_whitespace = false)
 	{
 		if (!toEnd)
 		{
@@ -380,12 +347,12 @@ public:
 			int start = m_parsePos;
 			while (m_parsePos < m_len)
 			{
-				if (m_pLine[m_parsePos] == ' ')
+				if (m_pLine[m_parsePos] == ' ' || m_pLine[m_parsePos] == '\t')
 				{
 					CToken *pToken = new CToken(m_pLine + start, m_parsePos - start);
 					m_Tokens.push_back(pToken);
 
-					while (m_pLine[m_parsePos] == ' ' && m_parsePos < m_len)
+					while ((m_pLine[m_parsePos] == ' ' || m_pLine[m_parsePos] == '\t') && m_parsePos < m_len)
 						m_parsePos++;
 
 					if (m_Tokens.size() > n)
@@ -414,6 +381,28 @@ public:
 		}
 		else
 		{
+			if (include_whitespace)
+			{
+				const wxChar* p;
+				if (!n)
+				{
+					CToken ref;
+					if (!GetToken(n, ref))
+						return false;
+					p = ref.GetToken() + ref.GetLength() + 1;
+				}
+				else
+				{
+					CToken ref;
+					if (!GetToken(n - 1, ref))
+						return false;
+					p = ref.GetToken() + ref.GetLength() + 1;
+				}
+
+				token = CToken(p, m_len - (p - m_pLine));
+				return true;
+			}
+
 			if (m_LineEndTokens.size() > n)
 			{
 				token = *(m_LineEndTokens[n]);
@@ -428,7 +417,7 @@ public:
 			{
 				const CToken *refToken = m_Tokens[i];
 				const wxChar* p = refToken->GetToken();
-				CToken *pToken = new CToken(p, m_len - (p - m_pLine));
+				CToken *pToken = new CToken(p, m_len - (p - m_pLine) - m_trailing_whitespace);
 				m_LineEndTokens.push_back(pToken);
 			}
 			token = *(m_LineEndTokens[n]);
@@ -444,7 +433,7 @@ public:
 		p[m_len] = ' ';
 		memcpy(p + m_len + 1, pLine->m_pLine, pLine->m_len * sizeof(wxChar));
 
-		return new CLine(p, m_len + pLine->m_len + 1);
+		return new CLine(p, m_len + pLine->m_len + 1, pLine->m_trailing_whitespace);
 	}
 
 protected:
@@ -452,6 +441,7 @@ protected:
 	std::vector<CToken *> m_LineEndTokens;
 	int m_parsePos;
 	int m_len;
+	int m_trailing_whitespace;
 	wxChar* m_pLine;
 };
 
@@ -612,6 +602,16 @@ CDirectoryListingParser::CDirectoryListingParser(CControlSocket* pControlSocket,
 		//Slovenian month names
 		m_MonthNamesMap[_T("avg")] = 8;
 
+		//Icelandic
+#if wxUSE_UNICODE
+		m_MonthNamesMap[_T("ma\x00ed")] = 5;
+		m_MonthNamesMap[_T("j\x00fan")] = 6;
+		m_MonthNamesMap[_T("j\x00fal")] = 7;
+		m_MonthNamesMap[_T("\x00e1g")] = 8;
+		m_MonthNamesMap[_T("n\x00f3v")] = 11;
+#endif
+		m_MonthNamesMap[_T("des")] = 12;
+
 		//There are more languages and thus month
 		//names, but as long as nobody reports a
 		//problem, I won't add them, there are way
@@ -725,8 +725,7 @@ CDirectoryListing CDirectoryListingParser::Parse(const CServerPath &path)
 			CDirentry entry;
 			entry.name = *iter;
 			entry.dir = false;
-			entry.hasDate = false;
-			entry.hasTime = false;
+			entry.hasTimestamp = CDirentry::timestamp_none;
 			entry.link = false;
 			entry.size = -1;
 			entry.unsure = false;
@@ -735,7 +734,6 @@ CDirectoryListing CDirectoryListingParser::Parse(const CServerPath &path)
 	}
 	else
 	{
-		listing.SetCount(m_entryList.size());
 		listing.Assign(m_entryList);
 	}
 
@@ -746,6 +744,7 @@ bool CDirectoryListingParser::ParseLine(CLine *pLine, const enum ServerType serv
 {
 	CDirentry entry;
 	bool res;
+	int ires;
 
 	if (serverType == ZVM)
 	{
@@ -753,16 +752,24 @@ bool CDirectoryListingParser::ParseLine(CLine *pLine, const enum ServerType serv
 		if (res)
 			goto done;
 	}
+	else if (serverType == HPNONSTOP)
+	{
+		res = ParseAsHPNonstop(pLine, entry);
+		if (res)
+			goto done;
+	}
 
-	res = ParseAsUnix(pLine, entry);
+	res = ParseAsUnix(pLine, entry, true); // Common 'ls -l'
 	if (res)
 		goto done;
 	res = ParseAsDos(pLine, entry);
 	if (res)
 		goto done;
-	res = ParseAsMlsd(pLine, entry);
-	if (res)
+	ires = ParseAsMlsd(pLine, entry);
+	if (ires == 1)
 		goto done;
+	else if (ires == 2)
+		goto skip;
 	res = ParseAsEplf(pLine, entry);
 	if (res)
 		goto done;
@@ -791,13 +798,19 @@ bool CDirectoryListingParser::ParseLine(CLine *pLine, const enum ServerType serv
 	if (serverType == MVS)
 #endif //LISTDEBUG_MVS
 	{
-		res = ParseAsIBM_MVS_PDS2(pLine, entry);
-		if (res)
-			goto done;
 		res = ParseAsIBM_MVS_Migrated(pLine, entry);
 		if (res)
 			goto done;
+		res = ParseAsIBM_MVS_PDS2(pLine, entry);
+		if (res)
+			goto done;
+		res = ParseAsIBM_MVS_Tape(pLine, entry);
+		if (res)
+			goto done;
 	}
+	res = ParseAsUnix(pLine, entry, false); // 'ls -l' but without the date/time
+	if (res)
+		goto done;
 
 	// Some servers just send a list of filenames. If a line could not be parsed,
 	// check if it's a filename. If that's the case, store it for later, else clear
@@ -843,21 +856,28 @@ done:
 			entry.name = entry.name.Left(pos);
 	}
 
-	int offset = m_server.GetTimezoneOffset();
-	if (offset && entry.hasTime)
 	{
-		// Apply timezone offset
-		wxTimeSpan span(0, offset, 0, 0);
-		entry.time.Add(span);
+		int offset = m_server.GetTimezoneOffset();
+		if (offset && entry.hasTimestamp >= CDirentry::timestamp_time)
+		{
+			// Apply timezone offset
+			wxTimeSpan span(0, offset, 0, 0);
+			entry.time.Add(span);
+		}
 	}
 
 	entry.unsure = false;
 	m_entryList.push_back(entry);
 
+skip:
+	m_maybeMultilineVms = false;
+	m_fileList.clear();
+	m_fileListOnly = false;
+
 	return true;
 }
 
-bool CDirectoryListingParser::ParseAsUnix(CLine *pLine, CDirentry &entry)
+bool CDirectoryListingParser::ParseAsUnix(CLine *pLine, CDirentry &entry, bool expect_date)
 {
 	int index = 0;
 	CToken token;
@@ -943,8 +963,13 @@ bool CDirectoryListingParser::ParseAsUnix(CLine *pLine, CDirentry &entry)
 			entry.ownerGroup += token.GetString(1);
 		}
 
-		if (!ParseUnixDateTime(pLine, index, entry))
-			continue;
+		if (expect_date)
+		{
+			if (!ParseUnixDateTime(pLine, index, entry))
+				continue;
+		}
+		else
+			entry.hasTimestamp = CDirentry::timestamp_none;
 
 		// Get the filename
 		if (!pLine->GetToken(++index, token, 1))
@@ -969,7 +994,7 @@ bool CDirectoryListingParser::ParseAsUnix(CLine *pLine, CDirentry &entry)
 			}
 		}
 
-		if (entry.hasTime)
+		if (entry.hasTimestamp >= CDirentry::timestamp_time)
 			entry.time.Add(m_timezoneOffset);
 
 		return true;
@@ -983,6 +1008,7 @@ bool CDirectoryListingParser::ParseUnixDateTime(CLine *pLine, int &index, CDiren
 {
 	bool mayHaveTime = true;
 	bool bHasYearAndTime = false;
+	bool hasTime = false;
 
 	CToken token;
 
@@ -1025,10 +1051,7 @@ bool CDirectoryListingParser::ParseUnixDateTime(CLine *pLine, int &index, CDiren
 				return false;
 
 			if (token[pos] == '.')
-			{
-				entry.hasTime = false;
 				return true;
-			}
 
 			year = entry.time.GetYear();
 			month = entry.time.GetMonth() - wxDateTime::Jan + 1;
@@ -1116,13 +1139,10 @@ bool CDirectoryListingParser::ParseUnixDateTime(CLine *pLine, int &index, CDiren
 			strMonth = strMonth.Left(i + 1);
 		}
 		// Check month name
-		if (strMonth.Right(1) == _T(",") || strMonth.Right(1) == _T("."))
+		while (strMonth.Right(1) == _T(",") || strMonth.Right(1) == _T("."))
 			strMonth.RemoveLast();
-		strMonth.MakeLower();
-		std::map<wxString, int>::iterator iter = m_MonthNamesMap.find(strMonth);
-		if (iter == m_MonthNamesMap.end())
+		if (!GetMonthFromName(strMonth, month))
 			return false;
-		month = iter->second;
 	}
 
 	// Get time/year field
@@ -1147,7 +1167,7 @@ bool CDirectoryListingParser::ParseUnixDateTime(CLine *pLine, int &index, CDiren
 		if (minute < 0 || minute > 59)
 			return false;
 
-		entry.hasTime = true;
+		hasTime = true;
 
 		// Some servers use times only for files newer than 6 months
 		if (!year)
@@ -1203,27 +1223,19 @@ bool CDirectoryListingParser::ParseUnixDateTime(CLine *pLine, int &index, CDiren
 				if (minute < 0 || minute > 59)
 					return false;
 
-				entry.hasTime = true;
+				hasTime = true;
 			}
 			else
-			{
-				entry.hasTime = false;
 				index--;
-			}
 		}
-		else
-			entry.hasTime = false;
 	}
 	else
-	{
-		entry.hasTime = false;
 		index--;
-	}
 
 	entry.time = wxDateTime();
 	if (!VerifySetDate(entry.time, year, (wxDateTime::Month)(wxDateTime::Jan + month - 1), day, hour, minute))
 		return false;
-	entry.hasDate = true;
+	entry.hasTimestamp = hasTime ? CDirentry::timestamp_time : CDirentry::timestamp_date;
 
 	return true;
 }
@@ -1254,11 +1266,8 @@ bool CDirectoryListingParser::ParseShortDate(CToken &token, CDirentry &entry, bo
 
 		// Check month name
 		wxString dateMonth = token.GetString().Mid(0, pos);
-		dateMonth.MakeLower();
-		std::map<wxString, int>::iterator iter = m_MonthNamesMap.find(dateMonth);
-		if (iter == m_MonthNamesMap.end())
+		if (!GetMonthFromName(dateMonth, month))
 			return false;
-		month = iter->second;
 		gotMonth = true;
 		gotMonthName = true;
 	}
@@ -1276,7 +1285,7 @@ bool CDirectoryListingParser::ParseShortDate(CToken &token, CDirentry &entry, bo
 		if (token[pos] == '.')
 		{
 			// Maybe dd.mm.yyyy
-			if (value < 1900 || value > 3000)
+			if (value < 1 || value > 31)
 				return false;
 			day = value.GetLo();
 			gotDay = true;
@@ -1343,11 +1352,8 @@ bool CDirectoryListingParser::ParseShortDate(CToken &token, CDirentry &entry, bo
 		// Month field in yyyy-mm-dd or dd-mm-yyyy
 		// Check month name
 		wxString dateMonth = token.GetString().Mid(pos + 1, pos2 - pos - 1);
-		dateMonth.MakeLower();
-		std::map<wxString, int>::iterator iter = m_MonthNamesMap.find(dateMonth);
-		if (iter == m_MonthNamesMap.end())
+		if (!GetMonthFromName(dateMonth, month))
 			return false;
-		month = iter->second;
 		gotMonth = true;
 	}
 	else
@@ -1383,15 +1389,13 @@ bool CDirectoryListingParser::ParseShortDate(CToken &token, CDirentry &entry, bo
 		gotYear = true;
 	}
 
-	entry.hasDate = true;
-
 	if (!gotMonth || !gotDay || !gotYear)
 		return false;
 
 	entry.time = wxDateTime();
 	if (!VerifySetDate(entry.time, year, (wxDateTime::Month)(wxDateTime::Jan + month - 1), day))
 		return false;
-	entry.hasDate = true;
+	entry.hasTimestamp = CDirentry::timestamp_date;
 
 	return true;
 }
@@ -1457,7 +1461,7 @@ bool CDirectoryListingParser::ParseAsDos(CLine *pLine, CDirentry &entry)
 	entry.ownerGroup = _T("");
 	entry.permissions = _T("");
 
-	if (entry.hasTime)
+	if (entry.hasTimestamp >= CDirentry::timestamp_time)
 		entry.time.Add(m_timezoneOffset);
 
 	return true;
@@ -1465,7 +1469,7 @@ bool CDirectoryListingParser::ParseAsDos(CLine *pLine, CDirentry &entry)
 
 bool CDirectoryListingParser::ParseTime(CToken &token, CDirentry &entry)
 {
-	if (!entry.hasDate)
+	if (entry.hasTimestamp == CDirentry::timestamp_none)
 		return false;
 
 	int pos = token.Find(':');
@@ -1492,6 +1496,7 @@ bool CDirectoryListingParser::ParseTime(CToken &token, CDirentry &entry)
 		return false;
 
 	wxLongLong seconds;
+	bool hasSeconds = false;
 	if (pos2 == -1)
 		seconds = 0;
 	else
@@ -1500,6 +1505,7 @@ bool CDirectoryListingParser::ParseTime(CToken &token, CDirentry &entry)
 		seconds = token.GetNumber(pos2 + 1, -1);
 		if (seconds < 0 || seconds > 59)
 			return false;
+		hasSeconds = true;
 	}
 
 	// Convert to 24h format
@@ -1517,7 +1523,7 @@ bool CDirectoryListingParser::ParseTime(CToken &token, CDirentry &entry)
 
 	wxTimeSpan span(hour.GetLo(), minute.GetLo(), seconds.GetLo());
 	entry.time.Add(span);
-	entry.hasTime = true;
+	entry.hasTimestamp = hasSeconds ? CDirentry::timestamp_seconds : CDirentry::timestamp_time;
 
 	return true;
 }
@@ -1541,8 +1547,7 @@ bool CDirectoryListingParser::ParseAsEplf(CLine *pLine, CDirentry &entry)
 	entry.permissions = _T("");
 	entry.dir = false;
 	entry.link = false;
-	entry.hasDate = false;
-	entry.hasTime = false;
+	entry.hasTimestamp = CDirentry::timestamp_none;
 	entry.size = -1;
 
 	int fact = 1;
@@ -1574,8 +1579,7 @@ bool CDirectoryListingParser::ParseAsEplf(CLine *pLine, CDirentry &entry)
 				return false;
 			entry.time = wxDateTime((time_t)number.GetValue());
 
-			entry.hasTime = true;
-			entry.hasDate = true;
+			entry.hasTimestamp = CDirentry::timestamp_seconds;
 		}
 		else if (type == 'u' && len > 2 && token[fact + 1] == 'p')
 			entry.permissions = token.GetString().Mid(fact + 2, len - 2);
@@ -1638,7 +1642,6 @@ bool CDirectoryListingParser::ParseAsVms(CLine *pLine, CDirentry &entry)
 
 	entry.ownerGroup = _T("");
 
-	bool gotSize = false;
 	// This field can either be the filesize, a username (at least that's what I think) enclosed in [] or a date.
 	if (!token.IsNumeric() && !token.IsLeftNumeric())
 	{
@@ -1656,16 +1659,44 @@ bool CDirectoryListingParser::ParseAsVms(CLine *pLine, CDirentry &entry)
 	}
 
 	// Current token is either size or date
+	bool gotSize = false;
 	pos = token.Find('/');
+	
+	if (!pos)
+		return false;
+
 	if (token.IsNumeric() || (pos != -1 && token.Find('/', pos + 1) == -1))
 	{
 		// Definitely size
+
+		CToken sizeToken;
+		if (pos == -1)
+			sizeToken = token;
+		else
+			sizeToken = CToken(token.GetToken(), pos);
+		if (!ParseComplexFileSize(sizeToken, entry.size, 512))
+			return false;
 		gotSize = true;
-		entry.size = token.GetNumber();
 
 		if (!pLine->GetToken(++index, token))
 			return false;
 	}
+	else if (pos == -1 && token.IsLeftNumeric())
+	{
+		// Perhaps size
+		CToken sizeToken;
+		if (pos == -1)
+			sizeToken = token;
+		else
+			sizeToken = CToken(token.GetToken(), pos);
+		if (ParseComplexFileSize(sizeToken, entry.size, 512))
+		{
+			gotSize = true;
+
+			if (!pLine->GetToken(++index, token))
+				return false;
+		}
+	}
 
 	// Get date
 	if (!ParseShortDate(token, entry))
@@ -1686,7 +1717,6 @@ bool CDirectoryListingParser::ParseAsVms(CLine *pLine, CDirentry &entry)
 			return false;
 		if (token[0] != '(' && token[len - 1] == ')')
 			return false;
-		entry.hasTime = false;
 		index--;
 	}
 
@@ -1699,7 +1729,17 @@ bool CDirectoryListingParser::ParseAsVms(CLine *pLine, CDirentry &entry)
 		if (!token.IsNumeric() && !token.IsLeftNumeric())
 			return false;
 
-		entry.size = token.GetNumber();
+		int pos = token.Find('/');
+		if (!pos)
+			return false;
+
+		CToken sizeToken;
+		if (pos == -1)
+			sizeToken = token;
+		else
+			sizeToken = CToken(token.GetToken(), pos);
+		if (!ParseComplexFileSize(sizeToken, entry.size, 512))
+			return false;
 	}
 
 	// Owner / group and permissions
@@ -1709,25 +1749,25 @@ bool CDirectoryListingParser::ParseAsVms(CLine *pLine, CDirentry &entry)
 		const int len = token.GetLength();
 		if (len > 2 && token[0] == '(' && token[len - 1] == ')')
 		{
-			if (entry.permissions != _T(""))
+			if (!entry.permissions.IsEmpty())
 				entry.permissions += _T(" ");
 			entry.permissions += token.GetString().Mid(1, len - 2);
 		}
 		else if (len > 2 && token[0] == '[' && token[len - 1] == ']')
 		{
-			if (entry.ownerGroup != _T(""))
+			if (!entry.ownerGroup.IsEmpty())
 				entry.ownerGroup += _T(" ");
 			entry.ownerGroup += token.GetString().Mid(1, len - 2);
 		}
 		else
 		{
-			if (entry.permissions != _T(""))
+			if (!entry.permissions.IsEmpty())
 				entry.permissions += _T(" ");
 			entry.ownerGroup += token.GetString();
 		}
 	}
 
-	if (entry.hasTime)
+	if (entry.hasTimestamp >= CDirentry::timestamp_time)
 		entry.time.Add(m_timezoneOffset);
 
 	return true;
@@ -1780,7 +1820,7 @@ bool CDirectoryListingParser::ParseAsIbm(CLine *pLine, CDirentry &entry)
 	else
 		entry.dir = false;
 
-	if (entry.hasTime)
+	if (entry.hasTimestamp >= CDirentry::timestamp_time)
 		entry.time.Add(m_timezoneOffset);
 
 	return true;
@@ -1838,8 +1878,7 @@ bool CDirectoryListingParser::ParseOther(CLine *pLine, CDirentry &entry)
 			return false;
 		entry.time = wxDateTime((time_t)number.GetValue());
 
-		entry.hasTime = true;
-		entry.hasDate = true;
+		entry.hasTimestamp = CDirentry::timestamp_seconds;
 
 		// Get filename
 		if (!pLine->GetToken(++index, token, true))
@@ -1861,9 +1900,8 @@ bool CDirectoryListingParser::ParseOther(CLine *pLine, CDirentry &entry)
 
 		// Get date
 		wxString dateMonth = token.GetString();
-		dateMonth.MakeLower();
-		std::map<wxString, int>::const_iterator iter = m_MonthNamesMap.find(dateMonth);
-		if (iter == m_MonthNamesMap.end())
+		int month = 0;
+		if (!GetMonthFromName(dateMonth, month))
 		{
 			// OS/2 or nortel.VxWorks
 			entry.dir = false;
@@ -1896,7 +1934,9 @@ bool CDirectoryListingParser::ParseOther(CLine *pLine, CDirentry &entry)
 				return false;
 
 			entry.name = token.GetString();
-			if (!skippedCount && entry.name.Right(5).MakeUpper() == _T("<DIR>"))
+			wxString type = entry.name.Right(5);
+			MakeLowerAscii(type);
+			if (!skippedCount && type == _T("<dir>"))
 			{
 				entry.dir = true;
 				entry.name = entry.name.Left(entry.name.Length() - 5);
@@ -1906,8 +1946,6 @@ bool CDirectoryListingParser::ParseOther(CLine *pLine, CDirentry &entry)
 		}
 		else
 		{
-			int month = iter->second;
-
 			// Get day
 			if (!pLine->GetToken(++index, token))
 				return false;
@@ -1936,7 +1974,7 @@ bool CDirectoryListingParser::ParseOther(CLine *pLine, CDirentry &entry)
 			if (!VerifySetDate(entry.time, year.GetLo(), (wxDateTime::Month)(month - 1), day.GetLo()))
 				return false;
 
-			entry.hasDate = true;
+			entry.hasTimestamp = CDirentry::timestamp_date;
 
 			// Get time
 			if (!pLine->GetToken(++index, token))
@@ -1964,7 +2002,7 @@ bool CDirectoryListingParser::ParseOther(CLine *pLine, CDirentry &entry)
 		entry.ownerGroup = _T("");
 		entry.permissions = _T("");
 
-		if (entry.hasTime)
+		if (entry.hasTimestamp >= CDirentry::timestamp_time)
 			entry.time.Add(m_timezoneOffset);
 	}
 
@@ -2039,13 +2077,11 @@ CLine *CDirectoryListingParser::GetLine(bool breakAtEnd /*=false*/)
 		int currentOffset = m_currentOffset;
 		while ((iter->p[currentOffset] != '\n') && (iter->p[currentOffset] != '\r'))
 		{
-			if (iter->p[currentOffset] != ' ' && iter->p[currentOffset] != '\t')
-			{
-				reslen += emptylen + 1;
-				emptylen = 0;
-			}
-			else
+			if (iter->p[currentOffset] == ' ' || iter->p[currentOffset] == '\t')
 				emptylen++;
+			else
+				emptylen = 0;
+			reslen++;
 
 			currentOffset++;
 			if (currentOffset >= len)
@@ -2063,7 +2099,7 @@ CLine *CDirectoryListingParser::GetLine(bool breakAtEnd /*=false*/)
 		}
 		m_currentOffset = currentOffset;
 
-		// Reslen is now the length of the line, excluding any terminating whitespace
+		// Reslen is now the length of the line, including any terminating whitespace
 		char *res = new char[reslen + 1];
 		res[reslen] = 0;
 
@@ -2084,19 +2120,13 @@ CLine *CDirectoryListingParser::GetLine(bool breakAtEnd /*=false*/)
 			delete [] i->p;
 			i++;
 		}
-		// Delete all extra whitespace
-		while (i != iter)
-		{
-			delete [] i->p;
-			i++;
-		}
 
 		// Copy last chunk
 		if (iter != m_DataList.end() && reslen)
 		{
 			int copylen = m_currentOffset-startpos;
-			if (copylen>reslen)
-				copylen=reslen;
+			if (copylen > reslen)
+				copylen = reslen;
 			memcpy(&res[respos], &iter->p[startpos], copylen);
 			if (reslen >= iter->len)
 			{
@@ -2118,10 +2148,10 @@ CLine *CDirectoryListingParser::GetLine(bool breakAtEnd /*=false*/)
 		else
 		{
 			wxString str(res, wxConvUTF8);
-			if (str == _T(""))
+			if (str.IsEmpty())
 			{
 				str = wxString(res, wxConvLocal);
-				if (str == _T(""))
+				if (str.IsEmpty())
 					str = wxString(res, wxConvISO8859_1);
 			}
 			buffer = new wxChar[str.Len() + 1];
@@ -2135,7 +2165,7 @@ CLine *CDirectoryListingParser::GetLine(bool breakAtEnd /*=false*/)
 			continue;
 		}
 
-		return new CLine(buffer);
+		return new CLine(buffer, -1, emptylen);
 	}
 
 	return 0;
@@ -2187,7 +2217,7 @@ bool CDirectoryListingParser::ParseAsWfFtp(CLine *pLine, CDirentry &entry)
 	entry.ownerGroup = _T("");
 	entry.permissions = _T("");
 
-	if (entry.hasTime)
+	if (entry.hasTimestamp >= CDirentry::timestamp_time)
 		entry.time.Add(m_timezoneOffset);
 
 	return true;
@@ -2211,7 +2241,7 @@ bool CDirectoryListingParser::ParseAsIBM_MVS(CLine *pLine, CDirentry &entry)
 		return false;
 
 	if (token.GetString() == _T("**NONE**"))
-		entry.hasDate = false;
+		entry.hasTimestamp = CDirentry::timestamp_none;
 	else if (!ParseShortDate(token, entry))
 	{
 		// Perhaps of the following type:
@@ -2230,8 +2260,7 @@ bool CDirectoryListingParser::ParseAsIBM_MVS(CLine *pLine, CDirentry &entry)
 		entry.dir = false;
 		entry.ownerGroup = _T("");
 		entry.permissions = _T("");
-		entry.hasDate = false;
-		entry.hasTime = false;
+		entry.hasTimestamp = CDirentry::timestamp_none;
 		entry.link = false;
 
 		return true;
@@ -2297,7 +2326,6 @@ bool CDirectoryListingParser::ParseAsIBM_MVS(CLine *pLine, CDirentry &entry)
 
 	entry.ownerGroup = _T("");
 	entry.permissions = _T("");
-	entry.hasTime = false;
 	entry.link = false;
 
 	return true;
@@ -2363,7 +2391,7 @@ bool CDirectoryListingParser::ParseAsIBM_MVS_PDS(CLine *pLine, CDirentry &entry)
 	entry.permissions = _T("");
 	entry.link = false;
 
-	if (entry.hasTime)
+	if (entry.hasTimestamp >= CDirentry::timestamp_time)
 		entry.time.Add(m_timezoneOffset);
 
 	return true;
@@ -2392,8 +2420,7 @@ bool CDirectoryListingParser::ParseAsIBM_MVS_Migrated(CLine *pLine, CDirentry &e
 	entry.ownerGroup = _T("");
 	entry.permissions = _T("");
 	entry.size = -1;
-	entry.hasDate = false;
-	entry.hasTime = false;
+	entry.hasTimestamp = CDirentry::timestamp_none;
 
 	if (pLine->GetToken(++index, token))
 		return false;
@@ -2415,8 +2442,7 @@ bool CDirectoryListingParser::ParseAsIBM_MVS_PDS2(CLine *pLine, CDirentry &entry
 	entry.ownerGroup = _T("");
 	entry.permissions = _T("");
 	entry.size = -1;
-	entry.hasDate = false;
-	entry.hasTime = false;
+	entry.hasTimestamp = CDirentry::timestamp_none;
 
 	if (!pLine->GetToken(++index, token))
 		return true;
@@ -2466,25 +2492,74 @@ bool CDirectoryListingParser::ParseAsIBM_MVS_PDS2(CLine *pLine, CDirentry &entry
 	return true;
 }
 
-bool CDirectoryListingParser::ParseComplexFileSize(CToken& token, wxLongLong& size)
+bool CDirectoryListingParser::ParseAsIBM_MVS_Tape(CLine *pLine, CDirentry &entry)
+{
+	int index = 0;
+	CToken token;
+
+	// volume
+	if (!pLine->GetToken(index++, token))
+		return false;
+
+	// unit
+	if (!pLine->GetToken(index++, token))
+		return false;
+
+	if (token.GetString().CmpNoCase(_T("Tape")))
+		return false;
+
+	// dsname
+	if (!pLine->GetToken(index++, token))
+		return false;
+
+	entry.name = token.GetString();
+	entry.dir = false;
+	entry.link = false;
+	entry.ownerGroup = _T("");
+	entry.permissions = _T("");
+	entry.size = -1;
+	entry.hasTimestamp = CDirentry::timestamp_none;
+
+	if (pLine->GetToken(index++, token))
+		return false;
+
+	return true;
+}
+
+bool CDirectoryListingParser::ParseComplexFileSize(CToken& token, wxLongLong& size, int blocksize /*=-1*/)
 {
 	if (token.IsNumeric())
 	{
 		size = token.GetNumber();
+		if (blocksize != -1)
+			size *= blocksize;
+
 		return true;
 	}
 
-	int len = token.GetLength() - 1;
+	int len = token.GetLength();
 
-	char last = token[len];
+	char last = token[len - 1];
 	if (last == 'B' || last == 'b')
 	{
-		char c = token[len];
+		if (len == 1)
+			return false;
+
+		char c = token[--len - 1];
 		if (c < '0' || c > '9')
 		{
-			last = token[len];
 			len--;
+			last = c;
 		}
+		else
+			last = 0;
+	}
+	else if (last >= '0' && last <= '9')
+		last = 0;
+	else
+	{
+		if (--len == 0)
+			return false;
 	}
 
 	size = 0;
@@ -2529,6 +2604,10 @@ bool CDirectoryListingParser::ParseComplexFileSize(CToken& token, wxLongLong& si
 	case 'b':
 	case 'B':
 		break;
+	case 0:
+		if (blocksize != -1)
+			size *= blocksize;
+		break;
 	default:
 		return false;
 	}
@@ -2538,7 +2617,7 @@ bool CDirectoryListingParser::ParseComplexFileSize(CToken& token, wxLongLong& si
 	return true;
 }
 
-bool CDirectoryListingParser::ParseAsMlsd(CLine *pLine, CDirentry &entry)
+int CDirectoryListingParser::ParseAsMlsd(CLine *pLine, CDirentry &entry)
 {
 	// MLSD format as described here: http://www.ietf.org/internet-drafts/draft-ietf-ftpext-mlst-16.txt
 
@@ -2547,38 +2626,53 @@ bool CDirectoryListingParser::ParseAsMlsd(CLine *pLine, CDirentry &entry)
 	CToken token;
 
 	if (!pLine->GetToken(0, token))
-		return false;
+		return 0;
 
 	wxString facts = token.GetString();
-	if (facts == _T(""))
-		return false;
+	if (facts.IsEmpty())
+		return 0;
 
 	entry.dir = false;
 	entry.link = false;
 	entry.size = -1;
-	entry.hasDate = false;
-	entry.hasTime = false;
+	entry.hasTimestamp = CDirentry::timestamp_none;
 	entry.ownerGroup = _T("");
 	entry.permissions = _T("");
 
-	while (facts != _T(""))
+	wxString owner, group, uid, gid;
+
+	while (!facts.IsEmpty())
 	{
 		int delim = facts.Find(';');
 		if (delim < 3)
-			return false;
+		{
+			if (delim != -1)
+				return 0;
+			else
+				delim = facts.Len();
+		}
 
 		int pos = facts.Find('=');
-		if (pos < 1 || (pos + 2) >= delim)
-			return false;
+		if (pos < 1 || pos > delim)
+			return 0;
 
-		wxString factname = facts.Left(pos).Lower();
+		wxString factname = facts.Left(pos);
+		MakeLowerAscii(factname);
 		wxString value = facts.Mid(pos + 1, delim - pos - 1);
 		if (factname == _T("type"))
 		{
-			if (!value.CmpNoCase(_T("dir")) ||
-				!value.CmpNoCase(_T("cdir")) ||
-				!value.CmpNoCase(_T("pdir")))
+			if (!value.CmpNoCase(_T("dir")))
 				entry.dir = true;
+			else if (!value.Left(13).CmpNoCase(_T("OS.unix=slink")))
+			{
+				entry.dir = true;
+				entry.link = true;
+				if (value[13] == ':' && value[14] != 0)
+					entry.target = value.Mid(14);
+			}
+			else if (!value.CmpNoCase(_T("cdir")) ||
+					 !value.CmpNoCase(_T("pdir")))
+				return 2;
 		}
 		else if (factname == _T("size"))
 		{
@@ -2587,45 +2681,77 @@ bool CDirectoryListingParser::ParseAsMlsd(CLine *pLine, CDirentry &entry)
 			for (unsigned int i = 0; i < value.Len(); i++)
 			{
 				if (value[i] < '0' || value[i] > '9')
-					return false;
+					return 0;
 				entry.size *= 10;
 				entry.size += value[i] - '0';
 			}
 		}
 		else if (factname == _T("modify") ||
-			(!entry.hasDate && factname == _T("create")))
+			(entry.hasTimestamp == CDirentry::timestamp_none && factname == _T("create")))
 		{
 			wxDateTime dateTime;
 			const wxChar* time = dateTime.ParseFormat(value, _T("%Y%m%d"));
 
 			if (!time)
-				return false;
+				return 0;
 
 			if (*time)
 			{
-				if (!dateTime.ParseFormat(time, _T("%H%M"), dateTime))
-					return false;
-				entry.hasTime = true;
+				if (!dateTime.ParseFormat(time, _T("%H%M%S"), dateTime))
+					return 0;
+				entry.hasTimestamp = CDirentry::timestamp_seconds;
 			}
 			else
-				entry.hasTime = false;
+				entry.hasTimestamp = CDirentry::timestamp_date;
 
 			entry.time = dateTime.FromTimezone(wxDateTime::GMT0);
-
-			entry.hasDate = true;
 		}
 		else if (factname == _T("perm"))
-			entry.permissions = value;
+		{
+			if (!value.empty())
+			{
+				if (!entry.permissions.empty())
+					entry.permissions = value + _T(" (") + entry.permissions + _T(")");
+				else
+					entry.permissions = value;
+			}
+		}
+		else if (factname == _T("unix.mode"))
+		{
+			if (!entry.permissions.empty())
+				entry.permissions = entry.permissions + _T(" (") + value + _T(")");
+			else
+				entry.permissions = value;
+		}
+		else if (factname == _T("unix.owner") || factname == _T("unix.user"))
+			owner = value;
+		else if (factname == _T("unix.group"))
+			group = value;
+		else if (factname == _T("unix.uid"))
+			uid = value;
+		else if (factname == _T("unix.gid"))
+			gid = value;
 
 		facts = facts.Mid(delim + 1);
 	}
 
-	if (!pLine->GetToken(1, token, true))
-		return false;
+	// The order of the facts is undefined, so assemble ownerGroup in correct
+	// order
+	if (!owner.empty())
+		entry.ownerGroup += owner;
+	else if (!uid.empty())
+		entry.ownerGroup += uid;
+	if (!group.empty())
+		entry.ownerGroup += _T(" ") + group;
+	else if (!gid.empty())
+		entry.ownerGroup += _T(" ") + gid;
+
+	if (!pLine->GetToken(1, token, true, true))
+		return 0;
 
 	entry.name = token.GetString();
 
-	return true;
+	return 1;
 }
 
 bool CDirectoryListingParser::ParseAsOS9(CLine *pLine, CDirentry &entry)
@@ -2688,7 +2814,6 @@ bool CDirectoryListingParser::ParseAsOS9(CLine *pLine, CDirentry &entry)
 
 	entry.name = token.GetString();
 
-	entry.hasTime = false;
 	entry.link = false;
 
 	return true;
@@ -2787,8 +2912,90 @@ bool CDirectoryListingParser::ParseAsZVM(CLine* pLine, CDirentry &entry)
 	entry.permissions = _T("");
 	entry.target = _T("");
 
-	if (entry.hasTime)
+	if (entry.hasTimestamp >= CDirentry::timestamp_time)
 		entry.time.Add(m_timezoneOffset);
 
 	return true;
 }
+
+bool CDirectoryListingParser::ParseAsHPNonstop(CLine *pLine, CDirentry &entry)
+{
+	int index = 0;
+	CToken token;
+
+	// Get name
+	if (!pLine->GetToken(index, token))
+		return false;
+
+	entry.name = token.GetString();
+
+	// File code, numeric, unsuded
+	if (!pLine->GetToken(++index, token))
+		return false;
+	if (!token.IsNumeric())
+		return false;
+
+	// Size
+	if (!pLine->GetToken(++index, token))
+		return false;
+	if (!token.IsNumeric())
+		return false;
+
+	entry.size = token.GetNumber();
+
+	// Date
+	if (!pLine->GetToken(++index, token))
+		return false;
+	if (!ParseShortDate(token, entry, false))
+		return false;
+
+	// Time
+	if (!pLine->GetToken(++index, token))
+		return false;
+	if (!ParseTime(token, entry))
+		return false;
+
+	// Owner
+	if (!pLine->GetToken(++index, token))
+		return false;
+	entry.ownerGroup = token.GetString();
+
+	if (token[token.GetLength() - 1] == ',')
+	{
+		// Owner, part 2
+		if (!pLine->GetToken(++index, token))
+			return false;
+		entry.ownerGroup += _T(" ") + token.GetString();
+	}
+
+	// Permissions
+	if (!pLine->GetToken(++index, token))
+		return false;
+	entry.permissions = token.GetString();
+
+	// Nothing
+	if (pLine->GetToken(++index, token))
+		return false;
+
+	entry.dir = false;
+	entry.link = false;
+		
+	return true;
+}
+
+bool CDirectoryListingParser::GetMonthFromName(const wxString& name, int &month)
+{
+	std::map<wxString, int>::iterator iter = m_MonthNamesMap.find(name.Lower());
+	if (iter == m_MonthNamesMap.end())
+	{
+		wxString lower(name);
+		MakeLowerAscii(lower);
+		iter = m_MonthNamesMap.find(lower);
+		if (iter == m_MonthNamesMap.end())
+			return false;
+	}
+	
+	month = iter->second;
+	
+	return true;
+}
