@@ -1,5 +1,4 @@
 #include "FileZilla.h"
-#include "asynchostresolver.h"
 #include "ControlSocket.h"
 #include "ftpcontrolsocket.h"
 #include "sftpcontrolsocket.h"
@@ -7,6 +6,7 @@
 #include "logging_private.h"
 #include "httpcontrolsocket.h"
 #include "ratelimiter.h"
+#include "pathcache.h"
 
 class wxFzEngineEvent : public wxEvent
 {
@@ -29,8 +29,7 @@ typedef void (wxEvtHandler::*fzEngineEventFunction)(wxFzEngineEvent&);
     ),
 
 std::list<CFileZillaEnginePrivate*> CFileZillaEnginePrivate::m_engineList;
-int CFileZillaEnginePrivate::m_activeStatusSend = 0;
-int CFileZillaEnginePrivate::m_activeStatusRecv = 0;
+int CFileZillaEnginePrivate::m_activeStatus[2] = {0, 0};
 std::list<CFileZillaEnginePrivate::t_failedLogins> CFileZillaEnginePrivate::m_failedLogins;
 
 DEFINE_EVENT_TYPE(fzEVT_ENGINE_NOTIFICATION);
@@ -49,7 +48,6 @@ wxEvent *wxFzEngineEvent::Clone() const
 
 BEGIN_EVENT_TABLE(CFileZillaEnginePrivate, wxEvtHandler)
 	EVT_FZ_ENGINE_NOTIFICATION(wxID_ANY, CFileZillaEnginePrivate::OnEngineEvent)
-	EVT_FZ_ASYNCHOSTRESOLVE(wxID_ANY, CFileZillaEnginePrivate::OnAsyncHostResolver)
 	EVT_TIMER(wxID_ANY, CFileZillaEnginePrivate::OnTimer)
 END_EVENT_TABLE();
 
@@ -67,6 +65,9 @@ CFileZillaEnginePrivate::CFileZillaEnginePrivate()
 	m_engineList.push_back(this);
 	m_retryCount = 0;
 
+	static int id = 0;
+	m_engine_id = id++;
+
 	m_pLogging = new CLogging(this);
 }
 
@@ -79,14 +80,6 @@ CFileZillaEnginePrivate::~CFileZillaEnginePrivate()
 	for (std::list<CNotification *>::iterator iter = m_NotificationList.begin(); iter != m_NotificationList.end(); iter++)
 		delete *iter;
 
-	for (std::list<CAsyncHostResolver *>::iterator iter = m_HostResolverThreads.begin(); iter != m_HostResolverThreads.end(); iter++)
-	{
-		CAsyncHostResolver* pResolver = *iter;
-		pResolver->SetObsolete();
-		pResolver->Wait();
-		delete pResolver;					
-	}
-
 	// Remove ourself from the engine list
 	for (std::list<CFileZillaEnginePrivate*>::iterator iter = m_engineList.begin(); iter != m_engineList.end(); iter++)
 		if (*iter == this)
@@ -99,6 +92,9 @@ CFileZillaEnginePrivate::~CFileZillaEnginePrivate()
 
 	if (m_pRateLimiter)
 		m_pRateLimiter->Free();
+
+	if (m_engineList.empty())
+		CSocket::Cleanup(true);
 }
 
 bool CFileZillaEnginePrivate::SendEvent(enum EngineNotificationType eventType, int data /*=0*/)
@@ -115,8 +111,10 @@ void CFileZillaEnginePrivate::OnEngineEvent(wxFzEngineEvent &event)
 	case engineCancel:
 		if (!IsBusy())
 			break;
-
-		m_pControlSocket->Cancel();
+		if (m_pControlSocket)
+			m_pControlSocket->Cancel();
+		else if (m_pCurrentCommand)
+			ResetOperation(FZ_REPLY_CANCELED);
 		break;
 	case engineTransferEnd:
 		if (m_pControlSocket)
@@ -126,40 +124,6 @@ void CFileZillaEnginePrivate::OnEngineEvent(wxFzEngineEvent &event)
 	}
 }
 
-void CFileZillaEnginePrivate::OnAsyncHostResolver(fzAsyncHostResolveEvent& event)
-{
-	if (m_HostResolverThreads.empty())
-		return;
-	CAsyncHostResolver *pResolver = m_HostResolverThreads.front();
-	m_HostResolverThreads.pop_front();
-
-	std::list<CAsyncHostResolver *> remaining;
-	for (std::list<CAsyncHostResolver *>::iterator iter = m_HostResolverThreads.begin(); iter != m_HostResolverThreads.end(); iter++)
-	{
-		CAsyncHostResolver* pResolver = *iter;
-		pResolver->SetObsolete();
-		if (!pResolver->Done())
-			remaining.push_back(pResolver);
-		else
-		{
-			pResolver->Wait();
-			delete pResolver;					
-		}
-	}
-	m_HostResolverThreads.clear();
-	m_HostResolverThreads = remaining;
-
-	if (!pResolver->Done())
-		m_HostResolverThreads.push_front(pResolver);
-	else
-	{
-		if (!pResolver->Obsolete())
-			m_pControlSocket->ContinueConnect(pResolver->Successful() ? &pResolver->m_Address : 0);
-		pResolver->Wait();
-		delete pResolver;
-	}
-}
-
 bool CFileZillaEnginePrivate::IsBusy() const
 {
 	return m_pCurrentCommand != 0;
@@ -206,6 +170,8 @@ void CFileZillaEnginePrivate::AddNotification(CNotification *pNotification)
 
 int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 {
+	m_pLogging->LogMessage(Debug_Debug, _T("CFileZillaEnginePrivate::ResetOperation(%d)"), nErrorCode);
+					
 	if (nErrorCode & FZ_REPLY_DISCONNECTED)
 		m_lastListDir.Clear();
 
@@ -219,21 +185,26 @@ int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 
 		if (m_pCurrentCommand->GetId() == cmd_connect)
 		{
-			if (!(nErrorCode & ~(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED | FZ_REPLY_TIMEOUT)) && 
-				nErrorCode & (FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED) &&
-				m_retryCount < m_pOptions->GetOptionVal(OPTION_RECONNECTCOUNT))
+			if (!(nErrorCode & ~(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED | FZ_REPLY_TIMEOUT | FZ_REPLY_CRITICALERROR | FZ_REPLY_PASSWORDFAILED)) && 
+				nErrorCode & (FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED))				
 			{
-				m_retryCount++;
-
 				const CConnectCommand *pConnectCommand = (CConnectCommand *)m_pCurrentCommand;
 
-				RegisterFailedLoginAttempt(pConnectCommand->GetServer());
-				unsigned int delay = GetRemainingReconnectDelay(pConnectCommand->GetServer());
-				if (!delay)
-					delay = 1;
-				m_pLogging->LogMessage(Status, _("Waiting to retry..."));
-				m_retryTimer.Start(delay, true);
-				return FZ_REPLY_WOULDBLOCK;
+				RegisterFailedLoginAttempt(pConnectCommand->GetServer(), (nErrorCode & FZ_REPLY_CRITICALERROR) == FZ_REPLY_CRITICALERROR);
+
+				if ((nErrorCode & FZ_REPLY_CRITICALERROR) != FZ_REPLY_CRITICALERROR)
+				{
+					m_retryCount++;
+					if (m_retryCount < m_pOptions->GetOptionVal(OPTION_RECONNECTCOUNT) && pConnectCommand->RetryConnecting())
+					{
+						unsigned int delay = GetRemainingReconnectDelay(pConnectCommand->GetServer());
+						if (!delay)
+							delay = 1;
+						m_pLogging->LogMessage(Status, _("Waiting to retry..."));
+						m_retryTimer.Start(delay, true);
+						return FZ_REPLY_WOULDBLOCK;
+					}
+				}
 			}
 		}
 
@@ -261,33 +232,17 @@ int CFileZillaEnginePrivate::ResetOperation(int nErrorCode)
 		}
 	}
 
-	if (!m_HostResolverThreads.empty())
-		m_HostResolverThreads.front()->SetObsolete();
-
 	return nErrorCode;
 }
 
-void CFileZillaEnginePrivate::SetActive(bool recv)
+void CFileZillaEnginePrivate::SetActive(int direction)
 {
 	if (m_pControlSocket)
 		m_pControlSocket->SetAlive();
 
-	if (recv)
-	{
-		if (!m_activeStatusRecv)
-		{
-			m_activeStatusRecv = 2;
-			AddNotification(new CActiveNotification(true));
-		}
-	}
-	else
-	{
-		if (!m_activeStatusSend)
-		{
-			m_activeStatusSend = 2;
-			AddNotification(new CActiveNotification(false));
-		}
-	}
+	if (!m_activeStatus[direction])
+		AddNotification(new CActiveNotification(direction));
+	m_activeStatus[direction] = 2;
 }
 
 unsigned int CFileZillaEnginePrivate::GetNextAsyncRequestNumber()
@@ -296,16 +251,6 @@ unsigned int CFileZillaEnginePrivate::GetNextAsyncRequestNumber()
 	return ++m_asyncRequestCounter;
 }
 
-void CFileZillaEnginePrivate::AddNewAsyncHostResolver(CAsyncHostResolver* pResolver)
-{
-	wxASSERT(pResolver);
-
-	if (!m_HostResolverThreads.empty())
-		m_HostResolverThreads.front()->SetObsolete();
-	
-	m_HostResolverThreads.push_front(pResolver);
-}
-
 // Command handlers
 int CFileZillaEnginePrivate::Connect(const CConnectCommand &command)
 {
@@ -316,11 +261,22 @@ int CFileZillaEnginePrivate::Connect(const CConnectCommand &command)
 		return FZ_REPLY_BUSY;
 
 	m_retryCount = 0;
+
+	if (m_pControlSocket)
+	{
+		// Need to delete before setting m_pCurrentCommand.
+		// The destructor can call CFileZillaEnginePrivate::ResetOperation
+		// which would delete m_pCurrentCommand
+		delete m_pControlSocket;
+		m_pControlSocket = 0;
+		m_nControlSocketError = 0;
+	}
+
 	m_pCurrentCommand = command.Clone();
 
 	if (command.GetServer().GetPort() != CServer::GetDefaultPort(command.GetServer().GetProtocol()))
 	{
-		ServerProtocol protocol = CServer::GetProtocolFromPort(command.GetServer().GetPort());
+		ServerProtocol protocol = CServer::GetProtocolFromPort(command.GetServer().GetPort(), true);
 		if (protocol != UNKNOWN && protocol != command.GetServer().GetProtocol())
 			m_pLogging->LogMessage(Status, _("Selected port usually in use by a different protocol."));
 	}
@@ -380,38 +336,57 @@ int CFileZillaEnginePrivate::List(const CListCommand &command)
 	if (!IsConnected())
 		return FZ_REPLY_NOTCONNECTED;
 
-	bool refresh = command.Refresh();
-	if (!command.Refresh() && !command.GetPath().IsEmpty())
+	if (command.GetPath().IsEmpty() && command.GetSubDir() != _T(""))
+		return FZ_REPLY_SYNTAXERROR;
+
+	if (command.GetFlags() & LIST_FLAG_LINK && command.GetSubDir() == _T(""))
+		return FZ_REPLY_SYNTAXERROR;
+
+	bool refresh = (command.GetFlags() & LIST_FLAG_REFRESH) != 0;
+	bool avoid = (command.GetFlags() & LIST_FLAG_AVOID) != 0;
+	if (refresh && avoid)
+		return FZ_REPLY_SYNTAXERROR;
+
+	if (!refresh && !command.GetPath().IsEmpty())
 	{
 		const CServer* pServer = m_pControlSocket->GetCurrentServer();
 		if (pServer)
 		{
-			CDirectoryListing *pListing = new CDirectoryListing;
-			CDirectoryCache cache;
-			bool found = cache.Lookup(*pListing, *pServer, command.GetPath(), command.GetSubDir(), true);
-			if (found)
+			CServerPath path(CPathCache::Lookup(*pServer, command.GetPath(), command.GetSubDir()));
+			if (!path.IsEmpty())
 			{
-				
-				if (pListing->m_hasUnsureEntries)
-					refresh = true;
-				else
+				CDirectoryListing *pListing = new CDirectoryListing;
+				CDirectoryCache cache;
+				bool is_outdated = false;
+				bool found = cache.Lookup(*pListing, *pServer, path, true, is_outdated);
+				if (found && !is_outdated)
 				{
-					m_lastListDir = pListing->path;
-					m_lastListTime = wxDateTime::Now();
-					CDirectoryListingNotification *pNotification = new CDirectoryListingNotification(pListing->path);
-					AddNotification(pNotification);
-					delete pListing;
-					return FZ_REPLY_OK;
+					if (pListing->m_hasUnsureEntries)
+						refresh = true;
+					else
+					{
+						if (!avoid)
+						{
+							m_lastListDir = pListing->path;
+							m_lastListTime = wxDateTime::Now();
+							CDirectoryListingNotification *pNotification = new CDirectoryListingNotification(pListing->path);
+							AddNotification(pNotification);
+						}
+						delete pListing;
+						return FZ_REPLY_OK;
+					}
 				}
+				if (is_outdated)
+					refresh = true;
+				delete pListing;
 			}
-			delete pListing;
 		}
 	}
 	if (IsBusy())
 		return FZ_REPLY_BUSY;
 
 	m_pCurrentCommand = command.Clone();
-	return m_pControlSocket->List(command.GetPath(), command.GetSubDir(), refresh);
+	return m_pControlSocket->List(command.GetPath(), command.GetSubDir(), command.GetFlags());
 }
 
 int CFileZillaEnginePrivate::FileTransfer(const CFileTransferCommand &command)
@@ -523,7 +498,6 @@ int CFileZillaEnginePrivate::Chmod(const CChmodCommand& command)
 void CFileZillaEnginePrivate::SendDirectoryListingNotification(const CServerPath& path, bool onList, bool modified, bool failed)
 {
 	wxASSERT(m_pControlSocket);
-	wxASSERT(onList || modified);
 	
 	const CServer* const pOwnServer = m_pControlSocket->GetCurrentServer();
 	wxASSERT(pOwnServer);
@@ -577,14 +551,14 @@ void CFileZillaEnginePrivate::SendDirectoryListingNotification(const CServerPath
 	}
 }
 
-void CFileZillaEnginePrivate::RegisterFailedLoginAttempt(const CServer& server)
+void CFileZillaEnginePrivate::RegisterFailedLoginAttempt(const CServer& server, bool critical)
 {
 	std::list<t_failedLogins>::iterator iter = m_failedLogins.begin();
 	while (iter != m_failedLogins.end())
 	{
 		const wxTimeSpan span = wxDateTime::UNow() - iter->time;
 		if (span.GetSeconds() >= m_pOptions->GetOptionVal(OPTION_RECONNECTDELAY) ||
-			(iter->host == server.GetHost() && iter->port == server.GetPort()))
+			iter->server == server || (!critical && (iter->server.GetHost() == server.GetHost() && iter->server.GetPort() == server.GetPort())))
 		{
 			std::list<t_failedLogins>::iterator prev = iter;
 			iter++;
@@ -595,8 +569,7 @@ void CFileZillaEnginePrivate::RegisterFailedLoginAttempt(const CServer& server)
 	}
 
 	t_failedLogins failure;
-	failure.host = server.GetHost();
-	failure.port = server.GetPort();
+	failure.server = server;
 	failure.time = wxDateTime::UNow();
 	m_failedLogins.push_back(failure);
 }
@@ -614,10 +587,10 @@ unsigned int CFileZillaEnginePrivate::GetRemainingReconnectDelay(const CServer&
 			iter++;
 			m_failedLogins.erase(prev);
 		}
-		else if (iter->host == server.GetHost() && iter->port == server.GetPort())
-		{
+		else if (!iter->critical && iter->server.GetHost() == server.GetHost() && iter->server.GetPort() == server.GetPort())
+			return delay * 1000 - span.GetMilliseconds().GetLo();
+		else if (iter->server == server)
 			return delay * 1000 - span.GetMilliseconds().GetLo();
-		}
 		else
 			iter++;
 	}
@@ -634,28 +607,28 @@ void CFileZillaEnginePrivate::OnTimer(wxTimerEvent& event)
 	}
 	wxASSERT(!IsConnected());
 
-#ifdef __WXDEBUG__
+#if 0//__WXDEBUG__
 	const CConnectCommand *pConnectCommand = (CConnectCommand *)m_pCurrentCommand;
 	wxASSERT(!GetRemainingReconnectDelay(pConnectCommand->GetServer()));
 #endif
 
-	ContinueConnect();
-}
-
-int CFileZillaEnginePrivate::ContinueConnect()
-{
 	if (m_pControlSocket)
 	{
 		delete m_pControlSocket;
 		m_pControlSocket = 0;
 	}
 
+	ContinueConnect();
+}
+
+int CFileZillaEnginePrivate::ContinueConnect()
+{
 	const CConnectCommand *pConnectCommand = (CConnectCommand *)m_pCurrentCommand;
 	const CServer& server = pConnectCommand->GetServer();
 	unsigned int delay = GetRemainingReconnectDelay(server);
 	if (delay)
 	{
-		m_pLogging->LogMessage(Status, _("Delaying connection due to previously failed connection attempt..."));
+		m_pLogging->LogMessage(Status, wxPLURAL("Delaying connection for %d second due to previously failed connection attempt...", "Delaying connection for %d seconds due to previously failed connection attempt...", (delay + 999) / 1000), (delay + 999) / 1000);
 		m_retryTimer.Start(delay, true);
 		return FZ_REPLY_WOULDBLOCK;
 	}
