@@ -1,6 +1,9 @@
 #include "FileZilla.h"
 #include "ControlSocket.h"
 #include "httpcontrolsocket.h"
+#include <errno.h>
+
+#define FZ_REPLY_REDIRECTED FZ_REPLY_ALREADYCONNECTED
 
 // Connect is special for HTTP: It is done on a per-command basis, so we need
 // to establish a connection before each command.
@@ -48,7 +51,7 @@ public:
 
 	wxLongLong m_totalSize;
 	wxLongLong m_receivedData;
-	
+
 	COpData* m_pOpData;
 
 	enum transferEncodings
@@ -70,8 +73,8 @@ public:
 class CHttpFileTransferOpData : public CFileTransferOpData, public CHttpOpData
 {
 public:
-	CHttpFileTransferOpData()
-		: CHttpOpData(this)
+	CHttpFileTransferOpData(const wxString& local_file, const wxString& remote_file, const CServerPath& remote_path)
+		: CFileTransferOpData(local_file, remote_file, remote_path), CHttpOpData(this)
 	{
 		pFile = 0;
 	}
@@ -88,18 +91,16 @@ CHttpControlSocket::CHttpControlSocket(CFileZillaEnginePrivate *pEngine)
 	: CRealControlSocket(pEngine)
 {
 	m_pRecvBuffer = 0;
-	m_pAddress = 0;
 }
 
 CHttpControlSocket::~CHttpControlSocket()
 {
 	delete [] m_pRecvBuffer;
-	delete m_pAddress;
 }
 
-int CHttpControlSocket::SendNextCommand(int prevResult /*=FZ_REPLY_OK*/)
+int CHttpControlSocket::SendNextCommand()
 {
-	LogMessage(Debug_Verbose, _T("CHttpControlSocket::SendNextCommand(%d)"), prevResult);
+	LogMessage(Debug_Verbose, _T("CHttpControlSocket::SendNextCommand()"));
 	if (!m_pCurOpData)
 	{
 		LogMessage(__TFILE__, __LINE__, this, Debug_Warning, _T("SendNextCommand called without active operation"));
@@ -116,7 +117,7 @@ int CHttpControlSocket::SendNextCommand(int prevResult /*=FZ_REPLY_OK*/)
 	switch (m_pCurOpData->opId)
 	{
 	case cmd_transfer:
-		return FileTransferSend(prevResult);
+		return FileTransferSend();
 	default:
 		LogMessage(__TFILE__, __LINE__, this, ::Debug_Warning, _T("Unknown opID (%d) in SendNextCommand"), m_pCurOpData->opId);
 		ResetOperation(FZ_REPLY_INTERNALERROR);
@@ -126,46 +127,33 @@ int CHttpControlSocket::SendNextCommand(int prevResult /*=FZ_REPLY_OK*/)
 	return FZ_REPLY_ERROR;
 }
 
-int CHttpControlSocket::ContinueConnect(const wxIPV4address *address)
+
+int CHttpControlSocket::ContinueConnect()
 {
-	LogMessage(__TFILE__, __LINE__, this, Debug_Verbose, _T("CHttpControlSocket::ContinueConnect(%p) m_pEngine=%p"), address, m_pEngine);
+	LogMessage(__TFILE__, __LINE__, this, Debug_Verbose, _T("CHttpControlSocket::ContinueConnect() m_pEngine=%p"), m_pEngine);
 	if (GetCurrentCommandId() != cmd_connect ||
 		!m_pCurrentServer)
 	{
 		LogMessage(Debug_Warning, _T("Invalid context for call to ContinueConnect(), cmd=%d, m_pCurrentServer=%p"), GetCurrentCommandId(), m_pCurrentServer);
 		return DoClose(FZ_REPLY_INTERNALERROR);
 	}
-	
-	if (!address)
-	{
-		if (m_pCurrentServer)
-		{
-			delete m_pCurrentServer;
-			m_pCurrentServer = 0;
-		}
 
-		LogMessage(::Error, _("Invalid hostname or host not found"));
-		return ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_CRITICALERROR);
-	}
-
-	if (m_pCurOpData && m_pCurOpData->opId == cmd_connect)
-	{
-		CHttpConnectOpData *pData = static_cast<CHttpConnectOpData *>(m_pCurOpData);
-		pData->host = address->IPAddress();
-		return DoInternalConnect();
-	}
-
-	if (m_pAddress)
-		delete m_pAddress;
-	m_pAddress = new wxIPV4address(*address);
-
-	// Now we got the server's IP address for later use, e.g. transfer commands
 	ResetOperation(FZ_REPLY_OK);
 	return FZ_REPLY_OK;
 }
 
 bool CHttpControlSocket::SetAsyncRequestReply(CAsyncRequestNotification *pNotification)
 {
+	if (m_pCurOpData)
+	{
+		if (!m_pCurOpData->waitForAsyncRequest)
+		{
+			LogMessage(__TFILE__, __LINE__, this, Debug_Info, _T("Not waiting for request reply, ignoring request reply %d"), pNotification->GetRequestID());
+			return false;
+		}
+		m_pCurOpData->waitForAsyncRequest = false;
+	}
+
 	switch (pNotification->GetRequestID())
 	{
 	case reqId_fileexists:
@@ -176,85 +164,14 @@ bool CHttpControlSocket::SetAsyncRequestReply(CAsyncRequestNotification *pNotifi
 				return false;
 			}
 
-			CHttpFileTransferOpData *pData = static_cast<CHttpFileTransferOpData *>(m_pCurOpData);
-
-			if (!pData->waitForAsyncRequest)
-			{
-				LogMessage(__TFILE__, __LINE__, this, Debug_Info, _T("Not waiting for request reply, ignoring request reply %d"), pNotification->GetRequestID());
-				return false;
-			}
-			pData->waitForAsyncRequest = false;
-
 			CFileExistsNotification *pFileExistsNotification = reinterpret_cast<CFileExistsNotification *>(pNotification);
 			switch (pFileExistsNotification->overwriteAction)
 			{
-			case CFileExistsNotification::overwrite:
-				SendNextCommand();
-				break;
-			case CFileExistsNotification::overwriteNewer:
-				if (!pFileExistsNotification->localTime.IsValid() || !pFileExistsNotification->remoteTime.IsValid())
-					SendNextCommand();
-				else if (pFileExistsNotification->download && pFileExistsNotification->localTime.IsEarlierThan(pFileExistsNotification->remoteTime))
-					SendNextCommand();
-				else if (!pFileExistsNotification->download && pFileExistsNotification->localTime.IsLaterThan(pFileExistsNotification->remoteTime))
-					SendNextCommand();
-				else
-				{
-					if (pData->download)
-					{
-						wxString filename = pData->remotePath.FormatFilename(pData->remoteFile);
-						LogMessage(Status, _("Skipping download of %s"), filename.c_str());
-					}
-					else
-					{
-						LogMessage(Status, _("Skipping upload of %s"), pData->localFile.c_str());
-					}
-					ResetOperation(FZ_REPLY_OK);
-				}
-				break;
 			case CFileExistsNotification::resume:
 				ResetOperation(FZ_REPLY_CRITICALERROR | FZ_REPLY_NOTSUPPORTED);
 				break;
-			case CFileExistsNotification::rename:
-				if (pData->download)
-				{
-					wxFileName fn = pData->localFile;
-					fn.SetFullName(pFileExistsNotification->newName);
-					pData->localFile = fn.GetFullPath();
-
-					wxStructStat buf;
-					int result;
-					result = wxStat(pData->localFile, &buf);
-					if (!result)
-						pData->localFileSize = buf.st_size;
-					else
-						pData->localFileSize = -1;
-
-					if (CheckOverwriteFile() == FZ_REPLY_OK)
-						SendNextCommand();
-				}
-				else
-				{
-					ResetOperation(FZ_REPLY_CRITICALERROR | FZ_REPLY_NOTSUPPORTED);
-					break;
-				}
-				break;
-			case CFileExistsNotification::skip:
-				if (pData->download)
-				{
-					wxString filename = pData->remotePath.FormatFilename(pData->remoteFile);
-					LogMessage(Status, _("Skipping download of %s"), filename.c_str());
-				}
-				else
-				{
-					LogMessage(Status, _("Skipping upload of %s"), pData->localFile.c_str());
-				}
-				ResetOperation(FZ_REPLY_OK);
-				break;
 			default:
-				LogMessage(__TFILE__, __LINE__, this, Debug_Warning, _T("Unknown file exists action: %d"), pFileExistsNotification->overwriteAction);
-				ResetOperation(FZ_REPLY_INTERNALERROR);
-				return false;
+				return SetFileExistsAction(pFileExistsNotification);
 			}
 		}
 		break;
@@ -269,49 +186,88 @@ bool CHttpControlSocket::SetAsyncRequestReply(CAsyncRequestNotification *pNotifi
 
 void CHttpControlSocket::OnReceive()
 {
-	if (!m_pRecvBuffer)
-	{
-		m_pRecvBuffer = new char[m_recvBufferLen];
-		m_recvBufferPos = 0;
-	}
+	DoReceive();
+}
 
-	unsigned int len = m_recvBufferLen - m_recvBufferPos;
-	Read(m_pRecvBuffer + m_recvBufferPos, len);
-	if (Error())
+int CHttpControlSocket::DoReceive()
+{
+	do
 	{
-		if (LastError() != wxSOCKET_WOULDBLOCK)
-			ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
-		return;
-	}
-		
-	int read;
-	if (!(read = LastCount()))
-	{
-		ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
-		return;
-	}
+		const enum CSocket::SocketState state = m_pSocket->GetState();
+		if (state != CSocket::connected && state != CSocket::closing)
+			return 0;
 
-	m_pEngine->SetActive(true);
+		if (!m_pRecvBuffer)
+		{
+			m_pRecvBuffer = new char[m_recvBufferLen];
+			m_recvBufferPos = 0;
+		}
 
-	if (!m_pCurOpData || m_pCurOpData->opId == cmd_connect)
-	{
-		// Just ignore all further data
-		m_recvBufferPos = 0;
-		return;
+		unsigned int len = m_recvBufferLen - m_recvBufferPos;
+		int error;
+		int read = m_pSocket->Read(m_pRecvBuffer + m_recvBufferPos, len, error);
+		if (read == -1)
+		{
+			if (error != EAGAIN)
+			{
+				ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
+			}
+			return 0;
+		}
+
+		m_pEngine->SetActive(CFileZillaEngine::recv);
+
+		if (!m_pCurOpData || m_pCurOpData->opId == cmd_connect)
+		{
+			// Just ignore all further data
+			m_recvBufferPos = 0;
+			return 0;
+		}
+
+		m_recvBufferPos += read;
+
+		if (!m_pHttpOpData->m_gotHeader)
+		{
+			if (!read)
+			{
+				ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
+				return 0;
+			}
+
+			int res = ParseHeader(m_pHttpOpData);
+			if ((res & FZ_REPLY_REDIRECTED) == FZ_REPLY_REDIRECTED)
+				return FZ_REPLY_REDIRECTED;
+			if (res != FZ_REPLY_WOULDBLOCK)
+				return 0;
+		}
+		else if (m_pHttpOpData->m_transferEncoding == CHttpOpData::chunked)
+		{
+			if (!read)
+			{
+				ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
+				return 0;
+			}
+			OnChunkedData(m_pHttpOpData);
+		}
+		else
+		{
+			if (!read)
+			{
+				wxASSERT(!m_recvBufferPos);
+				ProcessData(0, 0);
+				return 0;
+			}
+			else
+			{
+				m_pHttpOpData->m_receivedData += m_recvBufferPos;
+				ProcessData(m_pRecvBuffer, m_recvBufferPos);
+				m_recvBufferPos = 0;
+			}
+		}
 	}
+	while (m_pSocket);
 
-	m_recvBufferPos += read;
-
-	if (!m_pHttpOpData->m_gotHeader)
-		ParseHeader(m_pHttpOpData);
-	else if (m_pHttpOpData->m_transferEncoding == CHttpOpData::chunked)
-		OnChunkedData(m_pHttpOpData);
-	else
-	{
-		m_pHttpOpData->m_receivedData += m_recvBufferPos;
-		ProcessData(m_pRecvBuffer, m_recvBufferPos);
-		m_recvBufferPos = 0;
-	}
+	return 0;
 }
 
 void CHttpControlSocket::OnConnect()
@@ -332,7 +288,7 @@ int CHttpControlSocket::FileTransfer(const wxString localFile, const CServerPath
 							  const CFileTransferCommand::t_transferSettings& transferSettings)
 {
 	LogMessage(Debug_Verbose, _T("CHttpControlSocket::FileTransfer()"));
-	
+
 	LogMessage(Status, _("Downloading %s"), remotePath.FormatFilename(remoteFile).c_str());
 
 	if (!download)
@@ -347,13 +303,10 @@ int CHttpControlSocket::FileTransfer(const wxString localFile, const CServerPath
 		delete m_pCurOpData;
 	}
 
-	CHttpFileTransferOpData *pData = new CHttpFileTransferOpData;
+	CHttpFileTransferOpData *pData = new CHttpFileTransferOpData(localFile, remoteFile, remotePath);
 	m_pCurOpData = pData;
 	m_pHttpOpData = pData;
 
-	pData->localFile = localFile;
-	pData->remotePath = remotePath;
-	pData->remoteFile = remoteFile;
 	pData->download = download;
 
 	if (localFile != _T(""))
@@ -367,9 +320,7 @@ int CHttpControlSocket::FileTransfer(const wxString localFile, const CServerPath
 
 		pData->pFile = new wxFile();
 
-		// Create local directory
-		wxFileName fn(pData->localFile);
-		wxFileName::Mkdir(fn.GetPath(), 0777, wxPATH_MKDIR_FULL);
+		CreateLocalDir(pData->localFile);
 
 		if (!pData->pFile->Open(pData->localFile, wxFile::write))
 		{
@@ -381,16 +332,36 @@ int CHttpControlSocket::FileTransfer(const wxString localFile, const CServerPath
 	else
 		pData->opState = filetransfer_transfer;
 
-	int res = InternalConnect(m_pAddress->IPAddress(), m_pCurrentServer->GetPort());
+	int res = InternalConnect(m_pCurrentServer->GetHost(), m_pCurrentServer->GetPort());
 	if (res != FZ_REPLY_OK)
 		return res;
 
 	return FileTransferSend();
 }
 
-int CHttpControlSocket::FileTransferSend(int prevResult /*=FZ_RESULT_OK*/)
+int CHttpControlSocket::FileTransferSubcommandResult(int prevResult)
 {
-	LogMessage(Debug_Verbose, _T("CHttpControlSocket::FileTransferSend(%d)"), prevResult);
+	LogMessage(Debug_Verbose, _T("CHttpControlSocket::FileTransferSubcommandResult(%d)"), prevResult);
+
+	if (!m_pCurOpData)
+	{
+		LogMessage(__TFILE__, __LINE__, this, Debug_Info, _T("Empty m_pCurOpData"));
+		ResetOperation(FZ_REPLY_INTERNALERROR);
+		return FZ_REPLY_ERROR;
+	}
+
+	if (prevResult != FZ_REPLY_OK)
+	{
+		ResetOperation(prevResult);
+		return FZ_REPLY_ERROR;
+	}
+
+	return FileTransferSend();
+}
+
+int CHttpControlSocket::FileTransferSend()
+{
+	LogMessage(Debug_Verbose, _T("CHttpControlSocket::FileTransferSend()"));
 
 	if (!m_pCurOpData)
 	{
@@ -401,20 +372,12 @@ int CHttpControlSocket::FileTransferSend(int prevResult /*=FZ_RESULT_OK*/)
 
 	CHttpFileTransferOpData *pData = static_cast<CHttpFileTransferOpData *>(m_pCurOpData);
 
-	if (prevResult != FZ_REPLY_OK)
-	{
-		ResetOperation(prevResult);
-		return FZ_REPLY_ERROR;
-	}
-
 	if (pData->opState == filetransfer_waitfileexists)
 	{
 		pData->opState = filetransfer_transfer;
 		pData->pFile = new wxFile();
 
-		// Create local directory
-		wxFileName fn(pData->localFile);
-		wxFileName::Mkdir(fn.GetPath(), 0777, wxPATH_MKDIR_FULL);
+		CreateLocalDir(pData->localFile);
 
 		if (!pData->pFile->Open(pData->localFile, wxFile::write))
 		{
@@ -423,7 +386,7 @@ int CHttpControlSocket::FileTransferSend(int prevResult /*=FZ_RESULT_OK*/)
 			return FZ_REPLY_ERROR;
 		}
 
-		int res = InternalConnect(m_pAddress->IPAddress(), m_pCurrentServer->GetPort());
+		int res = InternalConnect(m_pCurrentServer->GetHost(), m_pCurrentServer->GetPort());
 		if (res != FZ_REPLY_OK)
 			return res;
 	}
@@ -432,8 +395,8 @@ int CHttpControlSocket::FileTransferSend(int prevResult /*=FZ_RESULT_OK*/)
 	wxString hostWithPort;
 	if (pData->m_newLocation == _T(""))
 	{
-		location = _T("http://") + m_pCurrentServer->GetHost() + pData->remotePath.FormatFilename(pData->remoteFile).c_str();
-		hostWithPort = wxString::Format(_T("%s:%d"), m_pCurrentServer->GetHost().c_str(), m_pCurrentServer->GetPort());
+		location = _T("http://") + m_pCurrentServer->FormatHost() + pData->remotePath.FormatFilename(pData->remoteFile).c_str();
+		hostWithPort = wxString::Format(_T("%s:%d"), m_pCurrentServer->FormatHost(true).c_str(), m_pCurrentServer->GetPort());
 	}
 	else
 	{
@@ -453,7 +416,7 @@ int CHttpControlSocket::FileTransferSend(int prevResult /*=FZ_RESULT_OK*/)
 	return FZ_REPLY_WOULDBLOCK;
 }
 
-int CHttpControlSocket::InternalConnect(const wxString& host, unsigned short port)
+int CHttpControlSocket::InternalConnect(wxString host, unsigned short port)
 {
 	LogMessage(Debug_Verbose, _T("CHttpControlSocket::InternalConnect()"));
 
@@ -463,16 +426,8 @@ int CHttpControlSocket::InternalConnect(const wxString& host, unsigned short por
 	pData->port = port;
 
 	if (!IsIpAddress(host))
-	{
-		LogMessage(Status, _("Resolving IP-Address for %s"), host.c_str());
-		CAsyncHostResolver *resolver = new CAsyncHostResolver(m_pEngine, ConvertDomainName(host));
-		m_pEngine->AddNewAsyncHostResolver(resolver);
+		LogMessage(Status, _("Resolving address of %s"), host.c_str());
 
-		resolver->Create();
-		resolver->Run();
-		return FZ_REPLY_WOULDBLOCK;
-	}
-	
 	pData->host = host;
 	return DoInternalConnect();
 }
@@ -489,23 +444,19 @@ int CHttpControlSocket::DoInternalConnect()
 	}
 
 	CHttpConnectOpData *pData = static_cast<CHttpConnectOpData *>(m_pCurOpData);
-	LogMessage(Status, _("Connecting to %s:%d..."), m_pAddress->IPAddress().c_str(), pData->port);
+	LogMessage(Status, _("Connecting to %s:%d..."), pData->host.c_str(), pData->port);
 
 	if (m_pBackend)
 		delete m_pBackend;
-	m_pBackend = new CSocketBackend(this, this);
+	m_pBackend = new CSocketBackend(this, m_pSocket);
 
-	wxIPV4address addr;
-	addr.Hostname(pData->host);
-	addr.Service(pData->port);
-
-	bool res = wxSocketClient::Connect(addr, false);
-
-	if (res)
+	int res = m_pSocket->Connect(pData->host, pData->port);
+	if (!res)
 		return FZ_REPLY_OK;
-	else if (LastError() != wxSOCKET_WOULDBLOCK)
-		return ResetOperation(FZ_REPLY_ERROR);
 
+	if (res && res != EINPROGRESS)
+		return ResetOperation(FZ_REPLY_ERROR);
+	
 	return FZ_REPLY_WOULDBLOCK;
 }
 
@@ -573,7 +524,7 @@ int CHttpControlSocket::ParseHeader(CHttpOpData* pData)
 			{
 				if (m_pRecvBuffer[i + 1] != '\n')
 				{
-					LogMessage(::Error, _("Malformed reply, server not sending proper lineendings"));
+					LogMessage(::Error, _("Malformed reply, server not sending proper line endings"));
 					ResetOperation(FZ_REPLY_ERROR);
 					return FZ_REPLY_ERROR;
 				}
@@ -630,7 +581,7 @@ int CHttpControlSocket::ParseHeader(CHttpOpData* pData)
 			if (pData->m_responseCode == 305)
 			{
 				// Unsupported redirect
-				LogMessage(::Error, _("Unsuppored redirect"));
+				LogMessage(::Error, _("Unsupported redirect"));
 				ResetOperation(FZ_REPLY_ERROR);
 				return FZ_REPLY_ERROR;
 			}
@@ -682,9 +633,15 @@ int CHttpControlSocket::ParseHeader(CHttpOpData* pData)
 						ResetOperation(FZ_REPLY_ERROR);
 						return FZ_REPLY_ERROR;
 					}
-					pData->m_newHostWithPort = wxString::Format(_T("%s:%d"), host.c_str(), port);
+					pData->m_newHostWithPort = wxString::Format(_T("%s:%d"), host.c_str(), (int)port);
 
-					return InternalConnect(host, port);
+					// International domain names
+					host = ConvertDomainName(host);
+
+					int res = InternalConnect(host, port);
+					if (res == FZ_REPLY_WOULDBLOCK)
+						res |= FZ_REPLY_REDIRECTED;
+					return res;
 				}
 
 				pData->m_gotHeader = true;
@@ -705,7 +662,7 @@ int CHttpControlSocket::ParseHeader(CHttpOpData* pData)
 					}
 					return res;
 				}
-				
+
 				return FZ_REPLY_WOULDBLOCK;
 			}
 			if (m_recvBufferPos > 12 && !memcmp(m_pRecvBuffer, "Location: ", 10))
@@ -784,7 +741,7 @@ int CHttpControlSocket::OnChunkedData(CHttpOpData* pData)
 			{
 				if (p[i + 1] != '\n')
 				{
-					LogMessage(::Error, _("Malformed chunk data: %s"), _("Wrong lineendings"));
+					LogMessage(::Error, _("Malformed chunk data: %s"), _("Wrong line endings"));
 					ResetOperation(FZ_REPLY_ERROR);
 					return FZ_REPLY_ERROR;
 				}
@@ -824,7 +781,7 @@ int CHttpControlSocket::OnChunkedData(CHttpOpData* pData)
 				// We're done
 				return ProcessData(0, 0);
 			}
-			
+
 			// Ignore the trailer
 		}
 		else
@@ -840,12 +797,12 @@ int CHttpControlSocket::OnChunkedData(CHttpOpData* pData)
 				}
 				else if (*q >= 'A' && *q <= 'F')
 				{
-					pData->m_chunkData.size *= 10;
+					pData->m_chunkData.size *= 16;
 					pData->m_chunkData.size += *q - 'A' + 10;
 				}
 				else if (*q >= 'a' && *q <= 'f')
 				{
-					pData->m_chunkData.size *= 10;
+					pData->m_chunkData.size *= 16;
 					pData->m_chunkData.size += *q - 'a' + 10;
 				}
 				else if (*q == ';' || *q == ' ')
@@ -901,11 +858,19 @@ int CHttpControlSocket::ResetOperation(int nErrorCode)
 	return CControlSocket::ResetOperation(nErrorCode);
 }
 
-void CHttpControlSocket::OnClose()
+void CHttpControlSocket::OnClose(int error)
 {
-	char tmp[1];
-	for (Peek(tmp, 1); !Error() && LastCount(); Peek(tmp, 1))
-		OnReceive();
+	LogMessage(Debug_Verbose, _T("CHttpControlSocket::OnClose(%d)"), error);
+
+	if (error)
+	{
+		LogMessage(::Error, _("Disconnected from server: %s"), CSocket::GetErrorDescription(error).c_str());
+		ResetOperation(FZ_REPLY_ERROR | FZ_REPLY_DISCONNECTED);
+		return;
+	}
+	
+	if (DoReceive() == FZ_REPLY_REDIRECTED)
+		return; // Socket got closed already
 
 	// HTTP socket isn't connected outside operations
 	if (!m_pCurOpData)
@@ -947,7 +912,7 @@ void CHttpControlSocket::ResetHttpData(CHttpOpData* pData)
 {
 	wxASSERT(pData);
 
-	delete m_pRecvBuffer;
+	delete [] m_pRecvBuffer;
 	m_pRecvBuffer = 0;
 
 	pData->m_gotHeader = false;
@@ -982,3 +947,26 @@ int CHttpControlSocket::ProcessData(char* p, int len)
 
 	return res;
 }
+
+int CHttpControlSocket::ParseSubcommandResult(int prevResult)
+{
+	LogMessage(Debug_Verbose, _T("CHttpControlSocket::SendNextCommand(%d)"), prevResult);
+	if (!m_pCurOpData)
+	{
+		LogMessage(__TFILE__, __LINE__, this, Debug_Warning, _T("SendNextCommand called without active operation"));
+		ResetOperation(FZ_REPLY_ERROR);
+		return FZ_REPLY_ERROR;
+	}
+
+	switch (m_pCurOpData->opId)
+	{
+	case cmd_transfer:
+		return FileTransferSubcommandResult(prevResult);
+	default:
+		LogMessage(__TFILE__, __LINE__, this, ::Debug_Warning, _T("Unknown opID (%d) in SendNextCommand"), m_pCurOpData->opId);
+		ResetOperation(FZ_REPLY_INTERNALERROR);
+		break;
+	}
+
+	return FZ_REPLY_ERROR;
+}
