@@ -6,11 +6,7 @@
 #include <gnutls/x509.h>
 #include <errno.h>
 
-BEGIN_EVENT_TABLE(CTlsSocket, wxEvtHandler)
-EVT_SOCKET(wxID_ANY, CTlsSocket::OnSocketEvent)
-END_EVENT_TABLE()
-
-CTlsSocket::CTlsSocket(wxEvtHandler* pEvtHandler, wxSocketBase* pSocket, CControlSocket* pOwner)
+CTlsSocket::CTlsSocket(CSocketEventHandler* pEvtHandler, CSocket* pSocket, CControlSocket* pOwner)
 	: CBackend(pEvtHandler), m_pOwner(pOwner)
 {
 	m_pSocket = pSocket;
@@ -19,20 +15,17 @@ CTlsSocket::CTlsSocket(wxEvtHandler* pEvtHandler, wxSocketBase* pSocket, CContro
 	m_session = 0;
 	m_initialized = false;
 	m_certCredentials = 0;
-	
+
 	m_canReadFromSocket = true;
 	m_canWriteToSocket = true;
 	m_canCheckCloseSocket = false;
 
-	m_canTriggerRead = true;
+	m_canTriggerRead = false;
 	m_canTriggerWrite = true;
 
 	m_tlsState = noconn;
 
-	m_lastError = 0;
-	m_lastCount = 0;
-	m_lastSuccessful = false;
-
+	m_lastReadFailed = true;
 	m_lastWriteFailed = false;
 	m_writeSkip = 0;
 
@@ -45,6 +38,8 @@ CTlsSocket::CTlsSocket(wxEvtHandler* pEvtHandler, wxSocketBase* pSocket, CContro
 	m_implicitTrustedCert.size = 0;
 
 	m_shutdown_requested = false;
+
+	m_socket_eof = false;
 }
 
 CTlsSocket::~CTlsSocket()
@@ -56,13 +51,14 @@ CTlsSocket::~CTlsSocket()
 bool CTlsSocket::Init()
 {
 	// This function initializes GnuTLS
+	m_initialized = true;
 	int res = gnutls_global_init();
 	if (res)
 	{
 		LogError(res);
+		Uninit();
 		return false;
 	}
-	m_initialized = true;
 
 	res = gnutls_certificate_allocate_credentials(&m_certCredentials);
 	if (res < 0)
@@ -87,7 +83,9 @@ bool CTlsSocket::Init()
 		Uninit();
 		return false;
 	}
-	
+
+	gnutls_dh_set_prime_bits(m_session, 512);
+
 	// Set which type of certificates we accept
 	const int cert_type_priority[] = { GNUTLS_CRT_X509, 0 };
 	gnutls_certificate_type_set_priority(m_session, cert_type_priority);
@@ -204,12 +202,12 @@ ssize_t CTlsSocket::PushFunction(const void* data, size_t len)
 		return -1;
 	}
 
-	m_pSocketBackend->Write(data, len);
+	int error;
+	int written = m_pSocketBackend->Write(data, len, error);
 
-	if (m_pSocketBackend->Error())
+	if (written < 0)
 	{
-		const int error = m_pSocketBackend->LastError();
-		if (error == wxSOCKET_WOULDBLOCK)
+		if (error == EAGAIN)
 		{
 			m_canWriteToSocket = false;
 			gnutls_transport_set_errno(m_session, EAGAIN);
@@ -220,7 +218,7 @@ ssize_t CTlsSocket::PushFunction(const void* data, size_t len)
 		return -1;
 	}
 
-	return m_pSocketBackend->LastCount();
+	return written;
 }
 
 ssize_t CTlsSocket::PullFunction(void* data, size_t len)
@@ -240,80 +238,76 @@ ssize_t CTlsSocket::PullFunction(void* data, size_t len)
 		return -1;
 	}
 
-	m_canReadFromSocket = false;
-
-	m_pSocketBackend->Read(data, len);
-	if (m_pSocketBackend->Error())
+	int error;
+	int read = m_pSocketBackend->Read(data, len, error);
+	if (read < 0)
 	{
-		if (m_pSocketBackend->LastError() == wxSOCKET_WOULDBLOCK)
+		if (error == EAGAIN)
 		{
+			m_canReadFromSocket = false;
 			if (m_canCheckCloseSocket && !m_pSocketBackend->IsWaiting(CRateLimiter::inbound))
 			{
-				wxSocketEvent evt(m_pSocketBackend->GetId());
-				evt.m_event = wxSOCKET_LOST;
-				wxPostEvent(this, evt);
+				CSocketEvent *evt = new CSocketEvent(this, m_pSocketBackend, CSocketEvent::close);
+				CSocketEventDispatcher::Get().SendEvent(evt);
 			}
-
-			gnutls_transport_set_errno(m_session, EAGAIN);
-			return -1;
 		}
 
-		gnutls_transport_set_errno(m_session, 0);
+		gnutls_transport_set_errno(m_session, error);
 		return -1;
 	}
 
 	if (m_canCheckCloseSocket)
 	{
-		wxSocketEvent evt(m_pSocketBackend->GetId());
-		evt.m_event = wxSOCKET_LOST;
-		wxPostEvent(this, evt);
+		CSocketEvent *evt = new CSocketEvent(this, m_pSocketBackend, CSocketEvent::close);
+		CSocketEventDispatcher::Get().SendEvent(evt);
 	}
 
-	return m_pSocketBackend->LastCount();
+	if (!read)
+		m_socket_eof = true;
+
+	return read;
 }
 
-void CTlsSocket::OnSocketEvent(wxSocketEvent& event)
+void CTlsSocket::OnSocketEvent(CSocketEvent& event)
 {
 	wxASSERT(m_pSocket);
 	if (!m_session)
 		return;
 
-	if (event.GetId() != m_pSocketBackend->GetId())
-		return;
-
-	switch (event.GetSocketEvent())
+	switch (event.GetType())
 	{
-	case wxSOCKET_INPUT:
+	case CSocketEvent::read:
 		OnRead();
 		break;
-	case wxSOCKET_OUTPUT:
+	case CSocketEvent::write:
 		OnSend();
 		break;
-	case wxSOCKET_LOST:
+	case CSocketEvent::close:
 		{
 			m_canCheckCloseSocket = true;
 			char tmp[100];
-			m_pSocketBackend->Peek(&tmp, 100);
-			if (!m_pSocketBackend->Error())
+			int error;
+			int peeked = m_pSocketBackend->Peek(&tmp, 100, error);
+			if (peeked >= 0)
 			{
-				int lastCount = m_pSocketBackend->LastCount();
-
-				if (lastCount)
-					m_pOwner->LogMessage(Debug_Verbose, _T("CTlsSocket::OnSocketEvent(): pending data, postponing wxSOCKET_LOST"));
+				if (peeked > 0)
+					m_pOwner->LogMessage(Debug_Verbose, _T("CTlsSocket::OnSocketEvent(): pending data, postponing close event"));
 				else
+				{
+					m_socket_eof = true;
 					m_socketClosed = true;
+				}
 				OnRead();
 
-				if (lastCount)
+				if (peeked)
 					return;
 			}
 
-			m_pOwner->LogMessage(Debug_Info, _T("CTlsSocket::OnSocketEvent(): wxSOCKET_LOST received"));
+			m_pOwner->LogMessage(Debug_Info, _T("CTlsSocket::OnSocketEvent(): close event received"));
 
 			//Uninit();
-			wxSocketEvent evt(GetId());
-			evt.m_event = wxSOCKET_LOST;
-			wxPostEvent(m_pEvtHandler, evt);
+			CSocketEvent *evt = new CSocketEvent(m_pEvtHandler, this, CSocketEvent::close);
+			CSocketEventDispatcher::Get().SendEvent(evt);
 		}
 		break;
 	default:
@@ -330,16 +324,17 @@ void CTlsSocket::OnRead()
 	if (!m_session)
 		return;
 
-	if (gnutls_record_get_direction(m_session))
+	const int direction = gnutls_record_get_direction(m_session);
+	if (direction && !m_lastReadFailed)
 		return;
-		
+
 	if (m_tlsState == handshake)
 		Handshake();
 	if (m_tlsState == closing)
 		ContinueShutdown();
 	else if (m_tlsState == conn)
 	{
-		CheckResumeFailedWrite();
+		CheckResumeFailedReadWrite();
 		TriggerEvents();
 	}
 }
@@ -353,7 +348,8 @@ void CTlsSocket::OnSend()
 	if (!m_session)
 		return;
 
-	if (!gnutls_record_get_direction(m_session))
+	const int direction = gnutls_record_get_direction(m_session);
+	if (!direction && !m_lastWriteFailed)
 		return;
 
 	if (m_tlsState == handshake)
@@ -362,12 +358,51 @@ void CTlsSocket::OnSend()
 		ContinueShutdown();
 	else if (m_tlsState == conn)
 	{
-		CheckResumeFailedWrite();
+		CheckResumeFailedReadWrite();
 		TriggerEvents();
 	}
 }
 
-int CTlsSocket::Handshake(const CTlsSocket* pPrimarySocket /*=0*/)
+bool CTlsSocket::CopySessionData(const CTlsSocket* pPrimarySocket)
+{
+	size_t session_data_size = 0;
+
+	// Get buffer size
+	int res = gnutls_session_get_data(pPrimarySocket->m_session, 0, &session_data_size);
+	if (res)
+	{
+		m_pOwner->LogMessage(Debug_Info, _T("gnutls_session_get_data on primary socket failed: %d"), res);
+		return false;
+	}
+
+	// Get session data
+	char *session_data = new char[session_data_size];
+	res = gnutls_session_get_data(pPrimarySocket->m_session, session_data, &session_data_size);
+	if (res)
+	{
+		delete [] session_data;
+		m_pOwner->LogMessage(Debug_Info, _T("gnutls_session_get_data on primary socket failed: %d"), res);
+		return false;
+	}
+
+	// Set session data
+	res = gnutls_session_set_data(m_session, session_data, session_data_size);
+	delete [] session_data;
+	if (res)
+	{
+		m_pOwner->LogMessage(Debug_Info, _T("gnutls_session_set_data failed: %d"), res);
+		return false;
+	}
+
+	return true;
+}
+
+bool CTlsSocket::ResumedSession() const
+{
+	return gnutls_session_is_resumed(m_session) != 0;
+}
+
+int CTlsSocket::Handshake(const CTlsSocket* pPrimarySocket /*=0*/, bool try_resume /*=false*/)
 {
 	m_pOwner->LogMessage(Debug_Verbose, _T("CTlsSocket::Handshake()"));
 	wxASSERT(m_session);
@@ -385,6 +420,9 @@ int CTlsSocket::Handshake(const CTlsSocket* pPrimarySocket /*=0*/)
 			memcpy(m_implicitTrustedCert.data, cert_list[0].data, cert_list[0].size);
 			m_implicitTrustedCert.size = cert_list[0].size;
 		}
+
+		if (try_resume)
+			CopySessionData(pPrimarySocket);
 	}
 
 	int res = gnutls_handshake(m_session);
@@ -392,19 +430,11 @@ int CTlsSocket::Handshake(const CTlsSocket* pPrimarySocket /*=0*/)
 	{
 		m_pOwner->LogMessage(Debug_Info, _T("Handshake successful"));
 
-		wxString cipherName;
-		const char* cipher = gnutls_cipher_get_name(gnutls_cipher_get(m_session));
-		if (cipher)
-			cipherName = wxString(cipher, wxConvUTF8);
-		else
-			cipherName = _T("unknown");
+		if (ResumedSession())
+			m_pOwner->LogMessage(Debug_Info, _T("Session resumed"));
 
-		wxString macName;
-		const char* mac = gnutls_mac_get_name(gnutls_mac_get(m_session));
-		if (mac)
-			macName = wxString(mac, wxConvUTF8);
-		else
-			macName = _T("unknown");
+		const wxString& cipherName = GetCipherName();
+		const wxString& macName = GetMacName();
 
 		m_pOwner->LogMessage(Debug_Info, _T("Cipher: %s, MAC: %s"), cipherName.c_str(), macName.c_str());
 
@@ -412,21 +442,13 @@ int CTlsSocket::Handshake(const CTlsSocket* pPrimarySocket /*=0*/)
 		if (res != FZ_REPLY_OK)
 			return res;
 
-		m_tlsState = conn;
-
-		wxSocketEvent evt(GetId());
-		evt.m_event = wxSOCKET_CONNECTION;
-		wxPostEvent(m_pEvtHandler, evt);
-		TriggerEvents();
-
 		if (m_shutdown_requested)
 		{
-			Shutdown();
-			if (!Error() || LastError() != wxSOCKET_WOULDBLOCK)
+			int error = Shutdown();
+			if (!error || error != EAGAIN)
 			{
-				wxSocketEvent evt(GetId());
-				evt.m_event = wxSOCKET_LOST;
-				wxPostEvent(m_pEvtHandler, evt);
+				CSocketEvent *evt = new CSocketEvent(m_pEvtHandler, this, CSocketEvent::close);
+				CSocketEventDispatcher::Get().SendEvent(evt);
 			}
 		}
 
@@ -435,33 +457,34 @@ int CTlsSocket::Handshake(const CTlsSocket* pPrimarySocket /*=0*/)
 	else if (res == GNUTLS_E_AGAIN || res == GNUTLS_E_INTERRUPTED)
 		return FZ_REPLY_WOULDBLOCK;
 
-	Failure(res);
+	Failure(res, 0);
 
 	return FZ_REPLY_ERROR;
 }
 
-void CTlsSocket::Read(void *buffer, unsigned int len)
+int CTlsSocket::Read(void *buffer, unsigned int len, int& error)
 {
-	if (m_tlsState == handshake)
+	if (m_tlsState == handshake || m_tlsState == verifycert)
 	{
-		m_lastError = wxSOCKET_WOULDBLOCK;
-		m_lastSuccessful = false;
-		return;
+		error = EAGAIN;
+		return -1;
 	}
 	else if (m_tlsState != conn)
 	{
-		m_lastError = wxSOCKET_IOERR;
-		m_lastSuccessful = false;
-		return;
+		error = ENOTCONN;
+		return -1;
+	}
+	else if (m_lastReadFailed)
+	{
+		error = EAGAIN;
+		return -1;
 	}
-
-	m_canTriggerRead = true;
 
 	if (m_peekDataLen)
 	{
 		unsigned int min = wxMin(len, m_peekDataLen);
 		memcpy(buffer, m_peekData, min);
-		
+
 		if (min == m_peekDataLen)
 		{
 			m_peekDataLen = 0;
@@ -476,63 +499,66 @@ void CTlsSocket::Read(void *buffer, unsigned int len)
 
 		TriggerEvents();
 
-		m_lastSuccessful = true;
-		m_lastCount = min;
-		return;
+		error = 0;
+		return min;
 	}
-	
+
 	int res = gnutls_record_recv(m_session, buffer, len);
 	if (res >= 0)
 	{
-		m_lastSuccessful = true;
-		m_lastCount = res;
-
 		if (res > 0)
 			TriggerEvents();
+		else
+		{
+			// Peer did already initiate a shutdown, reply to it
+			gnutls_bye(m_session, GNUTLS_SHUT_WR);
+			// Note: Theoretically this could return a write error.
+			// But we ignore it, since it is perfectly valid for peer
+			// to close the connection after sending its shutdown
+			// notification.
+		}
 
-		return;
+		error = 0;
+		return res;
 	}
 
 	if (res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN)
-		m_lastError = wxSOCKET_WOULDBLOCK;
+	{
+		error = EAGAIN;
+		m_lastReadFailed = true;
+	}
 	else
 	{
-		Failure(res);
-		m_lastError = wxSOCKET_IOERR;
+		Failure(res, 0);
+		error = ECONNABORTED;
 	}
 
-	m_lastSuccessful = false;
-	return;
+	return -1;
 }
 
-void CTlsSocket::Write(const void *buffer, unsigned int len)
+int CTlsSocket::Write(const void *buffer, unsigned int len, int& error)
 {
-	if (m_tlsState == handshake)
+	if (m_tlsState == handshake || m_tlsState == verifycert)
 	{
-		m_lastError = wxSOCKET_WOULDBLOCK;
-		m_lastSuccessful = false;
-		return;
+		error = EAGAIN;
+		return -1;
 	}
 	else if (m_tlsState != conn)
 	{
-		m_lastError = wxSOCKET_IOERR;
-		m_lastSuccessful = false;
-		return;
+		error = ENOTCONN;
+		return -1;
 	}
 
 	if (m_lastWriteFailed)
 	{
-		m_lastError = wxSOCKET_WOULDBLOCK;
-		m_lastSuccessful = false;
-		return;
+		error = EAGAIN;
+		return -1;
 	}
 
 	if (m_writeSkip >= len)
 	{
 		m_writeSkip -= len;
-		m_lastCount = len;
-		m_lastSuccessful = true;
-		return;
+		return len;
 	}
 
 	len -= m_writeSkip;
@@ -541,34 +567,35 @@ void CTlsSocket::Write(const void *buffer, unsigned int len)
 	int res = gnutls_record_send(m_session, buffer, len);
 	if (res >= 0)
 	{
-		m_lastSuccessful = true;
-		m_lastCount = res + m_writeSkip;
+		error = 0;
+		int written = res + m_writeSkip;
 		m_writeSkip = 0;
 
 		TriggerEvents();
-		return;
+		return written;
 	}
 
 	if (res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN)
 	{
 		if (m_writeSkip)
 		{
-			m_lastSuccessful = true;
-			m_lastCount = m_writeSkip;
+			error = 0;
+			int written = m_writeSkip;
 			m_writeSkip = 0;
+			return written;
 		}
 		else
 		{
-			m_lastSuccessful = false;
-			m_lastError = wxSOCKET_WOULDBLOCK;
+			error = EAGAIN;
 			m_lastWriteFailed = true;
+			return -1;
 		}
 	}
 	else
 	{
-		m_lastSuccessful = false;
-		Failure(res);
-		m_lastError = wxSOCKET_IOERR;
+		Failure(res, 0);
+		error = ECONNABORTED;
+		return -1;
 	}
 }
 
@@ -579,32 +606,30 @@ void CTlsSocket::TriggerEvents()
 
 	if (m_canTriggerRead)
 	{
-		wxSocketEvent evt(GetId());
-		evt.m_event = wxSOCKET_INPUT;
-		wxPostEvent(m_pEvtHandler, evt);
+		CSocketEvent *evt = new CSocketEvent(m_pEvtHandler, this, CSocketEvent::read);
+		CSocketEventDispatcher::Get().SendEvent(evt);
 		m_canTriggerRead = false;
 	}
 
 	if (m_canTriggerWrite)
 	{
-		wxSocketEvent evt(GetId());
-		evt.m_event = wxSOCKET_OUTPUT;
-		wxPostEvent(m_pEvtHandler, evt);
+		CSocketEvent *evt = new CSocketEvent(m_pEvtHandler, this, CSocketEvent::write);
+		CSocketEventDispatcher::Get().SendEvent(evt);
 		m_canTriggerWrite = false;
 	}
 }
 
-void CTlsSocket::CheckResumeFailedWrite()
+void CTlsSocket::CheckResumeFailedReadWrite()
 {
 	if (m_lastWriteFailed)
 	{
 		int res = gnutls_record_send(m_session, 0, 0);
 		if (res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN)
 			return;
-		
+
 		if (res < 0)
 		{
-			Failure(res);
+			Failure(res, ECONNABORTED);
 			return;
 		}
 
@@ -614,73 +639,97 @@ void CTlsSocket::CheckResumeFailedWrite()
 
 		wxASSERT(GNUTLS_E_INVALID_REQUEST == gnutls_record_send(m_session, 0, 0));
 	}
+	if (m_lastReadFailed)
+	{
+		wxASSERT(!m_peekData);
+
+		m_peekDataLen = 65536;
+		m_peekData = new char[m_peekDataLen];
+
+		int res = gnutls_record_recv(m_session, m_peekData, m_peekDataLen);
+		if (res < 0)
+		{
+			m_peekDataLen = 0;
+			delete [] m_peekData;
+			m_peekData = 0;
+			if (res != GNUTLS_E_INTERRUPTED && res != GNUTLS_E_AGAIN)
+				Failure(res, ECONNABORTED);
+			return;
+		}
+
+		if (!res)
+		{
+			m_peekDataLen = 0;
+			delete [] m_peekData;
+			m_peekData = 0;
+		}
+		else
+			m_peekDataLen = res;
+
+		m_lastReadFailed = false;
+		m_canTriggerRead = true;
+	}
 }
 
-void CTlsSocket::Failure(int code)
+void CTlsSocket::Failure(int code, int socket_error)
 {
 	if (code)
+	{
 		LogError(code);
+		if (code == GNUTLS_E_UNEXPECTED_PACKET_LENGTH && m_socket_eof)
+			m_pOwner->LogMessage(Status, _("Server did not properly shut down TLS connection"));
+	}
 	Uninit();
 
-	wxSocketEvent evt(GetId());
-	evt.m_event = wxSOCKET_LOST;
-	wxPostEvent(m_pEvtHandler, evt);
+	if (socket_error)
+	{
+		CSocketEvent *evt = new CSocketEvent(m_pEvtHandler, this, CSocketEvent::close, socket_error);
+		CSocketEventDispatcher::Get().SendEvent(evt);
+	}
 }
 
-void CTlsSocket::Peek(void *buffer, unsigned int len)
+int CTlsSocket::Peek(void *buffer, unsigned int len, int& error)
 {
 	if (m_peekData)
 	{
 		int min = wxMin(len, m_peekDataLen);
 		memcpy(buffer, m_peekData, min);
-		
-		m_lastCount = min;
-		m_lastSuccessful = true;
-		return;
+
+		error = 0;
+		return min;
 	}
 
-	Read(buffer, len);
-	if (Error() || !LastCount())
-		return;
+	int read = Read(buffer, len, error);
+	if (read <= 0)
+		return read;
 
-	m_peekDataLen = LastCount();
+	m_peekDataLen = read;
 	m_peekData = new char[m_peekDataLen];
 	memcpy(m_peekData, buffer, m_peekDataLen);
+
+	return read;
 }
 
-void CTlsSocket::Shutdown()
+int CTlsSocket::Shutdown()
 {
 	m_pOwner->LogMessage(Debug_Verbose, _T("CTlsSocket::Shutdown()"));
 
 	if (m_tlsState == closed)
-	{
-		m_lastSuccessful = true;
-		return;
-	}
+		return 0;
 
 	if (m_tlsState == closing)
-	{
-		m_lastSuccessful = false;
-		m_lastError = wxSOCKET_WOULDBLOCK;
-		return;
-	}
+		return EAGAIN;
 
 	if (m_tlsState == handshake || m_tlsState == verifycert)
 	{
 		// Shutdown during handshake is not a good idea.
 		m_pOwner->LogMessage(Debug_Verbose, _T("Shutdown during handshake, postponing"));
 		m_shutdown_requested = true;
-		m_lastSuccessful = false;
-		m_lastError = wxSOCKET_WOULDBLOCK;
-		return;
+		return EAGAIN;
 	}
 
 	if (m_tlsState != conn)
-	{
-		m_lastSuccessful = false;
-		m_lastError = wxSOCKET_IOERR;
-		return;
-	}
+		return ECONNABORTED;
 
 	m_tlsState = closing;
 
@@ -688,18 +737,14 @@ void CTlsSocket::Shutdown()
 	if (!res)
 	{
 		m_tlsState = closed;
-		m_lastSuccessful = true;
-		return;
+		return 0;
 	}
 
-	m_lastSuccessful = false;
 	if (res == GNUTLS_E_INTERRUPTED || res == GNUTLS_E_AGAIN)
-		m_lastError = wxSOCKET_WOULDBLOCK;
-	else
-	{
-		Failure(res);
-		m_lastError = wxSOCKET_IOERR;
-	}
+		return EAGAIN;
+
+	Failure(res, 0);
+	return ECONNABORTED;
 }
 
 void CTlsSocket::ContinueShutdown()
@@ -711,15 +756,14 @@ void CTlsSocket::ContinueShutdown()
 	{
 		m_tlsState = closed;
 
-		wxSocketEvent evt(GetId());
-		evt.m_event = wxSOCKET_LOST;
-		wxPostEvent(m_pEvtHandler, evt);
+		CSocketEvent *evt = new CSocketEvent(m_pEvtHandler, this, CSocketEvent::close);
+		CSocketEventDispatcher::Get().SendEvent(evt);
 
 		return;
 	}
 
 	if (res != GNUTLS_E_INTERRUPTED && res != GNUTLS_E_AGAIN)
-		Failure(res);
+		Failure(res, ECONNABORTED);
 }
 
 void CTlsSocket::TrustCurrentCert(bool trusted)
@@ -734,15 +778,23 @@ void CTlsSocket::TrustCurrentCert(bool trusted)
 	{
 		m_tlsState = conn;
 
-		wxSocketEvent evt(GetId());
-		evt.m_event = wxSOCKET_CONNECTION;
-		wxPostEvent(m_pEvtHandler, evt);
+		if (m_lastWriteFailed)
+			m_lastWriteFailed = false;
+		CheckResumeFailedReadWrite();
+
+		if (m_tlsState == conn)
+		{
+			CSocketEvent *evt = new CSocketEvent(m_pEvtHandler, this, CSocketEvent::connection);
+			CSocketEventDispatcher::Get().SendEvent(evt);
+		}
+
 		TriggerEvents();
+
 		return;
 	}
 
 	m_pOwner->LogMessage(::Error, _("Remote certificate not trusted."));
-	Failure(0);
+	Failure(0, ECONNABORTED);
 }
 
 static wxString bin2hex(const unsigned char* in, size_t size)
@@ -771,16 +823,16 @@ int CTlsSocket::VerifyCertificate()
 	if (gnutls_certificate_type_get(m_session) != GNUTLS_CRT_X509)
 	{
 		m_pOwner->LogMessage(::Debug_Warning, _T("Unsupported certificate type"));
-		Failure(0);
+		Failure(0, ECONNABORTED);
 		return FZ_REPLY_ERROR;
 	}
 
 	unsigned int cert_list_size;
-	const gnutls_datum_t* const cert_list = gnutls_certificate_get_peers(m_session, &cert_list_size);
+	const gnutls_datum_t* cert_list = gnutls_certificate_get_peers(m_session, &cert_list_size);
 	if (!cert_list || !cert_list_size)
 	{
 		m_pOwner->LogMessage(::Debug_Warning, _T("gnutls_certificate_get_peers returned no certificates"));
-		Failure(0);
+		Failure(0, ECONNABORTED);
 		return FZ_REPLY_ERROR;
 	}
 
@@ -790,134 +842,146 @@ int CTlsSocket::VerifyCertificate()
 			memcmp(m_implicitTrustedCert.data, cert_list[0].data, cert_list[0].size))
 		{
 			m_pOwner->LogMessage(::Error, _("Primary connection and data connection certificates don't match."));
-			Failure(0);
+			Failure(0, ECONNABORTED);
 			return FZ_REPLY_ERROR;
 		}
-		
+
 		TrustCurrentCert(true);
+
+		if (m_tlsState != conn)
+			return FZ_REPLY_ERROR;
 		return FZ_REPLY_OK;
 	}
 
-	gnutls_x509_crt_t cert;
-	if (gnutls_x509_crt_init(&cert))
+	std::vector<CCertificate> certificates;
+	for (unsigned int i = 0; i < cert_list_size; i++)
 	{
-		m_pOwner->LogMessage(::Debug_Warning, _T("gnutls_x509_crt_init failed"));
-		Failure(0);
-		return FZ_REPLY_ERROR;
-	}
-
-	if (gnutls_x509_crt_import(cert, cert_list, GNUTLS_X509_FMT_DER))
-	{
-		m_pOwner->LogMessage(::Debug_Warning, _T("gnutls_x509_crt_import failed"));
-		Failure(0);
-		gnutls_x509_crt_deinit(cert);
-		return FZ_REPLY_ERROR;
-	}
-
-	wxDateTime expirationTime = gnutls_x509_crt_get_expiration_time(cert);
-	wxDateTime activationTime = gnutls_x509_crt_get_activation_time(cert);
-
-	// Get the serial number of the certificate
-	unsigned char buffer[40];
-	size_t size = sizeof(buffer);
-	int res = gnutls_x509_crt_get_serial(cert, buffer, &size);
-
-	wxString serial = bin2hex(buffer, size);
-
-	unsigned int bits;
-	int algo = gnutls_x509_crt_get_pk_algorithm(cert, &bits);
-
-	wxString algoName;
-	const char* pAlgo = gnutls_pk_algorithm_get_name((gnutls_pk_algorithm_t)algo);
-	if (pAlgo)
-		algoName = wxString(pAlgo, wxConvUTF8);
-
-	//int version = gnutls_x509_crt_get_version(cert);
-	
-	wxString subject, issuer;
-
-	size = 0;
-	res = gnutls_x509_crt_get_dn(cert, 0, &size);
-	if (size)
-	{
-		char* dn = new char[size + 1];
-		dn[size] = 0;
-		if (!(res = gnutls_x509_crt_get_dn(cert, dn, &size)))
+		gnutls_x509_crt_t cert;
+		if (gnutls_x509_crt_init(&cert))
 		{
+			m_pOwner->LogMessage(::Debug_Warning, _T("gnutls_x509_crt_init failed"));
+			Failure(0, ECONNABORTED);
+			return FZ_REPLY_ERROR;
+		}
+
+		if (gnutls_x509_crt_import(cert, cert_list, GNUTLS_X509_FMT_DER))
+		{
+			m_pOwner->LogMessage(::Debug_Warning, _T("gnutls_x509_crt_import failed"));
+			Failure(0, ECONNABORTED);
+			gnutls_x509_crt_deinit(cert);
+			return FZ_REPLY_ERROR;
+		}
+
+		wxDateTime expirationTime = gnutls_x509_crt_get_expiration_time(cert);
+		wxDateTime activationTime = gnutls_x509_crt_get_activation_time(cert);
+
+		// Get the serial number of the certificate
+		unsigned char buffer[40];
+		size_t size = sizeof(buffer);
+		int res = gnutls_x509_crt_get_serial(cert, buffer, &size);
+
+		wxString serial = bin2hex(buffer, size);
+
+		unsigned int bits;
+		int algo = gnutls_x509_crt_get_pk_algorithm(cert, &bits);
+
+		wxString algoName;
+		const char* pAlgo = gnutls_pk_algorithm_get_name((gnutls_pk_algorithm_t)algo);
+		if (pAlgo)
+			algoName = wxString(pAlgo, wxConvUTF8);
+
+		//int version = gnutls_x509_crt_get_version(cert);
+
+		wxString subject, issuer;
+
+		size = 0;
+		res = gnutls_x509_crt_get_dn(cert, 0, &size);
+		if (size)
+		{
+			char* dn = new char[size + 1];
 			dn[size] = 0;
-			subject = wxString(dn, wxConvUTF8);
+			if (!(res = gnutls_x509_crt_get_dn(cert, dn, &size)))
+			{
+				dn[size] = 0;
+				subject = wxString(dn, wxConvUTF8);
+			}
+			else
+				LogError(res);
+			delete [] dn;
 		}
 		else
 			LogError(res);
-		delete [] dn;
-	}
-	else
-		LogError(res);
-	if (subject == _T(""))
-	{
-		m_pOwner->LogMessage(::Debug_Warning, _T("gnutls_x509_get_dn failed"));
-		Failure(0);
-		gnutls_x509_crt_deinit(cert);
-		return FZ_REPLY_ERROR;
-	}
-
-	size = 0;
-	res = gnutls_x509_crt_get_issuer_dn(cert, 0, &size);
-	if (size)
-	{
-		char* dn = new char[size + 1];
-		dn[size] = 0;
-		if (!(res = gnutls_x509_crt_get_issuer_dn(cert, dn, &size)))
+		if (subject == _T(""))
 		{
+			m_pOwner->LogMessage(::Debug_Warning, _T("gnutls_x509_get_dn failed"));
+			Failure(0, ECONNABORTED);
+			gnutls_x509_crt_deinit(cert);
+			return FZ_REPLY_ERROR;
+		}
+
+		size = 0;
+		res = gnutls_x509_crt_get_issuer_dn(cert, 0, &size);
+		if (size)
+		{
+			char* dn = new char[++size + 1];
 			dn[size] = 0;
-			issuer = wxString(dn, wxConvUTF8);
+			if (!(res = gnutls_x509_crt_get_issuer_dn(cert, dn, &size)))
+			{
+				dn[size] = 0;
+				issuer = wxString(dn, wxConvUTF8);
+			}
+			else
+				LogError(res);
+			delete [] dn;
 		}
 		else
 			LogError(res);
-		delete [] dn;
-	}
-	else
-		LogError(res);
-	if (issuer == _T(""))
-	{
-		m_pOwner->LogMessage(::Debug_Warning, _T("gnutls_x509_get_issuer_dn failed"));
-		Failure(0);
-		gnutls_x509_crt_deinit(cert);
-		return FZ_REPLY_ERROR;
-	}
+		if (issuer == _T(""))
+		{
+			m_pOwner->LogMessage(::Debug_Warning, _T("gnutls_x509_get_issuer_dn failed"));
+			Failure(0, ECONNABORTED);
+			gnutls_x509_crt_deinit(cert);
+			return FZ_REPLY_ERROR;
+		}
 
-	wxString fingerprint_md5;
-	wxString fingerprint_sha1;
+		wxString fingerprint_md5;
+		wxString fingerprint_sha1;
 
-	unsigned char digest[100];
-	size = sizeof(digest) - 1;
-	if (!gnutls_x509_crt_get_fingerprint(cert, GNUTLS_DIG_MD5, digest, &size))
-	{
-		digest[size] = 0;
-		fingerprint_md5 = bin2hex(digest, size);
-	}
-	size = sizeof(digest) - 1;
-	if (!gnutls_x509_crt_get_fingerprint(cert, GNUTLS_DIG_SHA1, digest, &size))
-	{
-		digest[size] = 0;
-		fingerprint_sha1 = bin2hex(digest, size);
+		unsigned char digest[100];
+		size = sizeof(digest) - 1;
+		if (!gnutls_x509_crt_get_fingerprint(cert, GNUTLS_DIG_MD5, digest, &size))
+		{
+			digest[size] = 0;
+			fingerprint_md5 = bin2hex(digest, size);
+		}
+		size = sizeof(digest) - 1;
+		if (!gnutls_x509_crt_get_fingerprint(cert, GNUTLS_DIG_SHA1, digest, &size))
+		{
+			digest[size] = 0;
+			fingerprint_sha1 = bin2hex(digest, size);
+		}
+
+		certificates.push_back(CCertificate(
+			cert_list->data, cert_list->size,
+			activationTime, expirationTime,
+			serial,
+			algoName, bits,
+			fingerprint_md5,
+			fingerprint_sha1,
+			subject,
+			issuer));
+
+		cert_list++;
 	}
 
 	CCertificateNotification *pNotification = new CCertificateNotification(
 		m_pOwner->GetCurrentServer()->GetHost(),
 		m_pOwner->GetCurrentServer()->GetPort(),
-		cert_list->data, cert_list->size,
-		activationTime, expirationTime,
-		serial,
-		algoName, bits,
-		fingerprint_md5,
-		fingerprint_sha1,
-		subject,
-		issuer);
+		GetCipherName(),
+		GetMacName(),
+		certificates);
 
-	pNotification->requestNumber = m_pOwner->GetEngine()->GetNextAsyncRequestNumber();
-	
-	m_pOwner->GetEngine()->AddNotification(pNotification);
+	m_pOwner->SendAsyncRequest(pNotification);
 
 	m_pOwner->LogMessage(Status, _("Verifying certificate..."));
 
@@ -927,3 +991,21 @@ int CTlsSocket::VerifyCertificate()
 void CTlsSocket::OnRateAvailable(enum CRateLimiter::rate_direction direction)
 {
 }
+
+wxString CTlsSocket::GetCipherName()
+{
+	const char* cipher = gnutls_cipher_get_name(gnutls_cipher_get(m_session));
+	if (cipher && *cipher)
+		return wxString(cipher, wxConvUTF8);
+	else
+		return _T("unknown");
+}
+
+wxString CTlsSocket::GetMacName()
+{
+	const char* mac = gnutls_mac_get_name(gnutls_mac_get(m_session));
+	if (mac && *mac)
+		return wxString(mac, wxConvUTF8);
+	else
+		return _T("unknown");
+}
