@@ -4,9 +4,10 @@
 #define ssize_t long
 #include <gnutls/gnutls.h>
 #include "backend.h"
+#include "socket.h"
 
 class CControlSocket;
-class CTlsSocket : protected wxEvtHandler, public CBackend
+class CTlsSocket : protected CSocketEventHandler, public CBackend, public CSocketEventSource
 {
 public:
 	enum TlsState
@@ -19,29 +20,33 @@ public:
 		closed
 	};
 
-	CTlsSocket(wxEvtHandler* pEvtHandler, wxSocketBase* pSocket, CControlSocket* pOwner);
-	~CTlsSocket();
+	CTlsSocket(CSocketEventHandler* pEvtHandler, CSocket* pSocket, CControlSocket* pOwner);
+	virtual ~CTlsSocket();
 
 	bool Init();
 	void Uninit();
 
-	int Handshake(const CTlsSocket* pPrimarySocket = 0);
+	int Handshake(const CTlsSocket* pPrimarySocket = 0, bool try_resume = false);
 
-	virtual void Read(void *buffer, unsigned int len);
-	virtual void Write(const void *buffer, unsigned int len);
-	virtual bool Error() const { return !m_lastSuccessful; }
-	virtual unsigned int LastCount() const { return m_lastCount; }
-	virtual int LastError() const { return m_lastError; }
-	virtual void Peek(void *buffer, unsigned int len);
+	virtual int Read(void *buffer, unsigned int size, int& error);
+	virtual int Peek(void *buffer, unsigned int size, int& error);
+	virtual int Write(const void *buffer, unsigned int size, int& error);
 
-	void Shutdown();
+	int Shutdown();
 
 	void TrustCurrentCert(bool trusted);
 
 	enum TlsState GetState() const { return m_tlsState; }
+
+	wxString GetCipherName();
+	wxString GetMacName();
+
+	bool ResumedSession() const;
 	
 protected:
 
+	bool CopySessionData(const CTlsSocket* pPrimarySocket);
+
 	virtual void OnRateAvailable(enum CRateLimiter::rate_direction direction);
 
 	void ContinueShutdown();
@@ -61,7 +66,7 @@ protected:
 	void PrintAlert();
 	
 	// Failure logs the error, uninits the session and sends a close event
-	void Failure(int code);
+	void Failure(int code, int socket_error);
 
 	static ssize_t PushFunction(gnutls_transport_ptr_t ptr, const void* data, size_t len);
 	static ssize_t PullFunction(gnutls_transport_ptr_t ptr, void* data, size_t len);
@@ -70,8 +75,7 @@ protected:
 
 	void TriggerEvents();
 
-	DECLARE_EVENT_TABLE();
-	void OnSocketEvent(wxSocketEvent& event);
+	void OnSocketEvent(CSocketEvent& event);
 	void OnRead();
 	void OnSend();
 
@@ -85,12 +89,8 @@ protected:
 	bool m_socketClosed;
 
 	CSocketBackend* m_pSocketBackend;
-	wxSocketBase* m_pSocket;
+	CSocket* m_pSocket;
 
-	// Used by LastError() and LastCount()
-	int m_lastError;
-	unsigned int m_lastCount;
-	bool m_lastSuccessful;
 	bool m_shutdown_requested;
 
 	// Due to the strange gnutls_record_send semantics, call it again
@@ -99,7 +99,8 @@ protected:
 	// application.
 	// This avoids the rule to call it again with the -same- data after
 	// GNUTLS_E_AGAIN.
-	void CheckResumeFailedWrite();
+	void CheckResumeFailedReadWrite();
+	bool m_lastReadFailed;
 	bool m_lastWriteFailed;
 	unsigned int m_writeSkip;
 
@@ -108,6 +109,8 @@ protected:
 	unsigned int m_peekDataLen;
 
 	gnutls_datum_t m_implicitTrustedCert;
+
+	bool m_socket_eof;
 };
 
 #endif //__TLSSOCKET_H__
