@@ -19,7 +19,12 @@ enum ServerType
 	DOS,
 	MVS,
 	VXWORKS,
-	ZVM
+	ZVM,
+	HPNONSTOP,
+	DOS_VIRTUAL,
+	CYGWIN,
+
+	SERVERTYPE_MAX
 };
 
 enum LogonType
@@ -28,7 +33,9 @@ enum LogonType
 	NORMAL,
 	ASK, // ASK should not be sent to the engine, it's intendet to be used by the interface
 	INTERACTIVE,
-	ACCOUNT
+	ACCOUNT,
+
+	LOGONTYPE_MAX
 };
 
 enum PasvMode
@@ -70,13 +77,14 @@ public:
 	int GetTimezoneOffset() const;
 	enum PasvMode GetPasvMode() const;
 	int MaximumMultipleConnections() const;
+	bool GetBypassProxy() const;
 	
 	// Return true if URL could be parsed correctly, false otherwise.
 	// If parsing fails, pError is filled with the reason and the CServer instance may be left an undefined state.
 	bool ParseUrl(wxString host, unsigned int, wxString user, wxString pass, wxString &error, CServerPath &path);
 	
 	void SetProtocol(enum ServerProtocol serverProtocol);
-	bool SetHost(wxString Host, unsigned int);
+	bool SetHost(wxString Host, unsigned int port);
 
 	void SetLogonType(enum LogonType logonType);
 	bool SetUser(const wxString& user, const wxString& pass = _T(""));
@@ -86,13 +94,14 @@ public:
 	bool operator==(const CServer &op) const;
 	bool operator<(const CServer &op) const;
 	bool operator!=(const CServer &op) const;
+	bool EqualsNoPass(const CServer &op) const;
 
 	bool SetTimezoneOffset(int minutes);	
 	void SetPasvMode(enum PasvMode pasvMode);
 	void MaximumMultipleConnections(int maximum);
 
-	wxString FormatHost() const;
-	wxString FormatServer() const;
+	wxString FormatHost(bool always_omit_port = false) const;
+	wxString FormatServer(const bool always_include_prefix = false) const;
 
 	bool SetEncodingType(enum CharsetEncoding type, const wxString& encoding = _T(""));
 	bool SetCustomEncoding(const wxString& encoding);
@@ -100,7 +109,7 @@ public:
 	wxString GetCustomEncoding() const;
 
 	static unsigned int GetDefaultPort(enum ServerProtocol protocol);
-	static enum ServerProtocol GetProtocolFromPort(unsigned int port);
+	static enum ServerProtocol GetProtocolFromPort(unsigned int port, bool defaultOnly = false);
 
 	static wxString GetProtocolName(enum ServerProtocol protocol);
 	static enum ServerProtocol GetProtocolFromName(const wxString& name);
@@ -108,9 +117,25 @@ public:
 	static enum ServerProtocol GetProtocolFromPrefix(const wxString& prefix);
 	static wxString GetPrefixFromProtocol(const enum ServerProtocol protocol);
 
+	// Some protocol distinguish between ASCII and binary files for line-ending
+	// conversion.
+	static bool ProtocolHasDataTypeConcept(const enum ServerProtocol protocol);
+
 	// These commands will be executed after a successful login.
 	const std::vector<wxString>& GetPostLoginCommands() const { return m_postLoginCommands; }
 	bool SetPostLoginCommands(const std::vector<wxString>& postLoginCommands);
+	void SetBypassProxy(bool val);
+
+	// Abstract server name.
+	// Not compared in ==, < and related operators
+	void SetName(const wxString& name) { m_name = name; }
+	wxString GetName() const { return m_name; }
+
+	static wxString GetNameFromServerType(enum ServerType type);
+	static enum ServerType GetServerTypeFromName(const wxString& name);
+
+	static wxString GetNameFromLogonType(enum LogonType type);
+	static enum LogonType GetLogonTypeFromName(const wxString& name);
 
 protected:
 	void Initialize();
@@ -128,8 +153,10 @@ protected:
 	int m_maximumMultipleConnections;
 	enum CharsetEncoding m_encodingType;
 	wxString m_customEncoding;
+	wxString m_name;
 
 	std::vector<wxString> m_postLoginCommands;
+	bool m_bypassProxy;
 };
 
 #endif
