@@ -1,23 +1,40 @@
 #ifndef __LOCALLISTVIEW_H__
 #define __LOCALLISTVIEW_H__
 
-#include "systemimagelist.h"
+#include "filelistctrl.h"
 #include "state.h"
 
 class CQueueView;
 class CLocalListViewDropTarget;
-class CLocalListViewSortObject;
+#ifdef __WXMSW__
+class CVolumeDescriptionEnumeratorThread;
+#endif
 
-class CLocalListView : public wxListCtrl, CSystemImageList, CStateEventHandler
+class CLocalFileData : public CGenericFileData
+{
+public:
+	wxString name;
+#ifdef __WXMSW__
+	wxString label;
+#endif
+	bool dir;
+	wxLongLong size;
+	bool hasTime;
+	wxDateTime lastModified;
+	int attributes;
+};
+
+class CLocalListView : public CFileListCtrl<CLocalFileData>, CSystemImageList, CStateEventHandler
 {
 	friend class CLocalListViewDropTarget;
+	friend class CLocalListViewSortType;
 
 public:
-	CLocalListView(wxWindow* parent, wxWindowID id, CState *pState, CQueueView *pQueue);
+	CLocalListView(wxWindow* parent, CState *pState, CQueueView *pQueue);
 	virtual ~CLocalListView();
 
 protected:
-	void OnStateChange(unsigned int event, const wxString& data);
+	void OnStateChange(enum t_statechange_notifications notification, const wxString& data);
 	bool DisplayDir(wxString dirname);
 	void ApplyCurrentFilter();
 
@@ -25,8 +42,8 @@ protected:
 	// Won't be fixed since a fix would break backwards compatibility
 	// Both functions use a const_cast<CLocalListView *>(this) and modify
 	// the instance.
-	virtual wxString OnGetItemText(long item, long column) const;
 	virtual int OnGetItemImage(long item) const;
+	virtual wxListItemAttr* OnGetItemAttr(long item) const;
 
 	// Clears all selections and returns the list of items that were selected
 	std::list<wxString> RememberSelectedItems(wxString& focused);
@@ -42,72 +59,52 @@ protected:
 #endif
 
 public:
-	wxString GetType(wxString name, bool dir);
-
-	struct t_fileData
-	{
-		wxString name;
-		bool dir;
-		int icon;
-		wxLongLong size;
-		wxString fileType;
-		bool hasTime;
-		wxDateTime lastModified;
-	};
-
 	void InitDateFormat();
 
+	virtual bool CanStartComparison(wxString* pError);
+	virtual void StartComparison();
+	virtual bool GetNextFile(wxString& name, bool &dir, wxLongLong &size, wxDateTime& date, bool &hasTime);
+	virtual void FinishComparison();
+
+	virtual bool ItemIsDir(int index) const;
+	virtual wxLongLong ItemGetSize(int index) const;
+
 protected:
+	virtual wxString GetItemText(int item, unsigned int column);
+
 	bool IsItemValid(unsigned int item) const;
-	t_fileData *GetData(unsigned int item);
+	CLocalFileData *GetData(unsigned int item);
 
-	void SortList(int column = -1, int direction = -1);
-	void SortList_UpdateSelections(bool* selections, int focus);
-	CLocalListViewSortObject GetComparisonObject();
-
-	int FindItemWithPrefix(const wxString& prefix, int start);
+	virtual CSortComparisonObject GetSortComparisonObject();
 
 	void RefreshFile(const wxString& file);
 
 	wxString m_dir;
 
-	std::vector<t_fileData> m_fileData;
-	std::vector<unsigned int> m_indexMapping;
-	std::map<wxString, wxString> m_fileTypeMap;
-
-#ifdef __WXMSW__
-	wxImageListEx *m_pHeaderImageList;
-#endif
-
-	CQueueView *m_pQueue;
-
-	int m_sortColumn;
-	int m_sortDirection;
-
-	wxDateTime m_lastKeyPress;
-	wxString m_prefix;
-
 	wxDropTarget* m_pDropTarget;
 	int m_dropTarget;
 
-	bool m_hasParent;
-
 	wxString m_dateFormat;
 
 	// Event handlers
 	DECLARE_EVENT_TABLE();
 	void OnItemActivated(wxListEvent& event);
-	void OnColumnClicked(wxListEvent& event);
 	void OnContextMenu(wxContextMenuEvent& event);
 	void OnMenuUpload(wxCommandEvent& event);
 	void OnMenuMkdir(wxCommandEvent& event);
 	void OnMenuDelete(wxCommandEvent& event);
 	void OnMenuRename(wxCommandEvent& event);
-	void OnChar(wxKeyEvent& event);
 	void OnKeyDown(wxKeyEvent& event);
 	void OnBeginLabelEdit(wxListEvent& event);
 	void OnEndLabelEdit(wxListEvent& event);
 	void OnBeginDrag(wxListEvent& event);
+	void OnMenuOpen(wxCommandEvent& event);
+	void OnMenuEdit(wxCommandEvent& event);
+	void OnMenuEnter(wxCommandEvent& event);
+#ifdef __WXMSW__
+	void OnVolumesEnumerated(wxCommandEvent& event);
+	CVolumeDescriptionEnumeratorThread* m_pVolumeEnumeratorThread;
+#endif
 };
 
 #endif
