@@ -1,6 +1,10 @@
 #ifndef __MAINFRM_H__
 #define __MAINFRM_H__
 
+#ifndef __WXMAC__
+#include <wx/taskbar.h>
+#endif
+
 class CStatusView;
 class CQueueView;
 class CLocalTreeView;
@@ -16,9 +20,14 @@ class CQuickconnectBar;
 #if FZ_MANUALUPDATECHECK && FZ_AUTOUPDATECHECK
 class CUpdateWizard;
 #endif //FZ_MANUALUPDATECHECK && FZ_AUTOUPDATECHECK
-class CSiteManagerItemData;
+class CSiteManagerItemData_Site;
 class CQueue;
 class CViewHeader;
+class CComparisonManager;
+class CWindowStateManager;
+class CStatusBar;
+class CMainFrameStateEventHandler;
+class CSplitterWindowEx;
 
 class CMainFrame : public wxFrame
 {
@@ -26,8 +35,7 @@ public:
 	CMainFrame();
 	virtual ~CMainFrame();
 
-	void UpdateSendLed();
-	void UpdateRecvLed();
+	void UpdateActivityLed(int direction);
 
 	void AddToRequestQueue(CFileZillaEngine* pEngine, CAsyncRequestNotification* pNotification);
 	CState* GetState() { return m_pState; }
@@ -37,37 +45,59 @@ public:
 	CQueueView* GetQueue() { return m_pQueueView; }
 	CQuickconnectBar* GetQuickconnectBar() { return m_pQuickconnectBar; }
 
-	void UpdateLayout(int layout = -1, int swap = -1);
+	void UpdateLayout(int layout = -1, int swap = -1, int messagelog_position = -1);
 
 	// Window size and position as well as pane sizes
-	void RememberSizes();
-	void RestoreSizes();
+	void RememberSplitterPositions();
+	bool RestoreSplitterPositions();
+	void SetDefaultSplitterPositions();
 
 	void CheckChangedSettings();
 
 	void ConnectNavigationHandler(wxEvtHandler* handler);
 
+	CStatusBar* GetStatusBar() { return m_pStatusBar; }
+
+	void UpdateToolbarState();
+	void UpdateMenubarState();
+
+	void ProcessCommandLine();
+
+	void ClearBookmarks();
+
+	CComparisonManager* GetComparisonManager() { return m_pComparisonManager; }
+
 protected:
 	bool CreateMenus();
 	bool CreateQuickconnectBar();
 	bool CreateToolBar();
 	void SetProgress(const CTransferStatus* pStatus);
-	void ConnectToSite(CSiteManagerItemData* const pData);
-
-	// If resizing the window, make sure the individual splitter windows don't get too small
-	void ApplySplitterConstraints();
+	bool ConnectToSite(CSiteManagerItemData_Site* const pData);
+	void OpenSiteManager(const CServer* pServer = 0);
+	void InitToolbarState();
+	void InitMenubarState();
 
 	void FocusNextEnabled(std::list<wxWindow*>& windowOrder, std::list<wxWindow*>::iterator iter, bool skipFirst, bool forward);
 
-	wxStatusBar* m_pStatusBar;
+	void UpdateBookmarkMenu();
+
+	std::list<int> m_bookmark_menu_ids;
+	std::map<int, wxString> m_bookmark_menu_id_map_global;
+	std::map<int, wxString> m_bookmark_menu_id_map_site;
+	wxString m_last_bookmark_path;
+	std::list<wxString> m_bookmarks;
+
+	CStatusBar* m_pStatusBar;
 	wxMenuBar* m_pMenuBar;
 	wxToolBar* m_pToolBar;
 	CQuickconnectBar* m_pQuickconnectBar;
-	wxSplitterWindow* m_pTopSplitter;
-	wxSplitterWindow* m_pBottomSplitter;
-	wxSplitterWindow* m_pViewSplitter;
-	wxSplitterWindow* m_pLocalSplitter;
-	wxSplitterWindow* m_pRemoteSplitter;
+
+	CSplitterWindowEx* m_pTopSplitter; // If log position is 0, splits message log from rest of panes
+	CSplitterWindowEx* m_pBottomSplitter; // Top contains view splitter, buttom queue (or queuelog splitter if in position 1)
+	CSplitterWindowEx* m_pViewSplitter; // Contains local and remote splitters
+	CSplitterWindowEx* m_pLocalSplitter;
+	CSplitterWindowEx* m_pRemoteSplitter;
+	CSplitterWindowEx* m_pQueueLogSplitter;
 
 	CStatusView* m_pStatusView;
 	CQueueView* m_pQueueView;
@@ -81,87 +111,85 @@ protected:
 	CRemoteListView* m_pRemoteListView;
 	CViewHeader* m_pLocalViewHeader;
 	CViewHeader* m_pRemoteViewHeader;
-	CLed* m_pRecvLed;
-	CLed* m_pSendLed;
+	CLed* m_pActivityLed[2];
 	wxTimer m_transferStatusTimer;
 	CThemeProvider* m_pThemeProvider;
 #if FZ_MANUALUPDATECHECK && FZ_AUTOUPDATECHECK
 	CUpdateWizard* m_pUpdateWizard;
 #endif //FZ_MANUALUPDATECHECK && FZ_AUTOUPDATECHECK
+	CComparisonManager* m_pComparisonManager;
+
+	void ShowLocalTree();
+	void ShowRemoteTree();
+
+#if defined(EVT_TOOL_DROPDOWN) && defined(__WXMSW__)
+	void MakeDropdownTool(wxToolBar* pToolBar, int id);
+#endif
+	void ShowDropdownMenu(wxMenu* pMenu, wxToolBar* pToolBar, wxCommandEvent& event);
+
+#ifdef __WXMSW__
+	virtual WXLRESULT MSWWindowProc(WXUINT nMsg, WXWPARAM wParam, WXLPARAM lParam);
+#endif
+
+	void HandleResize();
 
 	// Event handlers
 	DECLARE_EVENT_TABLE()
 	void OnSize(wxSizeEvent& event);
-	void OnViewSplitterPosChanged(wxSplitterEvent& event);
 	void OnMenuHandler(wxCommandEvent& event);
 	void OnMenuOpenHandler(wxMenuEvent& event);
 	void OnEngineEvent(wxEvent& event);
-	void OnUpdateToolbarDisconnect(wxUpdateUIEvent& event);
 	void OnDisconnect(wxCommandEvent& event);
-	void OnUpdateToolbarCancel(wxUpdateUIEvent& event);
 	void OnCancel(wxCommandEvent& event);
-	void OnSplitterSashPosChanging(wxSplitterEvent& event);
-	void OnSplitterSashPosChanged(wxSplitterEvent& event);
 	void OnClose(wxCloseEvent& event);
-	void OnUpdateToolbarReconnect(wxUpdateUIEvent& event);
 	void OnReconnect(wxCommandEvent& event);
 	void OnRefresh(wxCommandEvent& event);
-	void OnStatusbarSize(wxSizeEvent& event);
 	void OnTimer(wxTimerEvent& event);
 	void OnSiteManager(wxCommandEvent& event);
 	void OnProcessQueue(wxCommandEvent& event);
-	void OnUpdateToolbarProcessQueue(wxUpdateUIEvent& event);
 	void OnMenuEditSettings(wxCommandEvent& event);
 	void OnToggleLogView(wxCommandEvent& event);
-	void OnUpdateToggleLogView(wxUpdateUIEvent& event);
 	void OnToggleLocalTreeView(wxCommandEvent& event);
-	void OnUpdateToggleLocalTreeView(wxUpdateUIEvent& event);
 	void OnToggleRemoteTreeView(wxCommandEvent& event);
-	void OnUpdateToggleRemoteTreeView(wxUpdateUIEvent& event);
 	void OnToggleQueueView(wxCommandEvent& event);
-	void OnUpdateToggleQueueView(wxUpdateUIEvent& event);
 	void OnMenuHelpAbout(wxCommandEvent& event);
 	void OnFilter(wxCommandEvent& event);
-	void OnUpdateMenuCustomcommand(wxUpdateUIEvent& event);
+	void OnFilterRightclicked(wxCommandEvent& event);
 #if FZ_MANUALUPDATECHECK
 	void OnCheckForUpdates(wxCommandEvent& event);
 #endif //FZ_MANUALUPDATECHECK
 	void OnSitemanagerDropdown(wxCommandEvent& event);
-	void OnUpdateMenuShowHidden(wxUpdateUIEvent& event);
 	void OnNavigationKeyEvent(wxNavigationKeyEvent& event);
 	void OnGetFocus(wxFocusEvent& event);
 	void OnChar(wxKeyEvent& event);
-	void OnMoveEvent(wxMoveEvent& event);
 	void OnActivate(wxActivateEvent& event);
-#ifdef __WXMSW__
-	void OnSizePost(wxCommandEvent& event);
-
-	bool m_pendingPostSizing;
+	void OnToolbarComparison(wxCommandEvent& event);
+	void OnToolbarComparisonDropdown(wxCommandEvent& event);
+	void OnDropdownComparisonMode(wxCommandEvent& event);
+	void OnDropdownComparisonHide(wxCommandEvent& event);
+	void OnSyncBrowse(wxCommandEvent& event);
+#ifndef __WXMAC__
+	void OnIconize(wxIconizeEvent& event);
+	void OnTaskBarClick(wxTaskBarIconEvent& event);
 #endif
+	void OnSearch(wxCommandEvent& event);
 
-	float m_ViewSplitterSashPos;
 	bool m_bInitDone;
 	bool m_bQuit;
+	wxEventType m_closeEvent;
+	wxTimer m_closeEventTimer;
 
 	CAsyncRequestQueue* m_pAsyncRequestQueue;
 	CState* m_pState;
+	CMainFrameStateEventHandler* m_pStateEventHandler;
 
-	// Variables to remember the splitter position on unsplit
-	int m_lastLogViewSplitterPos;
-	int m_lastLocalTreeSplitterPos;
-	int m_lastRemoteTreeSplitterPos;
-	int m_lastQueueSplitterPos;
-
-#ifdef __WXMSW__
-	bool m_windowIsMaximized;
-#endif
+	CWindowStateManager* m_pWindowStateManager;
 
 	CQueue* m_pQueuePane;
 
-	bool m_lastMaximized;
-	wxPoint m_lastWindowPosition;
-	wxSize m_lastWindowSize;
+#ifndef __WXMAC__
+	wxTaskBarIcon* m_taskBarIcon;
+#endif
 };
 
 #endif
-
