@@ -1,11 +1,9 @@
 #include "FileZilla.h"
 #include "Options.h"
-#include "../tinyxml/tinyxml.h"
 #include "xmlfunctions.h"
 #include "filezillaapp.h"
 #include <wx/tokenzr.h>
 #include "ipcmutex.h"
-
 #ifdef _DEBUG
 #define new DEBUG_NEW
 #endif
@@ -50,7 +48,7 @@ static const t_Option options[OPTIONS_NUM] =
 	{ "Speedlimit burst tolerance", number, _T("0") },
 	{ "View hidden files", number, _T("0") },
 	{ "Preserve timestamps", number, _T("0") },
-	{ "Socket recv buffer size", number, _T("131072") }, // Make it large enough by default 
+	{ "Socket recv buffer size", number, _T("131072") }, // Make it large enough by default
 														 // to enable a large TCP window scale
 	{ "Socket send buffer size", number, _T("131072") },
 	{ "FTP Keep-alive commands", number, _T("0") },
@@ -59,23 +57,31 @@ static const t_Option options[OPTIONS_NUM] =
 	{ "FTP Proxy user", string, _T("") },
 	{ "FTP Proxy password", string, _T("") },
 	{ "FTP Proxy login sequence", string, _T("") },
+	{ "SFTP keyfiles", string, _T("") },
+	{ "Proxy type", number, _T("0") },
+	{ "Proxy host", string, _T("") },
+	{ "Proxy port", number, _T("0") },
+	{ "Proxy user", string, _T("") },
+	{ "Proxy password", string, _T("") },
+	{ "Logging file", string, _T("") },
+	{ "Logging filesize limit", number, _T("10") },
 
 	// Interface settings
 	{ "Number of Transfers", number, _T("2") },
 	{ "Ascii Binary mode", number, _T("0") },
-	{ "Auto Ascii files", string, _T("am|asp|bat|c|cfm|cgi|conf|cpp|css|dhtml|diz|h|hpp|htm|html|in|inc|js|m4|mak|nfo|nsi|pas|patch|php|phtml|pl|po|py|qmail|sh|shtml|sql|tcl|tpl|txt|vbs|xml|xrc") },
+	{ "Auto Ascii files", string, _T("am|asp|bat|c|cfm|cgi|conf|cpp|css|dhtml|diz|h|hpp|htm|html|in|inc|js|jsp|m4|mak|md5|nfo|nsi|pas|patch|php|phtml|pl|po|py|qmail|sh|shtml|sql|svg|tcl|tpl|txt|vbs|xhtml|xml|xrc") },
 	{ "Auto Ascii no extension", number, _T("1") },
 	{ "Auto Ascii dotfiles", number, _T("1") },
-	{ "Theme", string, _T("") },
-	{ "Language", string, _T("") },
+	{ "Theme", string, _T("opencrystal/") },
+	{ "Language Code", string, _T("") },
 	{ "Last Server Path", string, _T("") },
-	{ "Max Concurrent Uploads", number, _T("0") },
-	{ "Max Concurrent Downloads", number, _T("0") },
+	{ "Concurrent download limit", number, _T("0") },
+	{ "Concurrent upload limit", number, _T("0") },
 	{ "Update Check", number, _T("1") },
 	{ "Update Check Interval", number, _T("7") },
 	{ "Last automatic update check", string, _T("") },
 	{ "Update Check New Version", string, _T("") },
-	{ "Update Check Package URL", string, _T("") },
+	{ "Update Check Check Beta", number, _T("0") },
 	{ "Show debug menu", number, _T("0") },
 	{ "File exists action download", number, _T("0") },
 	{ "File exists action upload", number, _T("0") },
@@ -93,6 +99,7 @@ static const t_Option options[OPTIONS_NUM] =
 	{ "Local filelist colwidths", string, _T("") },
 	{ "Remote filelist colwidths", string, _T("") },
 	{ "Window position and size", string, _T("") },
+	{ "Splitter positions (v2)", string, _T("") },
 	{ "Local filelist sortorder", string, _T("") },
 	{ "Remote filelist sortorder", string, _T("") },
 	{ "Time Format", string, _T("") },
@@ -105,8 +112,53 @@ static const t_Option options[OPTIONS_NUM] =
 	{ "Always use default editor", number, _T("0") },
 	{ "Inherit system associations", number, _T("1") },
 	{ "Custom file associations", string, _T("") },
+	{ "Comparison mode", number, _T("1") },
+	{ "Comparison threshold", number, _T("1") },
+	{ "Site Manager position", string, _T("") },
+	{ "Theme icon size", string, _T("") },
+	{ "Timestamp in message log", number, _T("0") },
+	{ "Sitemanager last selected", string, _T("") },
+	{ "Local filelist shown columns", string, _T("") },
+	{ "Remote filelist shown columns", string, _T("") },
+	{ "Local filelist column order", string, _T("") },
+	{ "Remote filelist column order", string, _T("") },
+	{ "Filelist status bar", number, _T("1") },
+	{ "Filter toggle state", number, _T("0") },
+	{ "Size decimal places", number, _T("0") },
+	{ "Show quickconnect bar", number, _T("1") },
+	{ "Messagelog position", number, _T("0") },
+	{ "Last connected site", string, _T("") },
+	{ "File doubleclock action", number, _T("0") },
+	{ "Dir doubleclock action", number, _T("0") },
+	{ "Minimize to tray", number, _T("0") },
+	{ "Search column widths", string, _T("") },
+	{ "Search column shown", string, _T("") },
+	{ "Search column order", string, _T("") },
+	{ "Search window size", string, _T("") },
+	{ "Comparison hide identical", number, _T("0") },
+	{ "Search sort order", string, _T("") },
+	{ "Edit track local", number, _T("1") }
 };
 
+struct t_default_option
+{
+	const wxChar name[30];
+	enum Type type;
+	wxString value_str;
+	int value_number;
+};
+
+static t_default_option default_options[DEFAULTS_NUM] =
+{
+	{ _T("Config Location"), string, _T(""), 0 },
+	{ _T("Kiosk mode"), number, _T(""), 0 },
+	{ _T("Disable update check"), number, _T(""), 0 }
+};
+
+BEGIN_EVENT_TABLE(COptions, wxEvtHandler)
+EVT_TIMER(wxID_ANY, COptions::OnTimer)
+END_EVENT_TABLE()
+
 COptions::COptions()
 {
 	m_pLastServer = 0;
@@ -116,17 +168,21 @@ COptions::COptions()
 	for (unsigned int i = 0; i < OPTIONS_NUM; i++)
 		m_optionsCache[i].cached = false;
 
+	m_save_timer.SetOwner(this);
+
 	CInterProcessMutex mutex(MUTEX_OPTIONS);
 	m_pXmlFile = new CXmlFile(_T("filezilla"));
 	if (!m_pXmlFile->Load())
 	{
-		wxString msg = wxString::Format(_("Could not load \"%s\", make sure the file is valid.\nFor this session, default settings will be used and any changes to the settings are not persistent."), m_pXmlFile->GetFileName().GetFullPath().c_str());
+		wxString msg = m_pXmlFile->GetError() + _T("\n\n") + _("For this session the default settings will be used. Any changes to the settings will not be saved.");
 		wxMessageBox(msg, _("Error loading xml file"), wxICON_ERROR);
 		delete m_pXmlFile;
 		m_pXmlFile = 0;
 	}
 	else
 		CreateSettingsXmlElement();
+
+	LoadGlobalDefaultOptions();
 }
 
 COptions::~COptions()
@@ -186,20 +242,6 @@ wxString COptions::GetOption(unsigned int nID)
 }
 
 bool COptions::SetOption(unsigned int nID, int value)
-{
-	if (!SetOptionNoSave(nID, value))
-		return false;
-
-	if (m_pXmlFile)
-	{
-		CInterProcessMutex mutex(MUTEX_OPTIONS);
-		m_pXmlFile->Save();
-	}
-
-	return true;
-}
-
-bool COptions::SetOptionNoSave(unsigned int nID, int value)
 {
 	if (nID >= OPTIONS_NUM)
 		return false;
@@ -213,26 +255,17 @@ bool COptions::SetOptionNoSave(unsigned int nID, int value)
 	m_optionsCache[nID].numValue = value;
 
 	if (m_pXmlFile)
+	{
 		SetXmlValue(nID, wxString::Format(_T("%d"), value));
 
-	return true;
-}
-
-bool COptions::SetOption(unsigned int nID, wxString value)
-{
-	if (!SetOptionNoSave(nID, value))
-		return false;
-	
-	if (m_pXmlFile)
-	{
-		CInterProcessMutex mutex(MUTEX_OPTIONS);
-		m_pXmlFile->Save();
+		if (!m_save_timer.IsRunning())
+			m_save_timer.Start(15000, true);
 	}
 
 	return true;
 }
 
-bool COptions::SetOptionNoSave(unsigned int nID, wxString value)
+bool COptions::SetOption(unsigned int nID, wxString value)
 {
 	if (nID >= OPTIONS_NUM)
 		return false;
@@ -252,8 +285,13 @@ bool COptions::SetOptionNoSave(unsigned int nID, wxString value)
 	m_optionsCache[nID].strValue = value;
 
 	if (m_pXmlFile)
+	{
 		SetXmlValue(nID, value);
 
+		if (!m_save_timer.IsRunning())
+			m_save_timer.Start(15000, true);
+	}
+
 
 	return true;
 }
@@ -338,6 +376,7 @@ void COptions::SetXmlValue(unsigned int nID, wxString value)
 			return;
 		}
 	}
+	wxASSERT(options[nID].name[0]);
 	TiXmlElement setting("Setting");
 	setting.SetAttribute("name", options[nID].name);
 	setting.SetAttribute("type", (options[nID].type == string) ? "string" : "number");
@@ -380,7 +419,13 @@ bool COptions::GetXmlValue(unsigned int nID, wxString &value, TiXmlElement *sett
 			continue;
 
 		TiXmlNode *text = setting->FirstChild();
-		if (!text || !text->ToText())
+		if (!text)
+		{
+			value.clear();
+			return true;
+		}
+
+		if (!text->ToText())
 			return false;
 
 		value = ConvLocal(text->Value());
@@ -436,6 +481,27 @@ int COptions::Validate(unsigned int nID, int value)
 		if (value != -1 && (value < 4096 || value > 4096 * 1024))
 			value = 131072;
 		break;
+	case OPTION_COMPARISONMODE:
+		if (value < 0 || value > 0)
+			value = 1;
+		break;
+	case OPTION_COMPARISON_THRESHOLD:
+		if (value < 0 || value > 1440)
+			value = 1;
+		break;
+	case OPTION_SIZE_DECIMALPLACES:
+		if (value < 0 || value > 3)
+			value = 0;
+		break;
+	case OPTION_MESSAGELOG_POSITION:
+		if (value < 0 || value > 2)
+			value = 0;
+		break;
+	case OPTION_DOUBLECLICK_ACTION_FILE:
+	case OPTION_DOUBLECLICK_ACTION_DIRECTORY:
+		if (value < 0 || value > 3)
+			value = 0;
+		break;
 	}
 	return value;
 }
@@ -454,7 +520,7 @@ void COptions::SetServer(wxString path, const CServer& server)
 		return;
 
 	TiXmlElement *element = m_pXmlFile->GetElement();
-	
+
 	while (path != _T(""))
 	{
 		wxString sub;
@@ -491,6 +557,9 @@ void COptions::SetServer(wxString path, const CServer& server)
 
 	::SetServer(element, server);
 
+	if (GetDefaultVal(DEFAULT_KIOSKMODE) == 2)
+		return;
+	
 	CInterProcessMutex mutex(MUTEX_OPTIONS);
 	m_pXmlFile->Save();
 }
@@ -588,49 +657,90 @@ void COptions::Import(TiXmlElement* pElement)
 		if (!GetXmlValue(i, value, pElement))
 			continue;
 
-		SetOptionNoSave(i, value);
-	}
-
-	if (m_pXmlFile)
-	{
-		CInterProcessMutex mutex(MUTEX_OPTIONS);
-		m_pXmlFile->Save();
+		SetOption(i, value);
 	}
 }
 
-void COptions::SaveColumnWidths(const wxListCtrl* const pListCtrl, unsigned int optionId)
+void COptions::LoadGlobalDefaultOptions()
 {
-	wxCHECK_RET(pListCtrl, _T("SaveColumnWidths called with !pListCtrl"));
+	const wxString& defaultsDir = wxGetApp().GetDefaultsDir();
+	if (defaultsDir == _T(""))
+		return;
+	
+	wxFileName name(defaultsDir, _T("fzdefaults.xml"));
+	CXmlFile file(name);
+	if (!file.Load())
+		return;
 
-	wxString widths;
-	for (int i = 0; i < pListCtrl->GetColumnCount(); i++)
-		widths += wxString::Format(_T("%d "), pListCtrl->GetColumnWidth(i));
-	widths.RemoveLast();
+	TiXmlElement* pElement = file.GetElement();
+	if (!pElement)
+		return;
 
-	SetOption(optionId, widths);
-}
+	pElement = pElement->FirstChildElement("Settings");
+	if (!pElement)
+		return;
 
-bool COptions::ReadColumnWidths(unsigned int optionId, unsigned int count, unsigned long* widths)
-{
-	wxString savedWidths = GetOption(optionId);
-	wxStringTokenizer tokens(savedWidths, _T(" "));
-	if (tokens.CountTokens() != count)
-		return false;
-
-	unsigned long* newWidths = new unsigned long[count];
-	for (unsigned int i = 0; i < count; i++)
+	for (TiXmlElement* pSetting = pElement->FirstChildElement("Setting"); pSetting; pSetting = pSetting->NextSiblingElement("Setting"))
 	{
-		wxString token = tokens.GetNextToken();
-		if (!token.ToULong(newWidths + i) || newWidths[i] > 5000)
+		wxString name = GetTextAttribute(pSetting, "name");
+		for (int i = 0; i < DEFAULTS_NUM; i++)
 		{
-			delete [] newWidths;
-			return false;
+			if (name != default_options[i].name)
+				continue;
+
+			wxString value = GetTextElement(pSetting);
+			if (default_options[i].type == string)
+				default_options[i].value_str = value;
+			else
+			{
+				long v = 0;
+				if (!value.ToLong(&v))
+					v = 0;
+				default_options[i].value_number = v;
+			}
+
 		}
 	}
+}
 
-	for (unsigned int i = 0; i < count; i++)
-		widths[i] = newWidths[i];
+int COptions::GetDefaultVal(unsigned int nID) const
+{
+	if (nID >= DEFAULTS_NUM)
+		return 0;
 
-	delete [] newWidths;
-	return true;
+	return default_options[nID].value_number;
+}
+
+wxString COptions::GetDefault(unsigned int nID) const
+{
+	if (nID >= DEFAULTS_NUM)
+		return _T("");
+
+	return default_options[nID].value_str;
+}
+
+void COptions::OnTimer(wxTimerEvent& event)
+{
+	Save();
+}
+
+void COptions::Save()
+{
+	if (GetDefaultVal(DEFAULT_KIOSKMODE) == 2)
+		return;
+
+	if (!m_pXmlFile)
+		return;
+
+	CInterProcessMutex mutex(MUTEX_OPTIONS);
+	m_pXmlFile->Save();
+}
+
+void COptions::SaveIfNeeded()
+{
+	if (!m_save_timer.IsRunning())
+		return;
+
+	m_save_timer.Stop();
+	Save();
 }
