@@ -17,7 +17,7 @@ enum interfaceOptions
 	OPTION_UPDATECHECK_INTERVAL,
 	OPTION_UPDATECHECK_LASTDATE,
 	OPTION_UPDATECHECK_NEWVERSION,
-	OPTION_UPDATECHECK_URL,
+	OPTION_UPDATECHECK_CHECKBETA,
 	OPTION_DEBUG_MENU,
 	OPTION_FILEEXISTS_DOWNLOAD,
 	OPTION_FILEEXISTS_UPLOAD,
@@ -34,7 +34,8 @@ enum interfaceOptions
 	OPTION_QUEUE_COLUMN_WIDTHS,
 	OPTION_LOCALFILELIST_COLUMN_WIDTHS,
 	OPTION_REMOTEFILELIST_COLUMN_WIDTHS,
-	OPTION_WINDOW_POSITION,
+	OPTION_MAINWINDOW_POSITION,
+	OPTION_MAINWINDOW_SPLITTER_POSITION,
 	OPTION_LOCALFILELIST_SORTORDER,
 	OPTION_REMOTEFILELIST_SORTORDER,
 	OPTION_TIME_FORMAT,
@@ -47,11 +48,47 @@ enum interfaceOptions
 	OPTION_EDIT_ALWAYSDEFAULT,
 	OPTION_EDIT_INHERITASSOCIATIONS,
 	OPTION_EDIT_CUSTOMASSOCIATIONS,
+	OPTION_COMPARISONMODE,
+	OPTION_COMPARISON_THRESHOLD,
+	OPTION_SITEMANAGER_POSITION,
+	OPTION_THEME_ICONSIZE,
+	OPTION_MESSAGELOG_TIMESTAMP,
+	OPTION_SITEMANAGER_LASTSELECTED,
+	OPTION_LOCALFILELIST_COLUMN_SHOWN,
+	OPTION_REMOTEFILELIST_COLUMN_SHOWN,
+	OPTION_LOCALFILELIST_COLUMN_ORDER,
+	OPTION_REMOTEFILELIST_COLUMN_ORDER,
+	OPTION_FILELIST_STATUSBAR,
+	OPTION_FILTERTOGGLESTATE,
+	OPTION_SIZE_DECIMALPLACES,
+	OPTION_SHOW_QUICKCONNECT,
+	OPTION_MESSAGELOG_POSITION,
+	OPTION_LAST_CONNECTED_SITE,
+	OPTION_DOUBLECLICK_ACTION_FILE,
+	OPTION_DOUBLECLICK_ACTION_DIRECTORY,
+	OPTION_MINIMIZE_TRAY,
+	OPTION_SEARCH_COLUMN_WIDTHS,
+	OPTION_SEARCH_COLUMN_SHOWN,
+	OPTION_SEARCH_COLUMN_ORDER,
+	OPTION_SEARCH_SIZE,
+	OPTION_COMPARE_HIDEIDENTICAL,
+	OPTION_SEARCH_SORTORDER,
+	OPTION_EDIT_TRACK_LOCAL,
 
 	// Has to be last element
 	OPTIONS_NUM
 };
 
+enum defaultOptions
+{
+	DEFAULT_SETTINGSDIR,
+	DEFAULT_KIOSKMODE,
+	DEFAULT_DISABLEUPDATECHECK,
+
+	// Has to be last element
+	DEFAULTS_NUM
+};
+
 struct t_OptionsCache
 {
 	bool cached;
@@ -61,18 +98,18 @@ struct t_OptionsCache
 
 class CXmlFile;
 class TiXmlElement;
-class COptions : public COptionsBase
+class COptions : public wxEvtHandler, public COptionsBase
 {
 public:
 	virtual int GetOptionVal(unsigned int nID);
 	virtual wxString GetOption(unsigned int nID);
 
+	int GetDefaultVal(unsigned int nID) const;
+	wxString GetDefault(unsigned int nID) const;
+
 	virtual bool SetOption(unsigned int nID, int value);
 	virtual bool SetOption(unsigned int nID, wxString value);
 	
-	bool SetOptionNoSave(unsigned int nID, int value);
-	bool SetOptionNoSave(unsigned int nID, wxString value);
-
 	void SetLastServer(const CServer& server);
 	bool GetLastServer(CServer& server);
 
@@ -82,9 +119,8 @@ public:
 
 	void Import(TiXmlElement* pElement);
 
-	void SaveColumnWidths(const wxListCtrl* const pListCtrl, unsigned int optionId);
-	bool ReadColumnWidths(unsigned int optionId, unsigned int count, unsigned long* widths);
-	
+	void SaveIfNeeded();
+
 protected:
 	COptions();
 	virtual ~COptions();
@@ -101,6 +137,10 @@ protected:
 
 	void CreateSettingsXmlElement();
 
+	void LoadGlobalDefaultOptions();
+
+	void Save();
+
 	CXmlFile* m_pXmlFile;
 
 	t_OptionsCache m_optionsCache[OPTIONS_NUM];
@@ -109,6 +149,11 @@ protected:
 	CServer* m_pLastServer;
 
 	static COptions* m_theOptions;
+
+	wxTimer m_save_timer;
+
+	DECLARE_EVENT_TABLE();
+	void OnTimer(wxTimerEvent& event);
 };
 
 #endif //__OPTIONS_H__
