@@ -1,28 +1,39 @@
 #ifndef __REMOTELISTVIEW_H__
 #define __REMOTELISTVIEW_H__
 
-#include "systemimagelist.h"
 #include "state.h"
+#include "listctrlex.h"
+#include "filelistctrl.h"
 
 class CQueueView;
 class CChmodDialog;
 class CInfoText;
 class CRemoteListViewDropTarget;
 
-class CRemoteListView : public wxListCtrl, CSystemImageList, CStateEventHandler
+class CRemoteListView : public CFileListCtrl<CGenericFileData>, CSystemImageList, CStateEventHandler
 {
 	friend class CRemoteListViewDropTarget;
+	friend class CRemoteListViewSortType;
+
 public:
-	CRemoteListView(wxWindow* parent, wxWindowID id, CState* pState, CQueueView* pQueue);
+	CRemoteListView(wxWindow* pParent, CState* pState, CQueueView* pQueue);
 	virtual ~CRemoteListView();
 
-	void ListingFailed();
-
-	bool DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataObject, wxString path, bool queueOnly);
+	bool DownloadDroppedFiles(const CRemoteDataObject* pRemoteDataObject, const CLocalPath& path, bool queueOnly);
 
 	void InitDateFormat();
 
+	virtual bool CanStartComparison(wxString* pError);
+	virtual void StartComparison();
+	virtual bool GetNextFile(wxString& name, bool &dir, wxLongLong &size, wxDateTime& date, bool& hasTime);
+	virtual void FinishComparison();
+	virtual void OnExitComparisonMode();
+
+	void LinkIsNotDir(const CServerPath& path, const wxString& link);
+
 protected:
+	virtual wxString GetItemText(int item, unsigned int column);
+
 	// Clears all selections and returns the list of items that were selected
 	std::list<wxString> RememberSelectedItems(wxString& focused);
 
@@ -36,63 +47,40 @@ protected:
 	// Won't be fixed since a fix would break backwards compatibility
 	// Both functions use a const_cast<CLocalListView *>(this) and modify
 	// the instance.
-	virtual wxString OnGetItemText(long item, long column) const;
+	virtual wxListItemAttr* OnGetItemAttr(long item) const;
 	virtual int OnGetItemImage(long item) const;
 
-	int FindItemWithPrefix(const wxString& prefix, int start);
-
-public:
-	struct t_fileData
-	{
-		int icon;
-		wxString fileType;
-	};
-
-	wxString GetType(wxString name, bool dir);
+	virtual bool ItemIsDir(int index) const;
+	virtual wxLongLong ItemGetSize(int index) const;
 
 protected:
 	bool IsItemValid(unsigned int item) const;
 	int GetItemIndex(unsigned int item) const;
 
-	void SortList(int column = -1, int direction = -1);
-	void SortList_UpdateSelections(bool* selections, int focus);
+	virtual CSortComparisonObject GetSortComparisonObject();
 
-	virtual void OnStateChange(unsigned int event, const wxString& data);
+	virtual void OnStateChange(enum t_statechange_notifications notification, const wxString& data);
 	void ApplyCurrentFilter();
-	void SetDirectoryListing(const CDirectoryListing *pDirectoryListing, bool modified = false);
-	bool UpdateDirectoryListing(const CDirectoryListing *pDirectoryListing);
-	void UpdateDirectoryListing_Removed(const CDirectoryListing *pDirectoryListing);
+	void SetDirectoryListing(const CSharedPointer<const CDirectoryListing> &pDirectoryListing, bool modified = false);
+	bool UpdateDirectoryListing(const CSharedPointer<const CDirectoryListing> &pDirectoryListing);
+	void UpdateDirectoryListing_Removed(const CSharedPointer<const CDirectoryListing> &pDirectoryListing);
+	void UpdateDirectoryListing_Added(const CSharedPointer<const CDirectoryListing> &pDirectoryListing);
 
 #ifdef __WXDEBUG__
 	void ValidateIndexMapping();
 #endif
 
-	const CDirectoryListing *m_pDirectoryListing;
-	std::vector<t_fileData> m_fileData;
-	std::vector<unsigned int> m_indexMapping;
-	std::map<wxString, wxString> m_fileTypeMap;
+	CSharedPointer <const CDirectoryListing> m_pDirectoryListing;
 
 	// Caller is responsible to check selection is valid!
-	void TransferSelectedFiles(const wxString& localDir, bool queueOnly);
+	void TransferSelectedFiles(const CLocalPath& local_parent, bool queueOnly);
 
 	// Cache icon for directories, no need to calculate it multiple times
 	int m_dirIcon;
 
-#ifdef __WXMSW__
-	wxImageListEx* m_pHeaderImageList;
-#endif
-
-	CQueueView* m_pQueue;
-
-	int m_sortColumn;
-	int m_sortDirection;
-
-	wxDateTime m_lastKeyPress;
-	wxString m_prefix;
-
 	CInfoText* m_pInfoText;
 	void RepositionInfoText();
-	void SetInfoText(const wxString& text);
+	void SetInfoText();
 
 	wxDropTarget* m_pDropTarget;
 	int m_dropTarget;
@@ -100,15 +88,24 @@ protected:
 	wxString m_dateFormat;
 	wxString m_timeFormat;
 
+	// Used to track state for resolving symlinks
+	// While being resolved, global state might have changed
+	// already.
+	struct t_linkResolveState
+	{
+		CServer server;
+		CServerPath remote_path;
+		wxString link;
+		wxString local_path;
+	} *m_pLinkResolveState;
+
 	DECLARE_EVENT_TABLE()
 	void OnItemActivated(wxListEvent &event);
-	void OnColumnClicked(wxListEvent &event);
 	void OnContextMenu(wxContextMenuEvent& event);
 	void OnMenuDownload(wxCommandEvent& event);
 	void OnMenuMkdir(wxCommandEvent& event);
 	void OnMenuDelete(wxCommandEvent& event);
 	void OnMenuRename(wxCommandEvent& event);
-	void OnChar(wxKeyEvent& event);
 	void OnKeyDown(wxKeyEvent& event);
 	void OnBeginLabelEdit(wxListEvent& event);
 	void OnEndLabelEdit(wxListEvent& event);
@@ -116,6 +113,8 @@ protected:
 	void OnSize(wxSizeEvent& event);
 	void OnBeginDrag(wxListEvent& event);
 	void OnMenuEdit(wxCommandEvent& event);
+	void OnMenuEnter(wxCommandEvent& event);
+	void OnMenuGeturl(wxCommandEvent& event);
 };
 
 #endif
