@@ -6,6 +6,8 @@
 #include "chmoddialog.h"
 #include "recursive_operation.h"
 #include "inputdialog.h"
+#include "dragdropmanager.h"
+#include <wx/clipbrd.h>
 
 #ifdef _DEBUG
 #define new DEBUG_NEW
@@ -70,6 +72,10 @@ public:
 		if (!GetData())
 			return wxDragError;
 
+		CDragDropManager* pDragDropManager = CDragDropManager::Get();
+		if (pDragDropManager)
+			pDragDropManager->pDropTarget = m_pRemoteTreeView;
+
 		if (m_pDataObject->GetReceivedFormat() == m_pFileDataObject->GetFormat())
 			m_pRemoteTreeView->m_pState->UploadDroppedFiles(m_pFileDataObject, path, false);
 		else
@@ -221,13 +227,23 @@ EVT_MENU(XRCID("ID_RENAME"), CRemoteTreeView::OnMenuRename)
 EVT_TREE_BEGIN_LABEL_EDIT(wxID_ANY, CRemoteTreeView::OnBeginLabelEdit)
 EVT_TREE_END_LABEL_EDIT(wxID_ANY, CRemoteTreeView::OnEndLabelEdit)
 EVT_MENU(XRCID("ID_MKDIR"), CRemoteTreeView::OnMkdir)
+EVT_CHAR(CRemoteTreeView::OnChar)
+EVT_MENU(XRCID("ID_GETURL"), CRemoteTreeView::OnMenuGeturl)
 END_EVENT_TABLE()
 
 CRemoteTreeView::CRemoteTreeView(wxWindow* parent, wxWindowID id, CState* pState, CQueueView* pQueue)
 	: wxTreeCtrl(parent, id, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL | wxTR_EDIT_LABELS | wxTR_LINES_AT_ROOT | wxTR_HAS_BUTTONS | wxNO_BORDER | wxTR_HIDE_ROOT),
 	CSystemImageList(16),
-	CStateEventHandler(pState, STATECHANGE_REMOTE_DIR | STATECHANGE_REMOTE_DIR_MODIFIED | STATECHANGE_APPLYFILTER)
+	CStateEventHandler(pState)
 {
+#ifdef __WXMAC__
+	SetFont(wxSystemSettings::GetFont(wxSYS_DEFAULT_GUI_FONT));
+#endif
+
+	pState->RegisterHandler(this, STATECHANGE_REMOTE_DIR);
+	pState->RegisterHandler(this, STATECHANGE_REMOTE_DIR_MODIFIED);
+	pState->RegisterHandler(this, STATECHANGE_APPLYFILTER);
+
 	m_busy = false;
 	m_pQueue = pQueue;
 	AddRoot(_T(""));
@@ -245,23 +261,23 @@ CRemoteTreeView::~CRemoteTreeView()
 	delete m_pImageList;
 }
 
-void CRemoteTreeView::OnStateChange(unsigned int event, const wxString& data)
+void CRemoteTreeView::OnStateChange(enum t_statechange_notifications notification, const wxString& data)
 {
-	if (event == STATECHANGE_REMOTE_DIR)
+	if (notification == STATECHANGE_REMOTE_DIR)
 		SetDirectoryListing(m_pState->GetRemoteDir(), false);
-	else if (event == STATECHANGE_REMOTE_DIR_MODIFIED)
+	else if (notification == STATECHANGE_REMOTE_DIR_MODIFIED)
 		SetDirectoryListing(m_pState->GetRemoteDir(), true);
+	else if (notification == STATECHANGE_APPLYFILTER)
+		ApplyFilters();
 }
 
-void CRemoteTreeView::SetDirectoryListing(const CDirectoryListing* pListing, bool modified)
+void CRemoteTreeView::SetDirectoryListing(const CSharedPointer<const CDirectoryListing> &pListing, bool modified)
 {
 	m_busy = true;
 
-	m_pDirectoryListing = pListing;
-
 	if (!pListing)
 	{
-        m_ExpandAfterList = wxTreeItemId();
+		m_ExpandAfterList = wxTreeItemId();
 		DeleteAllItems();
 		AddRoot(_T(""));
 		m_busy = false;
@@ -279,32 +295,63 @@ void CRemoteTreeView::SetDirectoryListing(const CDirectoryListing* pListing, boo
 	}
 	Enable(true);
 
+	if (pListing->m_hasUnsureEntries && !(pListing->m_hasUnsureEntries & ~(CDirectoryListing::unsure_unknown | CDirectoryListing::unsure_file_mask)))
+	{
+		// Just files changed, does not affect directory tree
+		m_busy = false;
+		return;
+	}
+
+#ifndef __WXMSW__
+	Freeze();
+#endif
 	wxTreeItemId parent = MakeParent(pListing->path, !modified);
 	if (!parent)
 	{
 		m_busy = false;
+#ifndef __WXMSW__
+		Thaw();
+#endif
 		return;
 	}
 
 	if (!IsExpanded(parent) && parent != m_ExpandAfterList)
 	{
 		DeleteChildren(parent);
-		CFilterDialog filter;
+		CFilterManager filter;
 		if (HasSubdirs(*pListing, filter))
 			AppendItem(parent, _T(""), -1, -1);
 	}
 	else
 	{
-		RefreshItem(parent, *pListing);
+		RefreshItem(parent, *pListing, !modified);
 
 		if (m_ExpandAfterList == parent)
+		{
+#ifndef __WXMSW__
+			// Prevent CalculatePositions from being called
+			wxGenericTreeItem *anchor = m_anchor;
+			m_anchor = 0;
+#endif
 			Expand(parent);
+#ifndef __WXMSW__
+			m_anchor = anchor;
+#endif
+		}
 	}
 	m_ExpandAfterList = wxTreeItemId();
 
 	SetItemImages(parent, false);
 
-	Refresh(false);
+#ifndef __WXMSW__
+	m_freezeCount--;
+#endif
+	if (!modified)
+		SelectItem(parent);
+#ifndef __WXMSW__
+	else
+		Refresh();
+#endif
 
 	m_busy = false;
 }
@@ -320,25 +367,28 @@ wxTreeItemId CRemoteTreeView::MakeParent(CServerPath path, bool select)
 	wxASSERT(path.GetPath() != _T(""));
 	pieces.push_front(path.GetPath());
 
-	wxTreeItemId parent = GetRootItem();
+	const wxTreeItemId root = GetRootItem();
+	wxTreeItemId parent = root;
+
 	for (std::list<wxString>::const_iterator iter = pieces.begin(); iter != pieces.end(); iter++)
 	{
 		if (iter != pieces.begin())
 			path.AddSegment(*iter);
 
 		wxTreeItemIdValue cookie;
-		wxTreeItemId child;
+		wxTreeItemId child = GetFirstChild(parent, cookie);
+		if (child && GetItemText(child) == _T(""))
+		{
+			Delete(child);
+			child = wxTreeItemId();
+			if (parent != root)
+				ListExpand(parent);
+		}
 		for (child = GetFirstChild(parent, cookie); child; child = GetNextSibling(child))
 		{
 			const wxString& text = GetItemText(child);
 			if (text == *iter)
 				break;
-			if (text == _T(""))
-			{
-				Delete(child);
-				child = wxTreeItemId();
-				break;
-			}
 		}
 		if (!child)
 		{
@@ -346,12 +396,12 @@ wxTreeItemId CRemoteTreeView::MakeParent(CServerPath path, bool select)
 
 			if (m_pState->m_pEngine->CacheLookup(path, listing) == FZ_REPLY_OK)
 			{
-				child = AppendItem(parent, *iter, 0, 2, new CItemData(path));
+				child = AppendItem(parent, *iter, 0, 2, path.HasParent() ? 0 : new CItemData(path));
 				SetItemImages(child, false);
 			}
 			else
 			{
-				child = AppendItem(parent, *iter, 1, 3, new CItemData(path));
+				child = AppendItem(parent, *iter, 1, 3, path.HasParent() ? 0 : new CItemData(path));
 				SetItemImages(child, true);
 			}
 			SortChildren(parent);
@@ -362,12 +412,20 @@ wxTreeItemId CRemoteTreeView::MakeParent(CServerPath path, bool select)
 				DisplayItem(child, listing);
 		}
 		if (select && iter != pieces.begin())
+		{
+#ifndef __WXMSW__
+			// Prevent CalculatePositions from being called
+			wxGenericTreeItem *anchor = m_anchor;
+			m_anchor = 0;
+#endif
 			Expand(parent);
+#ifndef __WXMSW__
+			m_anchor = anchor;
+#endif
+		}
 
 		parent = child;
 	}
-	if (select)
-		SelectItem(parent);
 
 	return parent;
 }
@@ -428,17 +486,21 @@ void CRemoteTreeView::CreateImageList()
 	SetImageList(m_pImageList);
 }
 
-bool CRemoteTreeView::HasSubdirs(const CDirectoryListing& listing, const CFilterDialog& filter)
+bool CRemoteTreeView::HasSubdirs(const CDirectoryListing& listing, const CFilterManager& filter)
 {
 	if (!listing.m_hasDirs)
 		return false;
 
+	if (!filter.HasActiveFilters())
+		return true;
+
+	const wxString path = listing.path.GetPath();
 	for (unsigned int i = 0; i < listing.GetCount(); i++)
 	{
 		if (!listing[i].dir)
 			continue;
 
-		if (filter.FilenameFiltered(listing[i].name, true, -1, false))
+		if (filter.FilenameFiltered(listing[i].name, path, true, -1, false, 0))
 			continue;
 
 		return true;
@@ -451,13 +513,15 @@ void CRemoteTreeView::DisplayItem(wxTreeItemId parent, const CDirectoryListing&
 {
 	DeleteChildren(parent);
 
+	const wxString path = listing.path.GetPath();
+
 	CFilterDialog filter;
 	for (unsigned int i = 0; i < listing.GetCount(); i++)
 	{
 		if (!listing[i].dir)
 			continue;
 
-		if (filter.FilenameFiltered(listing[i].name, true, -1, false))
+		if (filter.FilenameFiltered(listing[i].name, path, true, -1, false, 0))
 			continue;
 
 		const wxString& name = listing[i].name;
@@ -468,7 +532,7 @@ void CRemoteTreeView::DisplayItem(wxTreeItemId parent, const CDirectoryListing&
 
 		if (m_pState->m_pEngine->CacheLookup(subdir, subListing) == FZ_REPLY_OK)
 		{
-			wxTreeItemId child = AppendItem(parent, name, 0, 2, new CItemData(subdir));
+			wxTreeItemId child = AppendItem(parent, name, 0, 2, 0);
 			SetItemImages(child, false);
 
 			if (HasSubdirs(subListing, filter))
@@ -476,7 +540,7 @@ void CRemoteTreeView::DisplayItem(wxTreeItemId parent, const CDirectoryListing&
 		}
 		else
 		{
-			wxTreeItemId child = AppendItem(parent, name, 1, 3, new CItemData(subdir));
+			wxTreeItemId child = AppendItem(parent, name, 1, 3, 0);
 			SetItemImages(child, true);
 		}
 	}
@@ -494,7 +558,7 @@ static bool sortfunc(const wxString& a, const wxString& b)
 }
 
 
-void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing& listing)
+void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing& listing, bool will_select_parent)
 {
 	SetItemImages(parent, false);
 
@@ -506,7 +570,9 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 		return;
 	}
 
-	CFilterDialog filter;
+	CFilterManager filter;
+
+	const wxString path = listing.path.GetPath();
 
 	std::list<wxString> dirs;
 	for (unsigned int i = 0; i < listing.GetCount(); i++)
@@ -514,7 +580,7 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 		if (!listing[i].dir)
 			continue;
 
-		if (!filter.FilenameFiltered(listing[i].name, true, -1, false))
+		if (!filter.FilenameFiltered(listing[i].name, path, true, -1, false, 0))
 			dirs.push_back(listing[i].name);
 	}
 
@@ -549,32 +615,25 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 		}
 		else if (cmp > 0)
 		{
+			// Child no longer exists
 			wxTreeItemId sel = GetSelection();
-			if (sel)
-			{
-				do
-				{
-					sel = GetItemParent(sel);
-					if (sel == parent)
-						break;
-				} while (sel);
-			}
-			if (!sel)
-			{
-				wxTreeItemId prev = GetPrevSibling(child);
+			while (sel && sel != child)
+				sel = GetItemParent(sel);
+			wxTreeItemId prev = GetPrevSibling(child);
+			if (!sel || will_select_parent)
 				Delete(child);
-				child = prev;
-			}
+			child = prev;
 		}
 		else if (cmp < 0)
 		{
+			// New directory
 			CServerPath path = listing.path;
 			path.AddSegment(*iter);
 
 			CDirectoryListing subListing;
 			if (m_pState->m_pEngine->CacheLookup(path, subListing) == FZ_REPLY_OK)
 			{
-				wxTreeItemId child = AppendItem(parent, *iter, 0, 2, new CItemData(path));
+				wxTreeItemId child = AppendItem(parent, *iter, 0, 2, 0);
 				SetItemImages(child, false);
 
 				if (HasSubdirs(subListing, filter))
@@ -582,9 +641,9 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 			}
 			else
 			{
-				wxTreeItemId child = AppendItem(parent, *iter, 1, 3, new CItemData(path));
-                if (child)
-				    SetItemImages(child, true);
+				wxTreeItemId child = AppendItem(parent, *iter, 1, 3, 0);
+				if (child)
+					SetItemImages(child, true);
 			}
 
 			iter++;
@@ -593,22 +652,14 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 	}
 	while (child)
 	{
+		// Child no longer exists
 		wxTreeItemId sel = GetSelection();
-		if (sel)
-		{
-			do
-			{
-				sel = GetItemParent(sel);
-				if (sel == parent)
-					break;
-			} while (sel);
-		}
-		if (!sel)
-		{
-			wxTreeItemId prev = GetPrevSibling(child);
+		while (sel && sel != child)
+			sel = GetItemParent(sel);
+		wxTreeItemId prev = GetPrevSibling(child);
+		if (!sel || will_select_parent)
 			Delete(child);
-			child = prev;
-		}
+		child = prev;
 	}
 	while (iter != dirs.rend())
 	{
@@ -618,7 +669,7 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 		CDirectoryListing subListing;
 		if (m_pState->m_pEngine->CacheLookup(path, subListing) == FZ_REPLY_OK)
 		{
-			wxTreeItemId child = AppendItem(parent, *iter, 0, 2, new CItemData(path));
+			wxTreeItemId child = AppendItem(parent, *iter, 0, 2, 0);
 			SetItemImages(child, false);
 
 			if (HasSubdirs(subListing, filter))
@@ -626,7 +677,7 @@ void CRemoteTreeView::RefreshItem(wxTreeItemId parent, const CDirectoryListing&
 		}
 		else
 		{
-			wxTreeItemId child = AppendItem(parent, *iter, 1, 3, new CItemData(path));
+			wxTreeItemId child = AppendItem(parent, *iter, 1, 3, 0);
 			SetItemImages(child, true);
 		}
 
@@ -658,37 +709,22 @@ void CRemoteTreeView::OnItemExpanding(wxTreeEvent& event)
 	if (!item)
 		return;
 
-	const CItemData* data = (CItemData*)GetItemData(item);
-	wxASSERT(data);
-	if (!data)
-		return;
-
-	CDirectoryListing listing;
-	if (m_pState->m_pEngine->CacheLookup(data->m_path, listing) == FZ_REPLY_OK)
-		RefreshItem(item, listing);
-	else
+	if (!ListExpand(item))
 	{
-		SetItemImages(item, true);
-
-		wxTreeItemId child = GetLastChild(item);
-		if (!child)
-		{
-			event.Veto();
-			return;
-		}
-		if (GetItemText(child) == _T(""))
-		{
-			DeleteChildren(item);
-			event.Veto();
-		}
+		event.Veto();
+		return;
 	}
+
 	Refresh(false);
 }
 
 void CRemoteTreeView::SetItemImages(wxTreeItemId item, bool unknown)
 {
+	const int old_image = GetItemImage(item, wxTreeItemIcon_Normal);
 	if (!unknown)
 	{
+		if (old_image == 0)
+			return;
 		SetItemImage(item, 0, wxTreeItemIcon_Normal);
 		SetItemImage(item, 2, wxTreeItemIcon_Selected);
 		SetItemImage(item, 0, wxTreeItemIcon_Expanded);
@@ -696,6 +732,8 @@ void CRemoteTreeView::SetItemImages(wxTreeItemId item, bool unknown)
 	}
 	else
 	{
+		if (old_image == 1)
+			return;
 		SetItemImage(item, 1, wxTreeItemIcon_Normal);
 		SetItemImage(item, 3, wxTreeItemIcon_Selected);
 		SetItemImage(item, 1, wxTreeItemIcon_Expanded);
@@ -720,12 +758,12 @@ void CRemoteTreeView::OnSelectionChanged(wxTreeEvent& event)
 	if (!item)
 		return;
 
-	const CItemData* data = (CItemData*)GetItemData(item);
-	wxASSERT(data);
-	if (!data)
+	const CServerPath path = GetPathFromItem(item);
+	wxASSERT(!path.IsEmpty());
+	if (path.IsEmpty())
 		return;
 
-	m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(data->m_path));
+	m_pState->ChangeRemoteDir(path);
 }
 
 void CRemoteTreeView::OnItemActivated(wxTreeEvent& event)
@@ -736,12 +774,28 @@ void CRemoteTreeView::OnItemActivated(wxTreeEvent& event)
 
 CServerPath CRemoteTreeView::GetPathFromItem(const wxTreeItemId& item) const
 {
-	const CItemData* const pData = (const CItemData*)GetItemData(item);
+	std::list<wxString> segments;
 
-	if (!pData)
-		return CServerPath();
+	wxTreeItemId i = item;
+	while (i != GetRootItem())
+	{
+		const CItemData* const pData = (const CItemData*)GetItemData(i);
+		if (pData)
+		{
+			CServerPath path = pData->m_path;
+			for (std::list<wxString>::const_iterator iter = segments.begin(); iter != segments.end(); iter++)
+			{
+				if (!path.AddSegment(*iter))
+					return CServerPath();
+			}
+			return path;
+		}
 
-	return pData->m_path;
+		segments.push_front(GetItemText(i));
+		i = GetItemParent(i);
+	}
+
+	return CServerPath();
 }
 
 void CRemoteTreeView::OnBeginDrag(wxTreeEvent& event)
@@ -773,7 +827,7 @@ void CRemoteTreeView::OnBeginDrag(wxTreeEvent& event)
 		return;
 
 	CRemoteDataObject *pRemoteDataObject = new CRemoteDataObject(*pServer, parent);
-	pRemoteDataObject->AddFile(lastSegment, true, -1);
+	pRemoteDataObject->AddFile(lastSegment, true, -1, false);
 
 	pRemoteDataObject->Finalize();
 
@@ -794,9 +848,19 @@ void CRemoteTreeView::OnBeginDrag(wxTreeEvent& event)
 	}
 #endif
 
+	CDragDropManager* pDragDropManager = CDragDropManager::Init();
+	pDragDropManager->pDragSource = this;
+	pDragDropManager->server = *pServer;
+	pDragDropManager->remoteParent = parent;
+
 	wxDropSource source(this);
 	source.SetData(object);
-	if (source.DoDragDrop() != wxDragCopy)
+
+	int res = source.DoDragDrop();
+
+	pDragDropManager->Release();
+
+	if (res != wxDragCopy)
 	{
 #if FZ3_USESHELLEXT
 		delete ext;
@@ -821,8 +885,8 @@ void CRemoteTreeView::OnBeginDrag(wxTreeEvent& event)
 				return;
 			}
 
-			wxString target = ext->GetTarget();
-			if (target == _T(""))
+			CLocalPath target(ext->GetTarget());
+			if (target.empty())
 			{
 				delete ext;
 				ext = 0;
@@ -867,7 +931,7 @@ void CRemoteTreeView::OnContextMenu(wxTreeEvent& event)
 	if (!pMenu)
 		return;
 
-	const CServerPath& path = GetPathFromItem(m_contextMenuItem);
+	const CServerPath& path = m_contextMenuItem ? GetPathFromItem(m_contextMenuItem) : CServerPath();
 	if (!m_pState->IsRemoteIdle() || path.IsEmpty())
 	{
 		pMenu->Enable(XRCID("ID_DOWNLOAD"), false);
@@ -877,10 +941,15 @@ void CRemoteTreeView::OnContextMenu(wxTreeEvent& event)
 		pMenu->Enable(XRCID("ID_CHMOD"), false);
 		pMenu->Enable(XRCID("ID_MKDIR"), false);
 		pMenu->Enable(XRCID("ID_RENAME"), false);
+		pMenu->Enable(XRCID("ID_GETURL"), false);
 	}
 	else if (!path.HasParent())
-	{
 		pMenu->Enable(XRCID("ID_RENAME"), false);
+
+	if (!m_pState->GetLocalDir().IsWriteable())
+	{
+		pMenu->Enable(XRCID("ID_DOWNLOAD"), false);
+		pMenu->Enable(XRCID("ID_ADDTOQUEUE"), false);
 	}
 
 	PopupMenu(pMenu);
@@ -959,7 +1028,7 @@ void CRemoteTreeView::OnMenuChmod(wxCommandEvent& event)
 		// Change directory permissions
 		if (!applyType || applyType == 2)
 		{
-			wxString newPerms = pChmodDlg->GetPermissions(permissions);
+			wxString newPerms = pChmodDlg->GetPermissions(permissions, true);
 
 			m_pState->m_pCommandQueue->ProcessCommand(new CChmodCommand(path.GetParent(), name, newPerms));
 		}
@@ -984,14 +1053,30 @@ void CRemoteTreeView::OnMenuChmod(wxCommandEvent& event)
 		const wxTreeItemId selected = GetSelection();
 		if (selected)
 			currentPath = GetPathFromItem(selected);
-		pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_chmod, hasParent ? path.GetParent() : path, !cached, currentPath);
+		CFilterManager filter;
+		pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_chmod, hasParent ? path.GetParent() : path, filter.GetActiveFilters(false), !cached, currentPath);
 	}
 	else
+	{
 		pChmodDlg->Destroy();
+		const wxTreeItemId selected = GetSelection();
+		if (selected)
+		{
+			CServerPath currentPath = GetPathFromItem(selected);
+			m_pState->ChangeRemoteDir(currentPath);
+		}
+	}
 }
 
 void CRemoteTreeView::OnMenuDownload(wxCommandEvent& event)
 {
+	CLocalPath localDir = m_pState->GetLocalDir();
+	if (!localDir.IsWriteable())
+	{
+		wxBell();
+		return;
+	}
+
 	if (!m_pState->IsRemoteIdle())
 		return;
 
@@ -1002,19 +1087,12 @@ void CRemoteTreeView::OnMenuDownload(wxCommandEvent& event)
 	if (path.IsEmpty())
 		return;
 
-	const bool hasParent = path.HasParent();
-
-	const wxString localDir = m_pState->GetLocalDir();
-
 	const wxString& name = GetItemText(m_contextMenuItem);
 
-	wxFileName fn = wxFileName(localDir, _T(""));
-	
-	if (hasParent)
-		fn.AppendDir(name);	
+	localDir.AddSegment(name);
 
 	CRecursiveOperation* pRecursiveOperation = m_pState->GetRecursiveOperationHandler();
-	pRecursiveOperation->AddDirectoryToVisit(path, _T(""), fn.GetFullPath());
+	pRecursiveOperation->AddDirectoryToVisit(path, _T(""), localDir.GetPath());
 
 	CServerPath currentPath;
 	const wxTreeItemId selected = GetSelection();
@@ -1022,7 +1100,8 @@ void CRemoteTreeView::OnMenuDownload(wxCommandEvent& event)
 		currentPath = GetPathFromItem(selected);
 
 	const bool addOnly = event.GetId() == XRCID("ID_ADDTOQUEUE");
-	pRecursiveOperation->StartRecursiveOperation(addOnly ? CRecursiveOperation::recursive_addtoqueue : CRecursiveOperation::recursive_download, path, true, currentPath);
+	CFilterManager filter;
+	pRecursiveOperation->StartRecursiveOperation(addOnly ? CRecursiveOperation::recursive_addtoqueue : CRecursiveOperation::recursive_download, path, filter.GetActiveFilters(false), true, currentPath);
 }
 
 void CRemoteTreeView::OnMenuDelete(wxCommandEvent& event)
@@ -1061,10 +1140,11 @@ void CRemoteTreeView::OnMenuDelete(wxCommandEvent& event)
 	const wxTreeItemId selected = GetSelection();
 	if (selected)
 		currentPath = GetPathFromItem(selected);
-	if (!currentPath.IsEmpty() && path.IsParentOf(currentPath, false))
+	if (!currentPath.IsEmpty() && (path == currentPath || path.IsParentOf(currentPath, false)))
 		currentPath = startDir;
 
-	pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_delete, startDir, !hasParent, currentPath);
+	CFilterManager filter;
+	pRecursiveOperation->StartRecursiveOperation(CRecursiveOperation::recursive_delete, startDir, filter.GetActiveFilters(false), !hasParent, currentPath);
 }
 
 void CRemoteTreeView::OnMenuRename(wxCommandEvent& event)
@@ -1121,20 +1201,15 @@ void CRemoteTreeView::OnEndLabelEdit(wxTreeEvent& event)
 		return;
 	}
 
-	const CServerPath& path = GetPathFromItem(event.GetItem());
-	if (path.IsEmpty())
+	CItemData* const pData = (CItemData*)GetItemData(event.GetItem());
+	if (pData)
 	{
 		event.Veto();
 		return;
 	}
 
-	if (!path.HasParent())
-	{
-		event.Veto();
-		return;
-	}
-
-	CServerPath parent = path.GetParent();
+	CServerPath old_path = GetPathFromItem(event.GetItem());
+	CServerPath parent = old_path.GetParent();
 
 	const wxString& oldName = GetItemText(event.GetItem());
 	const wxString& newName = event.GetLabel();
@@ -1145,7 +1220,7 @@ void CRemoteTreeView::OnEndLabelEdit(wxTreeEvent& event)
 	}
 
 	m_pState->m_pCommandQueue->ProcessCommand(new CRenameCommand(parent, oldName, parent, newName));
-	m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(parent));
+	m_pState->ChangeRemoteDir(parent);
 
 	CServerPath currentPath;
 	const wxTreeItemId selected = GetSelection();
@@ -1154,12 +1229,12 @@ void CRemoteTreeView::OnEndLabelEdit(wxTreeEvent& event)
 	if (currentPath.IsEmpty())
 		return;
 
-	if (currentPath == path || currentPath.IsSubdirOf(path, false))
+	if (currentPath == old_path || currentPath.IsSubdirOf(old_path, false))
 	{
 		// Previously selected path was below renamed one, list the new one
 
 		std::list<wxString> subdirs;
-		while (currentPath != path)
+		while (currentPath != old_path)
 		{
 			if (!currentPath.HasParent())
 			{
@@ -1173,10 +1248,10 @@ void CRemoteTreeView::OnEndLabelEdit(wxTreeEvent& event)
 		currentPath.AddSegment(newName);
 		for (std::list<wxString>::const_iterator iter = subdirs.begin(); iter != subdirs.end(); iter++)
 			currentPath.AddSegment(*iter);
-		m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(currentPath));
+		m_pState->ChangeRemoteDir(currentPath);
 	}
 	else if (currentPath != parent)
-		m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(currentPath));
+		m_pState->ChangeRemoteDir(currentPath);
 }
 
 void CRemoteTreeView::OnMkdir(wxCommandEvent& event)
@@ -1198,15 +1273,15 @@ void CRemoteTreeView::OnMkdir(wxCommandEvent& event)
 	CServerPath newPath = path;
 
 	// Append a long segment which does (most likely) not exist in the path and
-	// replace it with "New folder" later. This way we get the exact position of
-	// "New folder" and can preselect it in the dialog.
+	// replace it with "New directory" later. This way we get the exact position of
+	// "New directory" and can preselect it in the dialog.
 	wxString tmpName = _T("25CF809E56B343b5A12D1F0466E3B37A49A9087FDCF8412AA9AF8D1E849D01CF");
 	if (newPath.AddSegment(tmpName))
 	{
 		wxString pathName = newPath.GetPath();
 		int pos = pathName.Find(tmpName);
 		wxASSERT(pos != -1);
-		wxString newName = _("New folder");
+		wxString newName = _("New directory");
 		pathName.Replace(tmpName, newName);
 		dlg.SetValue(pathName);
 		dlg.SelectText(pos, pos + newName.Length());
@@ -1227,7 +1302,7 @@ void CRemoteTreeView::OnMkdir(wxCommandEvent& event)
 	if (newPath.HasParent())
 	{
 		listed = newPath.GetParent();
-		m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(listed));
+		m_pState->ChangeRemoteDir(listed);
 	}
 
 	CServerPath currentPath;
@@ -1235,5 +1310,152 @@ void CRemoteTreeView::OnMkdir(wxCommandEvent& event)
 	if (selected)
 		currentPath = GetPathFromItem(selected);
 	if (!currentPath.IsEmpty() && currentPath != listed)
-		m_pState->m_pCommandQueue->ProcessCommand(new CListCommand(currentPath));
+		m_pState->ChangeRemoteDir(currentPath);
+}
+
+bool CRemoteTreeView::ListExpand(wxTreeItemId item)
+{
+	const CServerPath path = GetPathFromItem(item);
+	wxASSERT(!path.IsEmpty());
+	if (path.IsEmpty())
+		return false;
+
+	CDirectoryListing listing;
+	if (m_pState->m_pEngine->CacheLookup(path, listing) == FZ_REPLY_OK)
+		RefreshItem(item, listing, false);
+	else
+	{
+		SetItemImages(item, true);
+
+		wxTreeItemId child = GetLastChild(item);
+		if (!child || GetItemText(child) == _T(""))
+			return false;
+	}
+
+	return true;
+}
+
+void CRemoteTreeView::OnChar(wxKeyEvent& event)
+{
+	m_contextMenuItem = GetSelection();
+
+	wxCommandEvent cmdEvt;
+	if (event.GetKeyCode() == WXK_F2)
+		OnMenuRename(cmdEvt);
+	else if (event.GetKeyCode() == WXK_DELETE)
+		OnMenuDelete(cmdEvt);
+	else
+		event.Skip();
+}
+
+struct _parents
+{
+	wxTreeItemId item;
+	CServerPath path;
+};
+
+void CRemoteTreeView::ApplyFilters()
+{
+	std::list<struct _parents> parents;
+
+	const wxTreeItemId root = GetRootItem();
+	wxTreeItemIdValue cookie;
+	for (wxTreeItemId child = GetFirstChild(root, cookie); child; child = GetNextSibling(child))
+	{
+		CServerPath path = GetPathFromItem(child);
+		if (path.IsEmpty())
+			continue;
+
+		struct _parents dir;
+		dir.item = child;
+		dir.path = path;
+		parents.push_back(dir);
+	}
+	
+	CFilterManager filter;
+	while (!parents.empty())
+	{
+		struct _parents parent = parents.back();
+		parents.pop_back();
+
+		CDirectoryListing listing;
+		if (m_pState->m_pEngine->CacheLookup(parent.path, listing) == FZ_REPLY_OK)
+			RefreshItem(parent.item, listing, false);
+		else if (filter.HasActiveFilters())
+		{
+			for (wxTreeItemId child = GetFirstChild(parent.item, cookie); child; child = GetNextSibling(child))
+			{
+				CServerPath path = GetPathFromItem(child);
+				if (path.IsEmpty())
+					continue;
+
+				if (filter.FilenameFiltered(GetItemText(child), path.GetPath(), true, -1, false, 0))
+				{
+					wxTreeItemId sel = GetSelection();
+					while (sel && sel != child)
+						sel = GetItemParent(sel);
+					if (!sel)
+					{
+						Delete(child);
+						continue;
+					}
+				}
+				
+				struct _parents dir;
+				dir.item = child;
+				dir.path = path;
+				parents.push_back(dir);
+			}
+
+			// The stuff below has already been done above in this one case
+			continue;
+		}
+		for (wxTreeItemId child = GetFirstChild(parent.item, cookie); child; child = GetNextSibling(child))
+		{
+			CServerPath path = GetPathFromItem(child);
+			if (path.IsEmpty())
+				continue;
+
+			struct _parents dir;
+			dir.item = child;
+			dir.path = path;
+			parents.push_back(dir);
+		}
+	}
+}
+
+void CRemoteTreeView::OnMenuGeturl(wxCommandEvent& event)
+{
+	if (!m_contextMenuItem)
+		return;
+
+	const CServerPath& path = GetPathFromItem(m_contextMenuItem);
+	if (path.IsEmpty())
+	{
+		wxBell();
+		return;
+	}
+
+	const CServer *pServer = m_pState->GetServer();
+	if (!pServer)
+	{
+		wxBell();
+		return;
+	}
+
+	if (!wxTheClipboard->Open())
+	{
+		wxMessageBox(_("Could not open clipboard"), _("Could not copy URLs"), wxICON_EXCLAMATION);
+		return;
+	}
+
+	wxString url = pServer->FormatServer(true);
+	url += path.GetPath();
+
+	// Poor mans URLencode
+	url.Replace(_T(" "), _T("%20"));
+
+	wxTheClipboard->SetData(new wxURLDataObject(url));
+
+	wxTheClipboard->Close();
 }
