@@ -18,22 +18,25 @@ public:
 
 protected:
 	wxTreeItemId MakeParent(CServerPath path, bool select);
-	void SetDirectoryListing(const CDirectoryListing* pListing, bool modified);
-	virtual void OnStateChange(unsigned int event, const wxString& data);
+	void SetDirectoryListing(const CSharedPointer<const CDirectoryListing> &pListing, bool modified);
+	virtual void OnStateChange(enum t_statechange_notifications notification, const wxString& data);
 
 	void DisplayItem(wxTreeItemId parent, const CDirectoryListing& listing);
-	void RefreshItem(wxTreeItemId parent, const CDirectoryListing& listing);
+	void RefreshItem(wxTreeItemId parent, const CDirectoryListing& listing, bool will_select_parent);
 
 	void SetItemImages(wxTreeItemId item, bool unknown);
 
-	bool HasSubdirs(const CDirectoryListing& listing, const CFilterDialog& filter);
+	bool HasSubdirs(const CDirectoryListing& listing, const CFilterManager& filter);
 
 	virtual int OnCompareItems(const wxTreeItemId& item1, const wxTreeItemId& item2);
 
 	CServerPath GetPathFromItem(const wxTreeItemId& item) const;
 
+	bool ListExpand(wxTreeItemId item);
+
+	void ApplyFilters();
+
 	CQueueView* m_pQueue;
-	const CDirectoryListing* m_pDirectoryListing;
 
 	void CreateImageList();
 	wxBitmap CreateIcon(int index, const wxString& overlay = _T(""));
@@ -63,6 +66,8 @@ protected:
 	void OnBeginLabelEdit(wxTreeEvent& event);
 	void OnEndLabelEdit(wxTreeEvent& event);
 	void OnMkdir(wxCommandEvent& event);
+	void OnChar(wxKeyEvent& event);
+	void OnMenuGeturl(wxCommandEvent& event);
 
 	wxTreeItemId m_contextMenuItem;
 };
