@@ -39,7 +39,7 @@ CNetConfWizard::~CNetConfWizard()
 
 bool CNetConfWizard::Load()
 {
-	if (!Create(m_parent, wxID_ANY, _("Firewall and router configuration wizard")))
+	if (!Create(m_parent, wxID_ANY, _("Firewall and router configuration wizard"), wxNullBitmap, wxPoint(0, 0)))
 		return false;
 
 	wxSize minPageSize = GetPageAreaSizer()->GetMinSize();
@@ -64,6 +64,8 @@ bool CNetConfWizard::Load()
 		windows.push_back(m_pages[i]);
 	wxGetApp().GetWrapEngine()->WrapRecursive(windows, 1.7, "Netconf", wxSize(), minPageSize);
 
+	CenterOnParent();
+
 	// Load values
 
 	switch (m_pOptions->GetOptionVal(OPTION_USEPASV))
@@ -310,7 +312,7 @@ void CNetConfWizard::OnSend()
 	int written = m_socket->LastCount();
 	if (written == len)
 	{
-		delete m_pSendBuffer;
+		delete [] m_pSendBuffer;
 		m_pSendBuffer = 0;
 	}
 	else
@@ -398,14 +400,14 @@ void CNetConfWizard::ParseResponse(const char* line)
 	if (len < 3)
 	{
 		m_testResult = servererror;
-		PrintMessage(_("Server sent invalid reply."), 1);
+		PrintMessage(_("Server sent unexpected reply."), 1);
 		CloseSocket();
 		return;
 	}
 	if (line[3] && line[3] != ' ')
 	{
 		m_testResult = servererror;
-		PrintMessage(_("Server sent invalid reply."), 1);
+		PrintMessage(_("Server sent unexpected reply."), 1);
 		CloseSocket();
 		return;
 	}
@@ -444,7 +446,7 @@ void CNetConfWizard::ParseResponse(const char* line)
 		else
 		{
 			m_testResult = servererror;
-			PrintMessage(_("Server sent invalid reply."), 1);
+			PrintMessage(_("Server sent unexpected reply."), 1);
 			CloseSocket();
 			return;
 		}
@@ -453,7 +455,7 @@ void CNetConfWizard::ParseResponse(const char* line)
 		if (line[0] != '2')
 		{
 			m_testResult = servererror;
-			PrintMessage(_("Server sent invalid reply."), 1);
+			PrintMessage(_("Server sent unexpected reply."), 1);
 			CloseSocket();
 			return;
 		}
@@ -465,7 +467,7 @@ void CNetConfWizard::ParseResponse(const char* line)
 				if (*p < '0' || *p > '9')
 				{
 					m_testResult = servererror;
-					PrintMessage(_("Server sent invalid reply."), 1);
+					PrintMessage(_("Server sent unexpected reply."), 1);
 					CloseSocket();
 					return;
 				}
@@ -489,14 +491,14 @@ void CNetConfWizard::ParseResponse(const char* line)
 		}
 
 		m_testResult = servererror;
-		PrintMessage(_("Server sent invalid reply."), 1);
+		PrintMessage(_("Server sent unexpected reply."), 1);
 		CloseSocket();
 		return;
 	case 6:
 		if (line[0] != '2' && line[0] != '3')
 		{
 			m_testResult = servererror;
-			PrintMessage(_("Server sent invalid reply."), 1);
+			PrintMessage(_("Server sent unexpected reply."), 1);
 			CloseSocket();
 			return;
 		}
@@ -505,7 +507,7 @@ void CNetConfWizard::ParseResponse(const char* line)
 			if (gotListReply)
 			{
 				m_testResult = servererror;
-				PrintMessage(_("Server sent invalid reply."), 1);
+				PrintMessage(_("Server sent unexpected reply."), 1);
 				CloseSocket();
 			}
 			gotListReply = true;
@@ -516,7 +518,7 @@ void CNetConfWizard::ParseResponse(const char* line)
 		if (line[0] != '2' && line[0] != '3')
 		{
 			m_testResult = servererror;
-			PrintMessage(_("Server sent invalid reply."), 1);
+			PrintMessage(_("Server sent unexpected reply."), 1);
 			CloseSocket();
 			return;
 		}
@@ -549,6 +551,7 @@ void CNetConfWizard::CloseSocket()
 	{
 		text[0] = _("Connection with the test server failed.");
 		text[1] = _("Please check on http://filezilla-project.org/probe.php that the server is running and carefully check your settings again.");
+		text[2] = _("If the problem persists, some router and/or firewall keeps blocking FileZilla.");
 	}
 	else
 	{
@@ -556,7 +559,9 @@ void CNetConfWizard::CloseSocket()
 		{
 		case unknown:
 			text[0] = _("Connection with server got closed prematurely.");
-			text[1] = _("Please ensure you have a stable internet connection and check on http://filezilla-project.org/probe.php that the server is running.");
+			text[1] = _("Please ensure you have a stable internet connection and carefully check your settings again.");
+			text[2] = _("If the problem persists, some router and/or firewall keeps interrupting the connection.");
+			text[3] = _("See also: http://wiki.filezilla-project.org/Network_Configuration");
 			break;
 		case successful:
 			text[0] = _("Congratulations, your configuration seems to be working.");
@@ -566,9 +571,10 @@ void CNetConfWizard::CloseSocket()
 			text[4] = _("Click on Finish to save your configuration.");
 			break;
 		case servererror:
-			text[0] = _("The server sent an unrecognized reply.");
-			text[1] = _("Please make sure you are running the latest version of FileZilla. Visit http://filezilla-project.org for details.");
-			text[2] = _("Submit a bug report if this problem persists even with the latest version of FileZilla.");
+			text[0] = _("The server sent an unexpected or unrecognized reply.");
+			text[1] = _("This means that some router and/or firewall is still interfering with FileZilla.");
+			text[2] = _("Re-run the wizard and carefully check your settings and configure all routers and firewalls accordingly.");
+			text[3] = _("See also: http://wiki.filezilla-project.org/Network_Configuration");
 			break;
 		case tainted:
 			text[0] = _("Active mode FTP test failed. FileZilla knows the correct external IP address, but your router or firewall has misleadingly modified the sent address.");
@@ -585,7 +591,7 @@ void CNetConfWizard::CloseSocket()
 		case mismatchandtainted:
 			text[0] = _("Active mode FTP test failed. FileZilla does not know the correct external IP address. In addition to that, your router has modified the sent address.");
 			text[1] = _("Please enter your external IP address on the active mode page of this wizard. In case you have a dynamic address or don't know your external address, use the external resolver option.");
-			text[2] = _("Please make sure your router is using the latest available firmware. Furthermore, your router has to be configured properly. You will have to manual port forwarding. Don't run your router in the so called 'DMZ mode' or 'game mode'.");
+			text[2] = _("Please make sure your router is using the latest available firmware. Furthermore, your router has to be configured properly. You will have to use manual port forwarding. Don't run your router in the so called 'DMZ mode' or 'game mode'.");
 			text[3] = _("If your router keeps changing the IP address, please contact your router manufacturer.");
 			text[4] = _("Unless these problems get fixed, active mode FTP will not work and passive mode has to be used.");
 			if (XRCCTRL(*this, "ID_ACTIVE", wxRadioButton)->GetValue())
@@ -608,12 +614,12 @@ void CNetConfWizard::CloseSocket()
 			break;
 		case externalfailed:
 			text[0] = _("Failed to retrieve the external IP address.");
-			text[1] = _("Pleae make sure FileZilla is allowed to establish outgoing connections and make sure you typed the address of the address resolver correctly.");
+			text[1] = _("Please make sure FileZilla is allowed to establish outgoing connections and make sure you typed the address of the address resolver correctly.");
 			text[2] = wxString::Format(_("The address you entered was: %s"), XRCCTRL(*this, "ID_ACTIVERESOLVER", wxTextCtrl)->GetValue().c_str());
 			break;
 		case datatainted:
 			text[0] = _("Transferred data got tainted.");
-			text[1] = _("You likely have a router or firewall which errorneously modified the transferred data.");
+			text[1] = _("You likely have a router or firewall which erroneously modified the transferred data.");
 			text[2] = _("Please disable settings like 'DMZ mode' or 'Game mode' on your router.");
 			text[3] = _("If this problem persists, please contact your router or firewall manufacturer for a solution.");
 			break;
@@ -878,7 +884,9 @@ int CNetConfWizard::CreateListenSocket()
 		XRCCTRL(*this, "ID_ACTIVE_PORTMIN", wxTextCtrl)->GetValue().ToLong(&low);
 		XRCCTRL(*this, "ID_ACTIVE_PORTMAX", wxTextCtrl)->GetValue().ToLong(&high);
 		
-		int mid = (rand() * (high - low)) / (RAND_MAX + 1) + low;
+		int mid = GetRandomNumber(low, high);
+		wxASSERT(mid >= low && mid <= high);
+
 		for (int port = mid; port <= high; port++)
 			if (CreateListenSocket(port))
 				return port;
@@ -939,7 +947,7 @@ void CNetConfWizard::OnAccept()
 	{
 		delete m_pDataSocket;
 		m_pDataSocket = 0;
-		PrintMessage(_("Failed to get peer address, connection closed."), 1);
+		PrintMessage(_("Failed to get peer address of control connection, connection closed."), 1);
 		CloseSocket();
 		return;
 	}
@@ -947,7 +955,7 @@ void CNetConfWizard::OnAccept()
 	{
 		delete m_pDataSocket;
 		m_pDataSocket = 0;
-		PrintMessage(_("Failed to get data peer address, connection closed."), 1);
+		PrintMessage(_("Failed to get peer address of data connection, connection closed."), 1);
 		CloseSocket();
 		return;
 	}
