@@ -9,24 +9,31 @@ CQueueItem::CQueueItem()
 	m_visibleOffspring = 0;
 	m_parent = 0;
 	m_maxCachedIndex = -1;
+
+	m_removed_at_front = 0;
 }
 
 CQueueItem::~CQueueItem()
 {
 	std::vector<CQueueItem*>::iterator iter;
-	for (iter = m_children.begin(); iter != m_children.end(); iter++)
+	for (iter = m_children.begin() + m_removed_at_front; iter != m_children.end(); iter++)
 		delete *iter;
 }
 
 void CQueueItem::SetPriority(enum QueuePriority priority)
 {
 	std::vector<CQueueItem*>::iterator iter;
-	for (iter = m_children.begin(); iter != m_children.end(); iter++)
+	for (iter = m_children.begin() + m_removed_at_front; iter != m_children.end(); iter++)
 		(*iter)->SetPriority(priority);
 }
 
 void CQueueItem::AddChild(CQueueItem* item)
 {
+	if (m_removed_at_front)
+	{
+		m_children.erase(m_children.begin(), m_children.begin() + m_removed_at_front);
+		m_removed_at_front = 0;
+	}
 	item->m_indent = m_indent + _T("  ");
 	m_children.push_back(item);
 
@@ -46,17 +53,17 @@ CQueueItem* CQueueItem::GetChild(unsigned int item, bool recursive /*=true*/)
 	std::vector<CQueueItem*>::iterator iter;
 	if (!recursive)
 	{
-		if (item >= m_children.size())
+		if (item + m_removed_at_front >= m_children.size())
 			return 0;
 		iter = m_children.begin();
-		iter += item;
+		iter += item + m_removed_at_front;
 		return *iter;
 	}
 
 	if ((int)item <= m_maxCachedIndex)
 	{
 		// Index is cached
-		iter = m_children.begin();
+		iter = m_children.begin() + m_removed_at_front;
 		iter += m_lookupCache[item].child;
 		item -= m_lookupCache[item].index;
 		if (!item)
@@ -64,10 +71,10 @@ CQueueItem* CQueueItem::GetChild(unsigned int item, bool recursive /*=true*/)
 		else
 			return (*iter)->GetChild(item - 1);
 	}
-	
+
 	int index;
 	int child;
-	iter = m_children.begin();
+	iter = m_children.begin() + m_removed_at_front;
 	if (m_maxCachedIndex == -1)
 	{
 		child = 0;
@@ -111,7 +118,7 @@ CQueueItem* CQueueItem::GetChild(unsigned int item, bool recursive /*=true*/)
 unsigned int CQueueItem::GetChildrenCount(bool recursive)
 {
 	if (!recursive)
-		return m_children.size();
+		return m_children.size() - m_removed_at_front;
 
 	return m_visibleOffspring;
 }
@@ -121,7 +128,7 @@ bool CQueueItem::RemoveChild(CQueueItem* pItem, bool destroy /*=true*/)
 	int oldVisibleOffspring = m_visibleOffspring;
 	std::vector<CQueueItem*>::iterator iter;
 	bool deleted = false;
-	for (iter = m_children.begin(); iter != m_children.end(); iter++)
+	for (iter = m_children.begin() + m_removed_at_front; iter != m_children.end(); iter++)
 	{
 		if (*iter == pItem)
 		{
@@ -129,7 +136,17 @@ bool CQueueItem::RemoveChild(CQueueItem* pItem, bool destroy /*=true*/)
 			m_visibleOffspring -= pItem->m_visibleOffspring;
 			if (destroy)
 				delete pItem;
-			m_children.erase(iter);
+
+			if (iter - m_children.begin() - m_removed_at_front <= 10)
+			{
+				m_removed_at_front++;
+				unsigned int end = iter - m_children.begin();
+				int c = 0;
+				for (int i = end; i >= m_removed_at_front; i--, c++)
+					m_children[i] = m_children[i - 1];
+			}
+			else
+				m_children.erase(iter);
 
 			deleted = true;
 			break;
@@ -139,17 +156,27 @@ bool CQueueItem::RemoveChild(CQueueItem* pItem, bool destroy /*=true*/)
 		if ((*iter)->RemoveChild(pItem, destroy))
 		{
 			m_visibleOffspring -= visibleOffspring - (*iter)->m_visibleOffspring;
-			if (!(*iter)->m_children.size())
+			if (!((*iter)->m_children.size() - (*iter)->m_removed_at_front))
 			{
 				m_visibleOffspring -= 1;
 				delete *iter;
-				m_children.erase(iter);
+
+				if (iter - m_children.begin() - m_removed_at_front <= 10)
+				{
+					m_removed_at_front++;
+					unsigned int end = iter - m_children.begin();
+					for (int i = end; i >= m_removed_at_front; i--)
+						m_children[i] = m_children[i - 1];
+				}
+				else
+					m_children.erase(iter);
 			}
 
 			deleted = true;
 			break;
 		}
 	}
+
 	if (!deleted)
 		return false;
 
@@ -173,7 +200,7 @@ bool CQueueItem::TryRemoveAll()
 	std::vector<CQueueItem*>::iterator iter;
 	std::vector<CQueueItem*> keepChildren;
 	m_visibleOffspring = 0;
-	for (iter = m_children.begin(); iter != m_children.end(); iter++)
+	for (iter = m_children.begin() + m_removed_at_front; iter != m_children.end(); iter++)
 	{
 		CQueueItem* pItem = *iter;
 		if (pItem->TryRemoveAll())
@@ -187,6 +214,7 @@ bool CQueueItem::TryRemoveAll()
 	}
 	m_children = keepChildren;
 	m_maxCachedIndex = -1;
+	m_removed_at_front = 0;
 
 	CQueueItem* parent = GetParent();
 	while (parent)
@@ -238,7 +266,7 @@ int CQueueItem::GetItemIndex() const
 		return 0;
 
 	int index = 1;
-	for (std::vector<CQueueItem*>::const_iterator iter = pParent->m_children.begin(); iter != pParent->m_children.end(); iter++)
+	for (std::vector<CQueueItem*>::const_iterator iter = pParent->m_children.begin() + pParent->m_removed_at_front; iter != pParent->m_children.end(); iter++)
 	{
 		if (*iter == this)
 			break;
@@ -251,22 +279,21 @@ int CQueueItem::GetItemIndex() const
 
 CFileItem::CFileItem(CServerItem* parent, bool queued, bool download, const wxString& localFile,
 					 const wxString& remoteFile, const CServerPath& remotePath, wxLongLong size)
+	: m_localFile(localFile), m_remoteFile(remoteFile), m_remotePath(remotePath), m_size(size)
 {
 	m_parent = parent;
 	m_priority = priority_normal;
 
 	m_download = download;
-	m_localFile = localFile;
-	m_remoteFile = remoteFile;
-	m_remotePath = remotePath;
-	m_size = size;
 	m_queued = queued;
 	m_active = false;
 	m_errorCount = 0;
 	m_remove = false;
 	m_pEngineData = 0;
-	m_defaultFileExistsAction = -1;
-	m_edit = false;
+	m_defaultFileExistsAction = CFileExistsNotification::unknown;
+	m_edit = CEditHandler::none;
+	m_madeProgress = false;
+	m_onetime_action = CFileExistsNotification::unknown;
 }
 
 CFileItem::~CFileItem()
@@ -301,7 +328,7 @@ void CFileItem::SetActive(const bool active)
 	}
 	else if (!active && m_active)
 	{
-		CQueueItem* pItem = m_children.front();
+		CQueueItem* pItem = GetChild(0, false);
 		RemoveChild(pItem);
 	}
 	m_active = active;
@@ -309,7 +336,7 @@ void CFileItem::SetActive(const bool active)
 
 void CFileItem::SaveItem(TiXmlElement* pElement) const
 {
-	if (m_edit)
+	if (m_edit != CEditHandler::none)
 		return;
 
 	TiXmlElement file("File");
@@ -317,13 +344,14 @@ void CFileItem::SaveItem(TiXmlElement* pElement) const
 	AddTextElement(&file, "LocalFile", m_localFile);
 	AddTextElement(&file, "RemoteFile", m_remoteFile);
 	AddTextElement(&file, "RemotePath", m_remotePath.GetSafePath());
-	AddTextElement(&file, "Download", m_download ? _T("1") : _T("0"));
+	AddTextElementRaw(&file, "Download", m_download ? "1" : "0");
 	if (m_size != -1)
 		AddTextElement(&file, "Size", m_size.ToString());
 	if (m_errorCount)
 		AddTextElement(&file, "ErrorCount", m_errorCount);
-	AddTextElement(&file, "Priority", m_priority);
-	AddTextElement(&file, "TransferMode", m_transferSettings.binary ? _T("1") : _T("0"));
+	if (m_priority != priority_normal)
+		AddTextElement(&file, "Priority", m_priority);
+	AddTextElementRaw(&file, "TransferMode", m_transferSettings.binary ? "1" : "0");
 
 	pElement->InsertEndChild(file);
 }
@@ -337,6 +365,18 @@ bool CFileItem::TryRemoveAll()
 	return false;
 }
 
+void CFileItem::SetLocalFile(const wxString &file)
+{
+	wxASSERT(!file.empty());
+	m_localFile = file;
+}
+
+void CFileItem::SetRemoteFile(const wxString &file)
+{
+	wxASSERT(!file.empty());
+	m_remoteFile = file;
+}
+
 CFolderItem::CFolderItem(CServerItem* parent, bool queued, const wxString& localFile)
 	: CFileItem(parent, queued, true, localFile, _T(""), CServerPath(), -1)
 {
@@ -358,7 +398,7 @@ void CFolderItem::SaveItem(TiXmlElement* pElement) const
 		AddTextElement(&file, "RemoteFile", m_remoteFile);
 		AddTextElement(&file, "RemotePath", m_remotePath.GetSafePath());
 	}
-	AddTextElement(&file, "Download", m_download ? _T("1") : _T("0"));
+	AddTextElementRaw(&file, "Download", m_download ? "1" : "0");
 
 	pElement->InsertEndChild(file);
 }
@@ -418,9 +458,9 @@ void CServerItem::RemoveFileItemFromList(CFileItem* pItem)
 	wxFAIL_MSG(_T("File item not deleted from m_fileList"));
 }
 
-void CServerItem::SetDefaultFileExistsAction(int action, const enum TransferDirection direction)
+void CServerItem::SetDefaultFileExistsAction(enum CFileExistsNotification::OverwriteAction action, const enum TransferDirection direction)
 {
-	for (std::vector<CQueueItem*>::iterator iter = m_children.begin(); iter != m_children.end(); iter++)
+	for (std::vector<CQueueItem*>::iterator iter = m_children.begin() + m_removed_at_front; iter != m_children.end(); iter++)
 	{
 		CQueueItem *pItem = *iter;
 		if (pItem->GetType() == QueueItemType_File)
@@ -553,7 +593,7 @@ void CServerItem::SaveItem(TiXmlElement* pElement) const
 	TiXmlElement server("Server");
 	SetServer(&server, m_server);
 
-	for (std::vector<CQueueItem*>::const_iterator iter = m_children.begin(); iter != m_children.end(); iter++)
+	for (std::vector<CQueueItem*>::const_iterator iter = m_children.begin() + m_removed_at_front; iter != m_children.end(); iter++)
 		(*iter)->SaveItem(&server);
 
 	pElement->InsertEndChild(server);
@@ -579,7 +619,7 @@ wxLongLong CServerItem::GetTotalSize(int& filesWithUnknownSize, int& queuedFiles
 		}
 	}
 
-	for (std::vector<CQueueItem*>::const_iterator iter = m_children.begin(); iter != m_children.end(); iter++)
+	for (std::vector<CQueueItem*>::const_iterator iter = m_children.begin() + m_removed_at_front; iter != m_children.end(); iter++)
 	{
 		if ((*iter)->GetType() == QueueItemType_File ||
 			(*iter)->GetType() == QueueItemType_Folder)
@@ -597,7 +637,7 @@ bool CServerItem::TryRemoveAll()
 	std::vector<CQueueItem*>::iterator iter;
 	std::vector<CQueueItem*> keepChildren;
 	m_visibleOffspring = 0;
-	for (iter = m_children.begin(); iter != m_children.end(); iter++)
+	for (iter = m_children.begin() + m_removed_at_front; iter != m_children.end(); iter++)
 	{
 		CQueueItem* pItem = *iter;
 		if (pItem->TryRemoveAll())
@@ -617,6 +657,7 @@ bool CServerItem::TryRemoveAll()
 		}
 	}
 	m_children = keepChildren;
+	m_removed_at_front = 0;
 
 	CQueueItem* parent = GetParent();
 	while (parent)
@@ -635,6 +676,7 @@ void CServerItem::DetachChildren()
 	m_children.clear();
 	m_visibleOffspring = 0;
 	m_maxCachedIndex = -1;
+	m_removed_at_front = 0;
 
 	for (int i = 0; i < 2; i++)
 		for (int j = 0; j < PRIORITY_COUNT; j++)
@@ -644,7 +686,7 @@ void CServerItem::DetachChildren()
 void CServerItem::SetPriority(enum QueuePriority priority)
 {
 	std::vector<CQueueItem*>::iterator iter;
-	for (iter = m_children.begin(); iter != m_children.end(); iter++)
+	for (iter = m_children.begin() + m_removed_at_front; iter != m_children.end(); iter++)
 	{
 		if ((*iter)->GetType() == QueueItemType_File)
 			((CFileItem*)(*iter))->SetPriorityRaw(priority);
@@ -686,12 +728,9 @@ CFolderScanItem::CFolderScanItem(CServerItem* parent, bool queued, bool download
 	m_remove = false;
 	m_active = false;
 	m_count = 0;
-	t_dirPair pair;
-	pair.localPath = localPath.c_str();
-	pair.remotePath.SetSafePath(remotePath.GetSafePath().c_str());
-	m_dirsToCheck.push_back(pair);
+	m_dir_is_empty = false;
 
-	m_defaultFileExistsAction = -1;
+	m_defaultFileExistsAction = CFileExistsNotification::unknown;
 }
 
 bool CFolderScanItem::TryRemoveAll()
@@ -707,15 +746,17 @@ bool CFolderScanItem::TryRemoveAll()
 // CQueueViewBase
 // --------------
 
-BEGIN_EVENT_TABLE(CQueueViewBase, wxListCtrl)
+BEGIN_EVENT_TABLE(CQueueViewBase, wxListCtrlEx)
 EVT_ERASE_BACKGROUND(CQueueViewBase::OnEraseBackground)
 EVT_NAVIGATION_KEY(CQueueViewBase::OnNavigationKey)
 EVT_CHAR(CQueueViewBase::OnChar)
 EVT_LIST_COL_END_DRAG(wxID_ANY, CQueueViewBase::OnEndColumnDrag)
+EVT_TIMER(wxID_ANY, CQueueViewBase::OnTimer)
+EVT_KEY_DOWN(CQueueViewBase::OnKeyDown)
 END_EVENT_TABLE()
 
 CQueueViewBase::CQueueViewBase(CQueue* parent, int index, const wxString& title)
-	: wxListCtrl(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxCLIP_CHILDREN | wxLC_REPORT | wxLC_VIRTUAL | wxSUNKEN_BORDER | wxTAB_TRAVERSAL),
+	: wxListCtrlEx(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxCLIP_CHILDREN | wxLC_REPORT | wxLC_VIRTUAL | wxSUNKEN_BORDER | wxTAB_TRAVERSAL),
 	  m_pageIndex(index), m_title(title)
 {
 	m_pQueue = parent;
@@ -738,6 +779,8 @@ CQueueViewBase::CQueueViewBase(CQueue* parent, int index, const wxString& title)
 	pImageList->Add(wxArtProvider::GetBitmap(_T("ART_FOLDER"),  wxART_OTHER, wxSize(16, 16)));
 
 	AssignImageList(pImageList, wxIMAGE_LIST_SMALL);
+
+	m_filecount_delay_timer.SetOwner(this);
 }
 
 CQueueViewBase::~CQueueViewBase()
@@ -789,7 +832,7 @@ void CQueueViewBase::OnEraseBackground(wxEraseEvent& event)
 }
 
 // Defined in LocalListView.cpp
-extern wxString FormatSize(const wxLongLong& size);
+extern wxString FormatSize(const wxLongLong& size, bool add_bytes_suffix = false);
 
 wxString CQueueViewBase::OnGetItemText(long item, long column) const
 {
@@ -978,10 +1021,15 @@ void CQueueViewBase::UpdateSelections_ItemAdded(int added)
 	// selections. Though worst case is still O(n), as with every algorithm to
 	// move selections.
 
+#ifndef __WXMSW__
+	// GetNextItem is O(n) if nothing is selected, GetSelectedItemCount() is O(1)
+	const int selection_count = GetSelectedItemCount();
+	if (!selection_count)
+		return;
+#endif
+
 	// Go through all items, keep record of the previous selected item
-	int item = GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
-	while (item != -1 && item < added)
-		item = GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
+	int item = GetNextItem(added - 1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
 
 	int prevItem = -1;
 	while (item != -1)
@@ -1017,16 +1065,21 @@ void CQueueViewBase::UpdateSelections_ItemAdded(int added)
 
 void CQueueViewBase::UpdateSelections_ItemRangeAdded(int added, int count)
 {
+#ifndef __WXMSW__
+	// GetNextItem is O(n) if nothing is selected, GetSelectedItemCount() is O(1)
+	const int selection_count = GetSelectedItemCount();
+	if (!selection_count)
+		return;
+#endif
+
 	std::list<int> itemsToSelect;
 
 	// Go through all selected items
-	int item = GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
-	while (item != -1 && item < added)
-		item = GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
+	int item = GetNextItem(added - 1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
 
 	while (item != -1)
 	{
-		// Select new items preceeding to current one
+		// Select new items preceding to current one
 		while (!itemsToSelect.empty() && itemsToSelect.front() < item)
 		{
 			SetItemState(itemsToSelect.front(), wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);
@@ -1047,13 +1100,18 @@ void CQueueViewBase::UpdateSelections_ItemRangeAdded(int added, int count)
 
 void CQueueViewBase::UpdateSelections_ItemRemoved(int removed)
 {
+#ifndef __WXMSW__
+	// GetNextItem is O(n) if nothing is selected, GetSelectedItemCount() is O(1)
+	const int selection_count = GetSelectedItemCount();
+	if (!selection_count)
+		return;
+#endif
+
 	SetItemState(removed, 0, wxLIST_STATE_SELECTED);
 
-	int prevItem = -1;
-	int item = GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
-	while (item != -1 && item < removed)
-		item = GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
+	int item = GetNextItem(removed - 1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
 
+	int prevItem = -1;
 	while (item != -1)
 	{
 		if (prevItem != -1)
@@ -1084,17 +1142,22 @@ void CQueueViewBase::UpdateSelections_ItemRemoved(int removed)
 
 void CQueueViewBase::UpdateSelections_ItemRangeRemoved(int removed, int count)
 {
+#ifndef __WXMSW__
+	// GetNextItem is O(n) if nothing is selected, GetSelectedItemCount() is O(1)
+	const int selection_count = GetSelectedItemCount();
+	if (!selection_count)
+		return;
+#endif
+
 	SetItemState(removed, 0, wxLIST_STATE_SELECTED);
 
 	std::list<int> itemsToUnselect;
 
-	int item = GetNextItem(-1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
-	while (item != -1 && item < removed)
-		item = GetNextItem(item, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
+	int item = GetNextItem(removed - 1, wxLIST_NEXT_ALL, wxLIST_STATE_SELECTED);
 
 	while (item != -1)
 	{
-		// Unselect new items preceeding to current one
+		// Unselect new items preceding to current one
 		while (!itemsToUnselect.empty() && itemsToUnselect.front() < item - count)
 		{
 			SetItemState(itemsToUnselect.front(), 0, wxLIST_STATE_SELECTED);
@@ -1118,24 +1181,17 @@ void CQueueViewBase::UpdateSelections_ItemRangeRemoved(int removed, int count)
 
 void CQueueViewBase::CreateColumns(const wxString& lastColumnName)
 {
-	static unsigned long widths[6] = { 180, 60, 180, 80, 60, 150 };
-	static bool widths_loaded = false;
+	const unsigned long widths[6] = { 180, 60, 180, 80, 60, 150 };
 
-	if (!widths_loaded)
-	{
-		widths_loaded = true;
-
-		if (!wxGetKeyState(WXK_SHIFT) || !wxGetKeyState(WXK_ALT) || !wxGetKeyState(WXK_CONTROL))
-			COptions::Get()->ReadColumnWidths(OPTION_QUEUE_COLUMN_WIDTHS, 6, widths);
-	}
-
-	InsertColumn(0, _("Server / Local file"), wxLIST_FORMAT_LEFT, widths[0]);
-	InsertColumn(1, _("Direction"), wxLIST_FORMAT_CENTER, widths[1]);
-	InsertColumn(2, _("Remote file"), wxLIST_FORMAT_LEFT, widths[2]);
-	InsertColumn(3, _("Size"), wxLIST_FORMAT_RIGHT, widths[3]);
-	InsertColumn(4, _("Priority"), wxLIST_FORMAT_LEFT, widths[4]);
+	AddColumn(_("Server/Local file"), wxLIST_FORMAT_LEFT, widths[0]);
+	AddColumn(_("Direction"), wxLIST_FORMAT_CENTER, widths[1]);
+	AddColumn(_("Remote file"), wxLIST_FORMAT_LEFT, widths[2]);
+	AddColumn(_("Size"), wxLIST_FORMAT_RIGHT, widths[3]);
+	AddColumn(_("Priority"), wxLIST_FORMAT_LEFT, widths[4]);
 	if (lastColumnName != _T(""))
-		InsertColumn(5, lastColumnName, wxLIST_FORMAT_LEFT, widths[5]);
+		AddColumn(lastColumnName, wxLIST_FORMAT_LEFT, widths[5]);
+
+	LoadColumnSettings(OPTION_QUEUE_COLUMN_WIDTHS, -1, -1);
 }
 
 CServerItem* CQueueViewBase::GetServerItem(const CServer& server)
@@ -1170,7 +1226,7 @@ CServerItem* CQueueViewBase::CreateServerItem(const CServer& server)
 
 void CQueueViewBase::CommitChanges()
 {
-	SetItemCount(m_itemCount);
+	SaveSetItemCount(m_itemCount);
 
 	if (m_insertionStart != -1)
 	{
@@ -1190,6 +1246,12 @@ void CQueueViewBase::CommitChanges()
 
 void CQueueViewBase::DisplayNumberQueuedFiles()
 {
+	if (m_filecount_delay_timer.IsRunning())
+	{
+		m_fileCountChanged = true;
+		return;
+	}
+
 	wxString str;
 	if (m_fileCount > 0)
 	{
@@ -1209,15 +1271,19 @@ void CQueueViewBase::DisplayNumberQueuedFiles()
 
 	m_fileCountChanged = false;
 	m_folderScanCountChanged = false;
+
+	m_filecount_delay_timer.Start(200, true);
 }
 
 void CQueueViewBase::InsertItem(CServerItem* pServerItem, CQueueItem* pItem)
 {
+	const int newIndex = GetItemIndex(pServerItem) + pServerItem->GetChildrenCount(true) + 1;
+
 	pServerItem->AddChild(pItem);
 	m_itemCount++;
 
 	if (m_insertionStart == -1)
-		m_insertionStart = GetItemIndex(pItem);
+		m_insertionStart = newIndex;
 	m_insertionCount++;
 
 	if (pItem->GetType() == QueueItemType_File || pItem->GetType() == QueueItemType_Folder)
@@ -1245,13 +1311,13 @@ bool CQueueViewBase::RemoveItem(CQueueItem* pItem, bool destroy, bool updateItem
 		wxASSERT(m_folderScanCount > 0);
 		m_folderScanCount--;
 		m_folderScanCountChanged = true;
-		
+
 	}
 
 	int index = 0;
 	if (updateSelections)
 		index = GetItemIndex(pItem);
-	
+
 	CQueueItem* topLevelItem = pItem->GetTopLevelItem();
 
 	int count = topLevelItem->GetChildrenCount(true);
@@ -1277,7 +1343,7 @@ bool CQueueViewBase::RemoveItem(CQueueItem* pItem, bool destroy, bool updateItem
 
 		m_itemCount -= count + 1;
 		if (updateItemCount)
-			SetItemCount(m_itemCount);
+			SaveSetItemCount(m_itemCount);
 
 		didRemoveParent = true;
 	}
@@ -1290,7 +1356,7 @@ bool CQueueViewBase::RemoveItem(CQueueItem* pItem, bool destroy, bool updateItem
 
 		m_itemCount -= count;
 		if (updateItemCount)
-			SetItemCount(m_itemCount);
+			SaveSetItemCount(m_itemCount);
 
 		didRemoveParent = false;
 	}
@@ -1302,7 +1368,7 @@ bool CQueueViewBase::RemoveItem(CQueueItem* pItem, bool destroy, bool updateItem
 		if (oldCount > m_itemCount)
 		{
 			bool eraseBackground = GetTopItem() + GetCountPerPage() + 1 >= m_itemCount;
-			Refresh(eraseBackground);
+			RefreshListOnly(eraseBackground);
 			if (eraseBackground)
 				Update();
 		}
@@ -1316,7 +1382,13 @@ void CQueueViewBase::RefreshItem(const CQueueItem* pItem)
 	wxASSERT(pItem);
 	int index = GetItemIndex(pItem);
 
+#ifdef __WXMSW__
+	wxRect rect;
+	GetItemRect(index, rect);
+	RefreshRect(rect, false);
+#else
 	wxListCtrl::RefreshItem(index);
+#endif
 }
 
 void CQueueViewBase::OnNavigationKey(wxNavigationKeyEvent& event)
@@ -1334,10 +1406,16 @@ void CQueueViewBase::OnChar(wxKeyEvent& event)
 		return;
 	}
 
+	bool forward;
+	if (GetLayoutDirection() != wxLayout_RightToLeft)
+		forward = code == WXK_RIGHT;
+	else
+		forward = code == WXK_LEFT;
+
 	int selection = m_pQueue->GetSelection();
-	if (selection > 0 && code == WXK_LEFT)
+	if (selection > 0 && !forward)
 		selection--;
-	else if (selection < (int)m_pQueue->GetPageCount() - 1 && code == WXK_RIGHT)
+	else if (selection < (int)m_pQueue->GetPageCount() - 1 && forward)
 		selection++;
 	else
 		return;
@@ -1349,15 +1427,42 @@ void CQueueViewBase::OnEndColumnDrag(wxListEvent& event)
 {
 	for (unsigned int i = 0; i < m_pQueue->GetPageCount(); i++)
 	{
-		CQueueViewBase* page = (CQueueViewBase*)m_pQueue->GetPage(i);
-		if (!page || page == this)
+		wxWindow *page = m_pQueue->GetPage(i);
+
+		wxListCtrl* queue_page = wxDynamicCast(page, wxListCtrl);
+		if (!queue_page || queue_page == this)
 			continue;
 
-		for (int col = 0; col < wxMin(GetColumnCount(), page->GetColumnCount()); col++)
-			page->SetColumnWidth(col, GetColumnWidth(col));
+		for (int col = 0; col < wxMin(GetColumnCount(), queue_page->GetColumnCount()); col++)
+			queue_page->SetColumnWidth(col, GetColumnWidth(col));
 	}
 }
 
+void CQueueViewBase::OnTimer(wxTimerEvent& event)
+{
+	if (event.GetId() != m_filecount_delay_timer.GetId())
+	{
+		event.Skip();
+		return;
+	}
+
+	if (m_fileCountChanged || m_folderScanCountChanged)
+		DisplayNumberQueuedFiles();
+}
+
+void CQueueViewBase::OnKeyDown(wxKeyEvent& event)
+{
+	const int code = event.GetKeyCode();
+	const int mods = event.GetModifiers();
+	if (code == 'A' && (mods == wxMOD_CMD || mods == (wxMOD_CONTROL | wxMOD_META)))
+	{
+		for (unsigned int i = 0; i < (unsigned int)GetItemCount(); i++)
+			SetItemState(i, wxLIST_STATE_SELECTED, wxLIST_STATE_SELECTED);
+	}
+	else
+		event.Skip();
+}
+
 // ------
 // CQueue
 // ------
