@@ -2,6 +2,8 @@
 #define __QUEUE_H__
 
 #include "aui_notebook_ex.h"
+#include "listctrlex.h"
+#include "edithandler.h"
 
 #define PRIORITY_COUNT 5
 enum QueuePriority
@@ -59,7 +61,6 @@ protected:
 
 	CQueueItem* m_parent;
 
-	std::vector<CQueueItem*> m_children;
 	int m_visibleOffspring; // Visible offspring over all sublevels
 	wxString m_indent;
 
@@ -73,6 +74,14 @@ protected:
 	std::vector<t_cacheItem> m_lookupCache;
 
 	friend class CServerItem;
+
+private:
+	std::vector<CQueueItem*> m_children;
+
+	// Number of items removed at front of list
+	// Increased instead of calling slow m_children.erase(0),
+	// resetted on insert.
+	int m_removed_at_front;
 };
 
 class CFileItem;
@@ -97,7 +106,7 @@ public:
 
 	virtual void SaveItem(TiXmlElement* pElement) const;
 
-	void SetDefaultFileExistsAction(int action, const enum TransferDirection direction);
+	void SetDefaultFileExistsAction(enum CFileExistsNotification::OverwriteAction action, const enum TransferDirection direction);
 
 	int m_activeCount;
 
@@ -154,10 +163,13 @@ public:
 	virtual bool TryRemoveAll(); // Removes a inactive childrens, queues active children for removal.
 								 // Returns true if item can be removed itself
 
+	void SetLocalFile(const wxString &file);
+	void SetRemoteFile(const wxString &file);
+
 	bool m_queued;
 	int m_errorCount;
 	bool m_remove;
-	bool m_edit;
+	CEditHandler::fileType m_edit;
 
 	wxString m_statusMessage;
 
@@ -165,7 +177,11 @@ public:
 
 	t_EngineData* m_pEngineData;
 
-	int m_defaultFileExistsAction;
+	enum CFileExistsNotification::OverwriteAction m_defaultFileExistsAction;
+
+	bool m_madeProgress;
+
+	enum CFileExistsNotification::OverwriteAction m_onetime_action;
 
 protected:
 	enum QueuePriority m_priority;
@@ -213,14 +229,11 @@ public:
 
 	int m_count;
 
-	struct t_dirPair
-	{
-		wxString localPath;
-		CServerPath remotePath;
-	};
-	std::list<t_dirPair> m_dirsToCheck;
+	enum CFileExistsNotification::OverwriteAction m_defaultFileExistsAction;
 
-	int m_defaultFileExistsAction;
+	bool m_dir_is_empty;
+	wxString m_current_local_path;
+	CServerPath m_current_remote_path;
 
 protected:
 	wxString m_localPath;
@@ -238,7 +251,7 @@ public:
 };
 
 class CQueue;
-class CQueueViewBase : public wxListCtrl
+class CQueueViewBase : public wxListCtrlEx
 {
 public:
 	CQueueViewBase(CQueue* parent, int index, const wxString& title);
@@ -302,11 +315,15 @@ protected:
 
 	const wxString m_title;
 
+	wxTimer m_filecount_delay_timer;
+
 	DECLARE_EVENT_TABLE();
 	void OnEraseBackground(wxEraseEvent& event);
 	void OnNavigationKey(wxNavigationKeyEvent& event);
 	void OnChar(wxKeyEvent& event);
 	void OnEndColumnDrag(wxListEvent& event);
+	void OnTimer(wxTimerEvent& event);
+	void OnKeyDown(wxKeyEvent& event);
 };
 
 class CQueueView;
