@@ -4,6 +4,7 @@
 #include "commandqueue.h"
 #include "state.h"
 #include "Options.h"
+#include "loginmanager.h"
 
 BEGIN_EVENT_TABLE(CQuickconnectBar, wxPanel)
 EVT_BUTTON(XRCID("ID_QUICKCONNECT_OK"), CQuickconnectBar::OnQuickconnect)
@@ -15,6 +16,10 @@ END_EVENT_TABLE();
 
 CQuickconnectBar::CQuickconnectBar()
 {
+	m_pHost = 0;
+	m_pUser = 0;
+	m_pPass = 0;
+	m_pPort = 0;
 }
 
 CQuickconnectBar::~CQuickconnectBar()
@@ -38,7 +43,18 @@ bool CQuickconnectBar::Create(wxWindow* pParent, CState* pState)
 		tlw->SetDefaultItem(0);
 #endif
 
-	XRCCTRL(*this, "ID_QUICKCONNECT_PORT", wxTextCtrl)->SetMaxLength(5);
+	m_pHost = XRCCTRL(*this, "ID_QUICKCONNECT_HOST", wxTextCtrl);
+	m_pUser = XRCCTRL(*this, "ID_QUICKCONNECT_USER", wxTextCtrl);
+	m_pPass = XRCCTRL(*this, "ID_QUICKCONNECT_PASS", wxTextCtrl);
+	m_pPort = XRCCTRL(*this, "ID_QUICKCONNECT_PORT", wxTextCtrl);
+
+	if (!m_pHost || !m_pUser || !m_pPass || !m_pPort)
+	{
+		wxLogError(_("Cannot load Quickconnect bar from resource file"));
+		return false;
+	}
+
+	m_pPort->SetMaxLength(5);
 
 	return true;
 }
@@ -51,10 +67,10 @@ void CQuickconnectBar::OnQuickconnect(wxCommandEvent& event)
 		return;
 	}
 
-	wxString host = XRCCTRL(*this, "ID_QUICKCONNECT_HOST", wxTextCtrl)->GetValue();
-	wxString user = XRCCTRL(*this, "ID_QUICKCONNECT_USER", wxTextCtrl)->GetValue();
-	wxString pass = XRCCTRL(*this, "ID_QUICKCONNECT_PASS", wxTextCtrl)->GetValue();
-	wxString port = XRCCTRL(*this, "ID_QUICKCONNECT_PORT", wxTextCtrl)->GetValue();
+	wxString host = m_pHost->GetValue();
+	wxString user = m_pUser->GetValue();
+	wxString pass = m_pPass->GetValue();
+	wxString port = m_pPort->GetValue();
 	
 	long numericPort = 0;
 	if (port != _T(""))
@@ -74,7 +90,7 @@ void CQuickconnectBar::OnQuickconnect(wxCommandEvent& event)
 		return;
 	}
 
-	host = server.GetHost();
+	host = server.FormatHost(true);
 	ServerProtocol protocol = server.GetProtocol();
 	switch (protocol)
 	{
@@ -93,22 +109,26 @@ void CQuickconnectBar::OnQuickconnect(wxCommandEvent& event)
 		break;
 	}
 	
-	XRCCTRL(*this, "ID_QUICKCONNECT_HOST", wxTextCtrl)->SetValue(host);
+	m_pHost->SetValue(host);
 	if (server.GetPort() != server.GetDefaultPort(server.GetProtocol()))
-		XRCCTRL(*this, "ID_QUICKCONNECT_PORT", wxTextCtrl)->SetValue(wxString::Format(_T("%d"), server.GetPort()));
+		m_pPort->SetValue(wxString::Format(_T("%d"), server.GetPort()));
 	else
-		XRCCTRL(*this, "ID_QUICKCONNECT_PORT", wxTextCtrl)->SetValue(_T(""));
+		m_pPort->SetValue(_T(""));
 
-	XRCCTRL(*this, "ID_QUICKCONNECT_USER", wxTextCtrl)->SetValue(server.GetUser());
-	XRCCTRL(*this, "ID_QUICKCONNECT_PASS", wxTextCtrl)->SetValue(server.GetPass());
+	m_pUser->SetValue(server.GetUser());
+	m_pPass->SetValue(server.GetPass());
 
 	if (protocol == HTTP)
 	{
-		wxMessageBox(_("FileZilla does not support the HTTP protocol"), _("Syntax error"), wxICON_EXCLAMATION);
+		wxString error = _("Invalid protocol specified. Valid protocols are:\nftp:// for normal FTP,\nsftp:// for SSH file transfer protocol,\nftps:// for FTP over SSL (implicit) and\nftpes:// for FTP over SSL (explicit).");
+		wxMessageBox(error, _("Syntax error"), wxICON_EXCLAMATION);
 		return;
 	}
 
-	if (!m_pState->Connect(server, true))
+	if (event.GetId() == 1)
+		server.SetBypassProxy(true);
+
+	if (!m_pState->Connect(server, true, path))
 		return;
 
 	CRecentServerList::SetMostRecentServer(server);
@@ -119,8 +139,10 @@ void CQuickconnectBar::OnQuickconnectDropdown(wxCommandEvent& event)
 	wxMenu* pMenu = new wxMenu;
 
 	// We have to start with id 1 since menu items with id 0 don't work under OS X
-	pMenu->Append(1, _("Clear quickconnect bar"));
-	pMenu->Append(2, _("Clear history"));
+	if (COptions::Get()->GetOptionVal(OPTION_FTP_PROXY_TYPE))
+		pMenu->Append(1, _("Connect bypassing proxy settings"));
+	pMenu->Append(2, _("Clear quickconnect bar"));
+	pMenu->Append(3, _("Clear history"));
 	pMenu->AppendSeparator();
 
 	m_recentServers = CRecentServerList::GetMostRecentServers();
@@ -141,8 +163,10 @@ void CQuickconnectBar::OnMenu(wxCommandEvent& event)
 {
 	const int id = event.GetId();
 	if (id == 1)
-		ClearFields();
+		OnQuickconnect(event);
 	else if (id == 2)
+		ClearFields();
+	else if (id == 3)
 		CRecentServerList::Clear();
 
 	if (id < 10)
@@ -155,15 +179,22 @@ void CQuickconnectBar::OnMenu(wxCommandEvent& event)
 	std::list<CServer>::const_iterator iter;
 	for (iter = m_recentServers.begin(); index; index--, iter++);
 
-	m_pState->Connect(*iter, true);
+	CServer server = *iter;
+	if (server.GetLogonType() == ASK)
+	{
+		if (!CLoginManager::Get().GetPassword(server, false))
+			return;
+	}
+
+	m_pState->Connect(server, true);
 }
 
 void CQuickconnectBar::ClearFields()
 {
-	XRCCTRL(*this, "ID_QUICKCONNECT_HOST", wxTextCtrl)->SetValue(_T(""));
-	XRCCTRL(*this, "ID_QUICKCONNECT_PORT", wxTextCtrl)->SetValue(_T(""));
-	XRCCTRL(*this, "ID_QUICKCONNECT_USER", wxTextCtrl)->SetValue(_T(""));
-	XRCCTRL(*this, "ID_QUICKCONNECT_PASS", wxTextCtrl)->SetValue(_T(""));
+	m_pHost->SetValue(_T(""));
+	m_pPort->SetValue(_T(""));
+	m_pUser->SetValue(_T(""));
+	m_pPass->SetValue(_T(""));
 }
 
 void CQuickconnectBar::OnKeyboardNavigation(wxNavigationKeyEvent& event)
@@ -173,7 +204,7 @@ void CQuickconnectBar::OnKeyboardNavigation(wxNavigationKeyEvent& event)
 		event.SetEventObject(this);
 		GetParent()->ProcessEvent(event);
 	}
-	else if (!event.GetDirection() && event.GetEventObject() == XRCCTRL(*this, "ID_QUICKCONNECT_HOST", wxTextCtrl))
+	else if (!event.GetDirection() && event.GetEventObject() == m_pHost)
 	{
 		event.SetEventObject(this);
 		GetParent()->ProcessEvent(event);
