@@ -44,6 +44,19 @@ static Config cfg;
  * Higher-level helper functions used in commands.
  */
 
+void compare_req_rreq(struct sftp_request *req, struct sftp_request *rreq)
+{
+    if (!rreq) {
+	fzprintf(sftpError, "%s", fxp_error());
+	cleanup_exit(1);
+    }
+
+    if (rreq != req) {
+	fzprintf(sftpError, "ID of request and reply do not match");
+	cleanup_exit(1);
+    }
+}
+
 /*
  * Attempt to canonify a pathname starting from the pwd. If
  * canonification fails, at least fall back to returning a _valid_
@@ -68,7 +81,7 @@ char *canonify(char *name)
 
     sftp_register(req = fxp_realpath_send(fullname));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     canonname = fxp_realpath_recv(pktin, rreq);
 
     if (canonname) {
@@ -129,7 +142,7 @@ char *canonify(char *name)
 	    sftp_register(req = fxp_realpath_send(fullname));
 	}
 	rreq = sftp_find_request(pktin = sftp_recv());
-	assert(rreq == req);
+	compare_req_rreq(req, rreq);
 	canonname = fxp_realpath_recv(pktin, rreq);
 
 	if (!canonname) {
@@ -214,6 +227,8 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
     uint64 offset;
     WFile *file;
     int ret, shown_err = FALSE;
+    _fztimer timer;
+    int winterval;
 
     /*
      * In recursive mode, see if we're dealing with a directory.
@@ -226,7 +241,7 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
 
 	sftp_register(req = fxp_stat_send(fname));
 	rreq = sftp_find_request(pktin = sftp_recv());
-	assert(rreq == req);
+	compare_req_rreq(req, rreq);
 	result = fxp_stat_recv(pktin, rreq, &attrs);
 
 	if (result &&
@@ -255,7 +270,7 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
 	     */
 	    sftp_register(req = fxp_opendir_send(fname));
 	    rreq = sftp_find_request(pktin = sftp_recv());
-	    assert(rreq == req);
+	    compare_req_rreq(req, rreq);
 	    dirhandle = fxp_opendir_recv(pktin, rreq);
 
 	    if (!dirhandle) {
@@ -270,7 +285,7 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
 
 		sftp_register(req = fxp_readdir_send(dirhandle));
 		rreq = sftp_find_request(pktin = sftp_recv());
-		assert(rreq == req);
+		compare_req_rreq(req, rreq);
 		names = fxp_readdir_recv(pktin, rreq);
 
 		if (names == NULL) {
@@ -304,7 +319,7 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
 	    }
 	    sftp_register(req = fxp_close_send(dirhandle));
 	    rreq = sftp_find_request(pktin = sftp_recv());
-	    assert(rreq == req);
+	    compare_req_rreq(req, rreq);
 	    fxp_close_recv(pktin, rreq);
 
 	    /*
@@ -387,7 +402,7 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
 
     sftp_register(req = fxp_open_send(fname, SSH_FXF_READ));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     fh = fxp_open_recv(pktin, rreq);
 
     if (!fh) {
@@ -406,7 +421,7 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
 
 	sftp_register(req = fxp_close_send(fh));
 	rreq = sftp_find_request(pktin = sftp_recv());
-	assert(rreq == req);
+	compare_req_rreq(req, rreq);
 	fxp_close_recv(pktin, rreq);
 
 	return 0;
@@ -415,14 +430,15 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
     if (restart) {
 	char decbuf[30];
 	if (seek_file(file, uint64_make(0,0) , FROM_END) == -1) {
+	    close_wfile(file);
 	    fzprintf(sftpError, "reget: cannot restart %s - file too large",
 		   outfname);
-	    	sftp_register(req = fxp_close_send(fh));
-		rreq = sftp_find_request(pktin = sftp_recv());
-		assert(rreq == req);
-		fxp_close_recv(pktin, rreq);
-		
-		return 0;
+	    sftp_register(req = fxp_close_send(fh));
+	    rreq = sftp_find_request(pktin = sftp_recv());
+	    compare_req_rreq(req, rreq);
+	    fxp_close_recv(pktin, rreq);
+
+	    return 0;
 	}
 	    
 	offset = get_file_posn(file);
@@ -434,6 +450,9 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
 
     fzprintf(sftpStatus, "remote:%s => local:%s", fname, outfname);
 
+    fz_timer_init(&timer);
+    winterval = 0;
+
     /*
      * FIXME: we can use FXP_FSTAT here to get the file size, and
      * thus put up a progress bar.
@@ -454,43 +473,50 @@ int sftp_get_file(char *fname, char *outfname, int recurse, int restart)
 		fzprintf(sftpError, "error while reading: %s", fxp_error());
 		shown_err = TRUE;
 	    }
-            ret = 0;
 	}
 
 	while (xfer_download_data(xfer, &vbuf, &len)) {
 	    unsigned char *buf = (unsigned char *)vbuf;
 
 	    wpos = 0;
-	    while (wpos < len) {
+	    while (file && wpos < len) {
 		wlen = write_to_file(file, buf + wpos, len - wpos);
 		if (wlen <= 0) {
-		    fzprintf(sftpError, "error while writing local file");
-		    ret = 0;
-		    xfer_set_error(xfer);
+		    fzprintf(shown_err ? sftpStatus : sftpError, "error while writing local file");
+		    shown_err = TRUE;
+		    close_wfile(file);
+		    file = 0;
 		    break;
 		}
-		else
-		    fzprintf(sftpWrite, "%d", wlen);
 		wpos += wlen;
 	    }
 	    if (wpos < len) {	       /* we had an error */
-		ret = 0;
 		xfer_set_error(xfer);
 	    }
-
+	    winterval += wpos;
 	    sfree(vbuf);
 	}
+
+	if (fz_timer_check(&timer)) {
+	    fzprintf(sftpWrite, "%d", winterval);
+	    winterval = 0;
+	}
+
     }
 
     xfer_cleanup(xfer);
 
-    close_wfile(file);
+    if (file)
+	close_wfile(file);
 
     sftp_register(req = fxp_close_send(fh));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     fxp_close_recv(pktin, rreq);
 
+    if (shown_err)
+	ret = 0;
+
     return ret;
 }
 
@@ -523,14 +549,14 @@ int sftp_put_file(char *fname, char *outfname, int recurse, int restart)
 	 */
 	sftp_register(req = fxp_stat_send(outfname));
 	rreq = sftp_find_request(pktin = sftp_recv());
-	assert(rreq == req);
+	compare_req_rreq(req, rreq);
 	result = fxp_stat_recv(pktin, rreq, &attrs);
 	if (!result ||
 	    !(attrs.flags & SSH_FILEXFER_ATTR_PERMISSIONS) ||
 	    !(attrs.permissions & 0040000)) {
 	    sftp_register(req = fxp_mkdir_send(outfname));
 	    rreq = sftp_find_request(pktin = sftp_recv());
-	    assert(rreq == req);
+	    compare_req_rreq(req, rreq);
 	    result = fxp_mkdir_recv(pktin, rreq);
 
 	    if (!result) {
@@ -583,7 +609,7 @@ int sftp_put_file(char *fname, char *outfname, int recurse, int restart)
                 nextoutfname = dupcat(outfname, "/", ournames[i], NULL);
                 sftp_register(req = fxp_stat_send(nextoutfname));
                 rreq = sftp_find_request(pktin = sftp_recv());
-                assert(rreq == req);
+                compare_req_rreq(req, rreq);
                 result = fxp_stat_recv(pktin, rreq, &attrs);
                 sfree(nextoutfname);
                 if (!result)
@@ -645,10 +671,11 @@ int sftp_put_file(char *fname, char *outfname, int recurse, int restart)
 					  SSH_FXF_CREAT | SSH_FXF_TRUNC));
     }
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     fh = fxp_open_recv(pktin, rreq);
 
     if (!fh) {
+	close_rfile(file);
 	fzprintf(sftpError, "%s: open for write: %s", outfname, fxp_error());
 	return 0;
     }
@@ -660,14 +687,16 @@ int sftp_put_file(char *fname, char *outfname, int recurse, int restart)
 
 	sftp_register(req = fxp_fstat_send(fh));
 	rreq = sftp_find_request(pktin = sftp_recv());
-	assert(rreq == req);
+	compare_req_rreq(req, rreq);
 	ret = fxp_fstat_recv(pktin, rreq, &attrs);
 
 	if (!ret) {
+	    close_rfile(file);
 	    fzprintf(sftpError, "read size of %s: %s", outfname, fxp_error());
 	    return 0;
 	}
 	if (!(attrs.flags & SSH_FILEXFER_ATTR_SIZE)) {
+	    close_rfile(file);
 	    fzprintf(sftpError, "read size of %s: size was not given", outfname);
 	    return 0;
 	}
@@ -708,18 +737,21 @@ int sftp_put_file(char *fname, char *outfname, int recurse, int restart)
 	if (!xfer_done(xfer)) {
 	    pktin = sftp_recv();
 	    ret = xfer_upload_gotpkt(xfer, pktin);
-	    if (!ret) {
+	    if (ret <= 0 && !err) {
 		fzprintf(sftpError, "error while writing: %s\n", fxp_error());
 		err = 1;
 	    }
 	}
     }
 
+    if (err)
+	ret = 0;
+
     xfer_cleanup(xfer);
 
     sftp_register(req = fxp_close_send(fh));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     fxp_close_recv(pktin, rreq);
 
     close_rfile(file);
@@ -775,7 +807,7 @@ SftpWildcardMatcher *sftp_begin_wildcard_matching(char *name)
 
     sftp_register(req = fxp_opendir_send(cdir));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     dirh = fxp_opendir_recv(pktin, rreq);
 
     if (dirh) {
@@ -810,7 +842,7 @@ char *sftp_wildcard_get_filename(SftpWildcardMatcher *swcm)
 	if (!swcm->names) {
 	    sftp_register(req = fxp_readdir_send(swcm->dirh));
 	    rreq = sftp_find_request(pktin = sftp_recv());
-	    assert(rreq == req);
+	    compare_req_rreq(req, rreq);
 	    swcm->names = fxp_readdir_recv(pktin, rreq);
 
 	    if (!swcm->names) {
@@ -857,7 +889,7 @@ void sftp_finish_wildcard_matching(SftpWildcardMatcher *swcm)
 
     sftp_register(req = fxp_close_send(swcm->dirh));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     fxp_close_recv(pktin, rreq);
 
     if (swcm->names)
@@ -959,6 +991,93 @@ int sftp_cmd_quit(struct sftp_command *cmd)
     return -1;
 }
 
+int sftp_cmd_keyfile(struct sftp_command *cmd)
+{
+    if (cmd->nwords != 2) {
+	fzprintf(sftpError, "No keyfile given");
+	return 0;
+    }
+
+    Keyfile_list *list = snew(Keyfile_list);
+    list->file = filename_from_str(cmd->words[1]);
+    list->next = cfg.keyfile_list;
+    cfg.keyfile_list = list;
+
+    fznotify1(sftpDone, 1);
+    return 1;
+}
+
+int sftp_cmd_proxy(struct sftp_command *cmd)
+{
+    int proxy_type;
+    int portnumber;
+
+    if (cmd->nwords < 2) {
+	fzprintf(sftpError, "Not enough arguments to proxy command");
+	return 0;
+    }
+
+    if (!strcmp(cmd->words[1], "0")) {
+	cfg.proxy_type = PROXY_NONE;
+        fznotify1(sftpDone, 1);
+	return 1;
+    }
+    if (!strcmp(cmd->words[1], "1")) {
+	proxy_type = PROXY_HTTP;
+    }
+    else if (!strcmp(cmd->words[1], "2")) {
+	proxy_type = PROXY_SOCKS5;
+    }
+    else {
+	fzprintf(sftpError, "Unknown proxy type");
+	return 0;
+    }
+
+    if (cmd->nwords < 4) {
+	fzprintf(sftpError, "Not enough arguments to proxy command");
+	return 0;
+    }
+
+    portnumber = atoi(cmd->words[3]);
+    if (portnumber < 0 || portnumber > 65535) {
+	fzprintf(sftpError, "Invalid port");
+	return 0;
+    }
+
+    if (strlen(cmd->words[2]) >= sizeof(cfg.proxy_host)) {
+	fzprintf(sftpError,  "Host too long");
+	return 0;
+    }
+    if (cmd->nwords > 4 && strlen(cmd->words[4]) >= sizeof(cfg.proxy_username)) {
+	fzprintf(sftpError,  "User too long");
+	return 0;
+    }
+    if (cmd->nwords > 5 && strlen(cmd->words[5]) >= sizeof(cfg.proxy_password)) {
+	fzprintf(sftpError,  "Password too long");
+	return 0;
+    }
+
+    if (cmd->nwords > 5) {
+	strcpy(cfg.proxy_username, cmd->words[4]);
+	strcpy(cfg.proxy_password, cmd->words[5]);
+    }
+    else if (cmd->nwords > 4) {
+	strcpy(cfg.proxy_username, cmd->words[4]);
+	cfg.proxy_password[0] = 0;
+    }
+    else {
+	cfg.proxy_username[0] = 0;
+	cfg.proxy_password[0] = 0;
+    }
+
+    cfg.proxy_type = proxy_type;
+    strcpy(cfg.proxy_host, cmd->words[2]);
+    cfg.proxy_port = portnumber;
+
+    fznotify1(sftpDone, 1);
+    return 1;
+}
+
 int sftp_cmd_close(struct sftp_command *cmd)
 {
     if (back == NULL) {
@@ -990,6 +1109,7 @@ int sftp_cmd_ls(struct sftp_command *cmd)
     struct sftp_packet *pktin;
     struct sftp_request *req, *rreq;
     int i;
+    int ret;
 
     if (back == NULL) {
 	not_connected();
@@ -1037,11 +1157,12 @@ int sftp_cmd_ls(struct sftp_command *cmd)
 
     sftp_register(req = fxp_opendir_send(cdir));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     dirh = fxp_opendir_recv(pktin, rreq);
 
     if (dirh == NULL) {
 	fzprintf(sftpError, "Unable to open %s: %s", dir, fxp_error());
+	ret = 1;
     } else {
 	nnames = namesize = 0;
 	ournames = NULL;
@@ -1050,7 +1171,7 @@ int sftp_cmd_ls(struct sftp_command *cmd)
 
 	    sftp_register(req = fxp_readdir_send(dirh));
 	    rreq = sftp_find_request(pktin = sftp_recv());
-	    assert(rreq == req);
+	    compare_req_rreq(req, rreq);
 	    names = fxp_readdir_recv(pktin, rreq);
 
 	    if (names == NULL) {
@@ -1077,7 +1198,7 @@ int sftp_cmd_ls(struct sftp_command *cmd)
 	}
 	sftp_register(req = fxp_close_send(dirh));
 	rreq = sftp_find_request(pktin = sftp_recv());
-	assert(rreq == req);
+	compare_req_rreq(req, rreq);
 	fxp_close_recv(pktin, rreq);
 
 	/*
@@ -1094,12 +1215,14 @@ int sftp_cmd_ls(struct sftp_command *cmd)
 	    fxp_free_name(ournames[i]);
 	}
 	sfree(ournames);
+	ret = 0;
     }
 
     sfree(cdir);
     sfree(unwcdir);
 
-    fznotify1(sftpDone, 1);
+    if (!ret)
+	fznotify1(sftpDone, 1);
 
     return 1;
 }
@@ -1132,7 +1255,7 @@ int sftp_cmd_cd(struct sftp_command *cmd)
 
     sftp_register(req = fxp_opendir_send(dir));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     dirh = fxp_opendir_recv(pktin, rreq);
 
     if (!dirh) {
@@ -1143,7 +1266,7 @@ int sftp_cmd_cd(struct sftp_command *cmd)
 
     sftp_register(req = fxp_close_send(dirh));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     fxp_close_recv(pktin, rreq);
 
     sfree(pwd);
@@ -1188,9 +1311,10 @@ int sftp_general_get(struct sftp_command *cmd, int restart, int multiple)
     }
 
     i = 1;
+    /* FZ unused
     while (i < cmd->nwords && cmd->words[i][0] == '-') {
 	if (!strcmp(cmd->words[i], "--")) {
-	    /* finish processing options */
+	    /* finish processing options *//* FZ unused
 	    i++;
 	    break;
 	} else if (!strcmp(cmd->words[i], "-r")) {
@@ -1201,6 +1325,7 @@ int sftp_general_get(struct sftp_command *cmd, int restart, int multiple)
 	}
 	i++;
     }
+    */
 
     if (i >= cmd->nwords) {
 	fzprintf(sftpError, "%s: expects a filename", cmd->words[0]);
@@ -1426,7 +1551,7 @@ int sftp_cmd_mkdir(struct sftp_command *cmd)
 
 	sftp_register(req = fxp_mkdir_send(dir));
 	rreq = sftp_find_request(pktin = sftp_recv());
-	assert(rreq == req);
+	compare_req_rreq(req, rreq);
 	result = fxp_mkdir_recv(pktin, rreq);
 
 	if (!result) {
@@ -1449,7 +1574,7 @@ static int sftp_action_rmdir(void *vctx, char *dir)
 
     sftp_register(req = fxp_rmdir_send(dir));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     result = fxp_rmdir_recv(pktin, rreq);
 
     if (!result) {
@@ -1491,7 +1616,7 @@ static int sftp_action_rm(void *vctx, char *fname)
 
     sftp_register(req = fxp_remove_send(fname));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     result = fxp_remove_recv(pktin, rreq);
 
     if (!result) {
@@ -1539,7 +1664,7 @@ static int check_is_dir(char *dstfname)
 
     sftp_register(req = fxp_stat_send(dstfname));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     result = fxp_stat_recv(pktin, rreq, &attrs);
 
     if (result &&
@@ -1586,7 +1711,7 @@ static int sftp_action_mv(void *vctx, char *srcfname)
 
     sftp_register(req = fxp_rename_send(srcfname, finalfname));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     result = fxp_rename_recv(pktin, rreq);
 
     error = result ? NULL : fxp_error();
@@ -1666,7 +1791,7 @@ static int sftp_action_chmod(void *vctx, char *fname)
 
     sftp_register(req = fxp_stat_send(fname));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     result = fxp_stat_recv(pktin, rreq, &attrs);
 
     if (!result || !(attrs.flags & SSH_FILEXFER_ATTR_PERMISSIONS)) {
@@ -1686,7 +1811,7 @@ static int sftp_action_chmod(void *vctx, char *fname)
 
     sftp_register(req = fxp_setstat_send(fname, attrs));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     result = fxp_setstat_recv(pktin, rreq);
 
     if (!result) {
@@ -1830,6 +1955,160 @@ int sftp_cmd_chmod(struct sftp_command *cmd)
     return ret;
 }
 
+static int sftp_action_chmtime(void *vmtime, char *fname)
+{
+    struct fxp_attrs attrs = {0};
+    struct sftp_packet *pktin;
+    struct sftp_request *req, *rreq;
+    int result;
+    uint64 *mtime = (uint64*)vmtime;
+
+    attrs.flags = SSH_FILEXFER_ATTR_ACCESSTIME | SSH_FILEXFER_ATTR_CREATETIME | SSH_FILEXFER_ATTR_MODIFICATIONTIME;
+    sftp_register(req = fxp_stat_send(fname));
+    rreq = sftp_find_request(pktin = sftp_recv());
+    compare_req_rreq(req, rreq);
+    result = fxp_stat_recv(pktin, rreq, &attrs);
+
+    if (!result || !(attrs.flags & (SSH_FILEXFER_ATTR_ACCESSTIME | SSH_FILEXFER_ATTR_CREATETIME | SSH_FILEXFER_ATTR_MODIFICATIONTIME))) {
+	fzprintf(sftpError, "get attrs for %s: %s", fname,
+	       result ? "times not provided" : fxp_error());
+	return 0;
+    }
+
+/*FZ: Does not work
+    if (attrs.flags & SSH_FILEXFER_ATTR_MODIFICATIONTIME)
+	attrs.flags = SSH_FILEXFER_ATTR_MODIFICATIONTIME;
+    else
+*/
+	attrs.flags = SSH_FILEXFER_ATTR_ACMODTIME;
+    
+    if (!uint64_compare(attrs.mtime, *mtime)) {
+	fzprintf(sftpVerbose, "Nothing");
+	return 1;		       /* no need to do anything! */
+    }
+    attrs.mtime = *mtime;
+
+    sftp_register(req = fxp_setstat_send(fname, attrs));
+    rreq = sftp_find_request(pktin = sftp_recv());
+    compare_req_rreq(req, rreq);
+    result = fxp_setstat_recv(pktin, rreq);
+
+    if (!result) {
+	fzprintf(sftpError, "set attrs for %s: %s", fname, fxp_error());
+	return 0;
+    }
+
+    return 1;
+}
+
+static int sftp_cmd_chmtime(struct sftp_command *cmd)
+{
+    char *p;
+    int i, ret;
+    uint64 mtime;
+
+    if (back == NULL) {
+	not_connected();
+	return 0;
+    }
+
+    if (cmd->nwords < 3) {
+	fzprintf(sftpError, "chmtime: expects a the time and a filename");
+	return 0;
+    }
+
+    // Make sure mtime is valid
+    p = cmd->words[1];
+    while (*p) {
+	char c = *p++;
+	if (c < '0' || c > '9') {
+	    fzprintf(sftpError, "chmtime: not a valid time");
+	    return 0;
+	}
+    }
+    mtime = uint64_from_decimal(cmd->words[1]);
+
+    ret = 1;
+    for (i = 2; i < cmd->nwords; i++)
+	ret &= wildcard_iterate(cmd->words[i], sftp_action_chmtime, &mtime);
+
+    if (ret)
+	fznotify1(sftpDone, 1);
+
+    return ret;
+}
+
+static int sftp_cmd_mtime(struct sftp_command *cmd)
+{
+    char* unwcfname, *filename, *cname, *output;
+    int result, is_wc;
+    uint64 mtime;
+    struct fxp_attrs attrs = {0};
+    struct sftp_packet *pktin;
+    struct sftp_request *req, *rreq;
+    
+    if (back == NULL) {
+	not_connected();
+	return 0;
+    }
+
+    if (cmd->nwords != 2) {
+	fzprintf(sftpError, "mtime: expects exactly one filename as argument");
+	return 0;
+    }
+
+    filename = cmd->words[1];
+    unwcfname = snewn(strlen(filename) + 1, char);
+    is_wc = !wc_unescape(unwcfname, filename);
+    if (is_wc) {
+	fzprintf(sftpError, "mtime does not support wildcards");
+	sfree(unwcfname);
+	return 0;
+    }
+
+    cname = canonify(unwcfname);
+    sfree(unwcfname);
+    if (!cname) {
+	fzprintf(sftpError, "%s: canonify: %s", filename, fxp_error());
+	return 0;
+    }
+    attrs.flags = SSH_FILEXFER_ATTR_ACCESSTIME | SSH_FILEXFER_ATTR_CREATETIME | SSH_FILEXFER_ATTR_MODIFICATIONTIME;
+    sftp_register(req = fxp_stat_send(cname));
+    rreq = sftp_find_request(pktin = sftp_recv());
+    compare_req_rreq(req, rreq);
+    result = fxp_stat_recv(pktin, rreq, &attrs);
+
+    if (!result) {
+	fzprintf(sftpError, "get attrs for %s: %s", cname,
+	       fxp_error());
+
+	sfree(cname);
+	return 0;
+    }
+
+    attrs.flags &= SSH_FILEXFER_ATTR_ACCESSTIME | SSH_FILEXFER_ATTR_CREATETIME | SSH_FILEXFER_ATTR_MODIFICATIONTIME;
+
+    if (attrs.flags & SSH_FILEXFER_ATTR_ACMODTIME)
+	mtime = attrs.mtime;
+/*FZ: Does not work    else if (attrs.flags & SSH_FILEXFER_ATTR_MODIFICATIONTIME)
+	mtime = attrs.mtime;
+*/
+    else {
+	fzprintf(sftpError, "get attrs for %s: %s", cname,
+	       "mtime not provided");
+	sfree(cname);
+	return 0;
+    }
+    
+    sfree(cname);
+
+    output = snewn(25, char);
+    uint64_decimal(mtime, output);
+    fzprintf(sftpReply, output);
+    sfree(output);
+    return 1;
+}
+
 static int sftp_cmd_open(struct sftp_command *cmd)
 {
     int portnumber;
@@ -1983,6 +2262,12 @@ static struct sftp_cmd_lookup {
 	    "  use commas to separate different modifiers (\"u+rwx,g+s\").\n",
 	    sftp_cmd_chmod
     },
+    {
+	"chmtime", TRUE, "change file modification time",
+	    " <mtime> <filename>\n"
+	    "  time in seconds since Jan 1st, 1970 in UTC\n",
+	    sftp_cmd_chmtime
+    },
     {
 	"close", TRUE, "finish your SFTP session but do not quit PSFTP",
 	    "\n"
@@ -2022,6 +2307,14 @@ static struct sftp_cmd_lookup {
 	    "  If -r specified, recursively fetch a directory.\n",
 	    sftp_cmd_get
     },
+    {
+	"keyfile", TRUE, "add a keyfile to use",
+	    " <filename>\n"
+	    "  One of the keyfiles used during authentication.\n"
+	    "  Has to be in PuTTY's .ppk format.\n",
+	    sftp_cmd_keyfile
+    },
+
 /* BEGIN FZ UNUSED
     {
 	"help", TRUE, "give help",
@@ -2074,6 +2367,12 @@ static struct sftp_cmd_lookup {
 	    "  If -r specified, recursively store files and directories.\n",
 	    sftp_cmd_mput
     },
+    {
+	"mtime", TRUE, "get file modification time",
+	    " <filename>\n"
+	    "  Returns time in seconds since Jan 1st, 1970 in UTC\n",
+	    sftp_cmd_mtime
+    },
     {
 	"mv", TRUE, "move or rename file(s) on the remote server",
 	    " <source> [ <source>... ] <destination>\n"
@@ -2093,6 +2392,12 @@ static struct sftp_cmd_lookup {
 	    "  when you are not already connected to a server.\n",
 	    sftp_cmd_open
     },
+    {
+	"proxy", TRUE, "set a proxy",
+	    " <type> <host> [ <port> <user> ] <pass>\n"
+	    "  Type is 0 for no proxy, 1 for HTTP/1.1 and 2 for SOCKS5\n",
+	    sftp_cmd_proxy
+    },
     {
 	"put", TRUE, "upload a file from your local machine to the server",
 	    " [ -r ] [ -- ] <filename> [ <remote-filename> ]\n"
@@ -2274,6 +2579,11 @@ struct sftp_command *sftp_getcmd(FILE *fp, int mode, int modeflags)
 	cmd->words = sresize(cmd->words, cmd->wordssize, char *);
 	cmd->words[0] = dupstr("!");
 	cmd->words[1] = dupstr(p+1);
+    } else if (*p == '#') {
+	/*
+	 * Special case: comment. Entire line is ignored.
+	 */
+	cmd->nwords = cmd->wordssize = 0;
     } else {
 
 	/*
@@ -2359,7 +2669,7 @@ static int do_sftp_init(void)
      */
     sftp_register(req = fxp_realpath_send("."));
     rreq = sftp_find_request(pktin = sftp_recv());
-    assert(rreq == req);
+    compare_req_rreq(req, rreq);
     homedir = fxp_realpath_recv(pktin, rreq);
 
     if (!homedir) {
@@ -2465,7 +2775,7 @@ void fatalbox(char *fmt, ...)
     va_start(ap, fmt);
     str = dupvprintf(fmt, ap);
     va_end(ap);
-    fzprintf(sftpError, str);
+    fzprintf(sftpError, "%s", str);
     sfree(str);
 
     cleanup_exit(1);
@@ -2478,7 +2788,7 @@ void modalfatalbox(char *fmt, ...)
     va_start(ap, fmt);
     str = dupvprintf(fmt, ap);
     va_end(ap);
-    fzprintf(sftpError, str);
+    fzprintf(sftpError, "%s", str);
     sfree(str);
 
     cleanup_exit(1);
@@ -2756,7 +3066,9 @@ static int psftp_connect(char *userhost, char *user, int portnumber)
     /*
      * Trim a colon suffix off the hostname if it's there.
      */
+    /* FZ: Doing so ruins ability to connect to IPv6 addresses
     cfg.host[strcspn(cfg.host, ":")] = '\0';
+    */
 
     /*
      * Remove any remaining whitespace from the hostname.
@@ -2956,6 +3268,9 @@ int psftp_init_utf8_locale()
  */
 int psftp_main(int argc, char *argv[])
 {
+    // FZ: Set proxy to none
+    cfg.proxy_type = PROXY_NONE;
+
     int i;
     int portnumber = 0;
     char *userhost, *user;
@@ -2982,6 +3297,7 @@ int psftp_main(int argc, char *argv[])
     userhost = user = NULL;
 
     /* Load Default Settings before doing anything else. */
+    cfg.keyfile_list = 0;
     do_defaults(NULL, &cfg);
     loaded_session = FALSE;
 
@@ -3074,5 +3390,15 @@ int psftp_main(int argc, char *argv[])
     console_provide_logctx(NULL);
     sk_cleanup();
 
+    {
+	Keyfile_list *list = cfg.keyfile_list;
+	while (list)
+	{
+	    Keyfile_list *next = list->next;
+	    sfree(list);
+	    list = next;
+	}
+    }
+
     return 0;
 }
