@@ -14,6 +14,7 @@ typedef struct
 {
 	std::string data;
 	CDirentry reference;
+	enum ServerType serverType;
 } t_entry;
 
 class CDirectoryListingParserTest : public CppUnit::TestFixture
@@ -31,6 +32,7 @@ public:
 
 	void testIndividual();
 	void testAll();
+	void testSpecial();
 
 	static std::vector<t_entry> m_entries;
 
@@ -78,11 +80,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(8, wxDateTime::Apr, 1994),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// This one is a recent file with a time instead of the year.
@@ -96,11 +98,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(29, wxDateTime::Mar, calcYear(3, 29), 3, 26),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Group omitted
@@ -114,11 +116,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(8, wxDateTime::Apr, 1994),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Symbolic link
@@ -132,11 +134,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				true,
 				_T("usr/bin"),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(25, wxDateTime::Jan, calcYear(1, 25), 0, 17),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Some listings with uncommon date/time format
@@ -152,11 +154,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(26, wxDateTime::Sep, 2000),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -169,11 +171,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(26, wxDateTime::Sep, calcYear(9, 26), 13, 45),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -186,11 +188,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(7, wxDateTime::Jun, 2005, 21, 22),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 
@@ -205,11 +207,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(5, wxDateTime::Oct, calcYear(10, 5), 21, 22),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// NetWare style listings
@@ -225,11 +227,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(16, wxDateTime::Jan, calcYear(1, 16), 18, 53),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -242,11 +244,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(20, wxDateTime::Oct, calcYear(10, 20), 15, 27),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// NetPresenz for the Mac
@@ -266,11 +268,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(22, wxDateTime::Nov, 1995),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -283,11 +285,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(10, wxDateTime::May, 1996),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// A format with domain field some windows servers send
@@ -301,11 +303,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(29, wxDateTime::Jan, calcYear(1, 29), 3, 26),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// EPLF directory listings
@@ -325,11 +327,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				utc,
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	utc = wxDateTime(13, wxDateTime::Feb, 1996, 23, 58, 27);
@@ -344,11 +346,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				utc,
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// MSDOS type listing used by old IIS
@@ -364,11 +366,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(27, wxDateTime::Apr, 2000, 12, 9),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Ambiguous date and AM/PM crap. Some evil manager must have forced the poor devs to implement this
@@ -382,11 +384,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(6, wxDateTime::Apr, 2000, 15, 47),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -399,11 +401,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(2, wxDateTime::Sep, 2002, 18, 48),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -416,11 +418,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(2, wxDateTime::Sep, 2002, 19, 6),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Numerical unix style listing
@@ -436,11 +438,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				utc,
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// VShell servers
@@ -456,11 +458,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(4, wxDateTime::Apr, 2000, 21, 6),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -473,11 +475,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(12, wxDateTime::Dec, 2002, 2, 13),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* This type of directory listings is sent by some newer versions of VShell
@@ -492,11 +494,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(8, wxDateTime::Oct, 2002, 9, 47),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// OS/2 server format
@@ -513,11 +515,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(23, wxDateTime::Apr, 2003, 10, 57),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -530,11 +532,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(14, wxDateTime::Jul, 1999, 12, 37),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Another server not aware of Y2K
@@ -548,11 +550,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(11, wxDateTime::Feb, 2003, 16, 15),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Again Y2K
@@ -566,11 +568,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(5, wxDateTime::Oct, 2000, 23, 38),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Localized date formats
@@ -586,11 +588,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(26, wxDateTime::Jul, calcYear(7, 26), 20, 10),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -603,11 +605,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(2, wxDateTime::Oct, 2003),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -620,11 +622,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(12, wxDateTime::Oct, 1999, 17, 12),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -637,11 +639,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(24, wxDateTime::Apr, 2003, 17, 12),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Japanese listing
@@ -657,11 +659,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(18, wxDateTime::Apr, 2003),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Some other asian listing format. Those >127 chars are just examples
@@ -676,11 +678,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(20, wxDateTime::Mar, 2003),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -693,11 +695,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(17, wxDateTime::Aug, calcYear(8, 17), 8, 52),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -710,11 +712,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(15, wxDateTime::Jul, 2004),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// VMS listings
@@ -724,17 +726,17 @@ void CDirectoryListingParserTest::InitEntries()
 			"36-vms-dir.DIR;1  1 19-NOV-2001 21:41 [root,root] (RWE,RWE,RE,RE)",
 			{
 				_T("36-vms-dir"),
-				1,
+				512,
 				_T("RWE,RWE,RE,RE"),
 				_T("root,root"),
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(19, wxDateTime::Nov, 2001, 21, 41),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 
@@ -742,17 +744,17 @@ void CDirectoryListingParserTest::InitEntries()
 			"37-vms-file;1       155   2-JUL-2003 10:30:13.64",
 			{
 				_T("37-vms-file;1"),
-				155,
+				79360,
 				_T(""),
 				_T(""),
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				wxDateTime(2, wxDateTime::Jul, 2003, 10, 30, 13),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* VMS style listing without time */
@@ -760,17 +762,17 @@ void CDirectoryListingParserTest::InitEntries()
 			"38-vms-notime-file;1    2/8    7-JAN-2000    [IV2_XXX]   (RWED,RWED,RE,)",
 			{
 				_T("38-vms-notime-file;1"),
-				2,
+				1024,
 				_T("RWED,RWED,RE,"),
 				_T("IV2_XXX"),
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(7, wxDateTime::Jan, 2000),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* Localized month */
@@ -778,51 +780,51 @@ void CDirectoryListingParserTest::InitEntries()
 			"39-vms-notime-file;1    6/8    15-JUI-2002    PRONAS   (RWED,RWED,RE,)",
 			{
 				_T("39-vms-notime-file;1"),
-				6,
+				3072,
 				_T("RWED,RWED,RE,"),
 				_T("PRONAS"),
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(15, wxDateTime::Jul, 2002),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
 			"40-vms-multiline-file;1\r\n170774/170775     24-APR-2003 08:16:15  [FTP_CLIENT,SCOT]      (RWED,RWED,RE,)",
 			{
 				_T("40-vms-multiline-file;1"),
-				170774,
+				87436288,
 				_T("RWED,RWED,RE,"),
 				_T("FTP_CLIENT,SCOT"),
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				wxDateTime(24, wxDateTime::Apr, 2003, 8, 16, 15),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
 			"41-vms-multiline-file;1\r\n10     2-JUL-2003 10:30:08.59  [FTP_CLIENT,SCOT]      (RWED,RWED,RE,)",
 			{
 				_T("41-vms-multiline-file;1"),
-				10,
+				5120,
 				_T("RWED,RWED,RE,"),
 				_T("FTP_CLIENT,SCOT"),
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				wxDateTime(2, wxDateTime::Jul, 2003, 10, 30, 8),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// VMS style listings with a different field order
@@ -830,34 +832,34 @@ void CDirectoryListingParserTest::InitEntries()
 			"42-vms-alternate-field-order-file;1   [SUMMARY]    1/3     2-AUG-2006 13:05  (RWE,RWE,RE,)",
 			{
 				_T("42-vms-alternate-field-order-file;1"),
-				1,
+				512,
 				_T("RWE,RWE,RE,"),
 				_T("SUMMARY"),
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(2, wxDateTime::Aug, 2006, 13, 5),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
 			"43-vms-alternate-field-order-file;1       17-JUN-1994 17:25:37     6308/13     (RWED,RWED,R,)",
 			{
 				_T("43-vms-alternate-field-order-file;1"),
-				6308,
+				3229696,
 				_T("RWED,RWED,R,"),
 				_T(""),
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				wxDateTime(17, wxDateTime::Jun, 1994, 17, 25, 37),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	// Miscellaneous listings
@@ -874,11 +876,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				wxDateTime(23, wxDateTime::Feb, 2000, 15, 9, 55),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -891,11 +893,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				wxDateTime(23, wxDateTime::Feb, 2000, 15, 9, 55),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* aligned directory listing with too long size */
@@ -909,11 +911,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(12, wxDateTime::Feb, calcYear(2, 12), 17, 20),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* short directory listing with month name */
@@ -927,11 +929,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(1, wxDateTime::Jun, 1999),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* Nortel wfFtp router */
@@ -945,11 +947,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				wxDateTime(3, wxDateTime::Jun, 2004, 10, 20, 3),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* VxWorks based server used in Nortel routers */
@@ -963,11 +965,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_seconds,
 				wxDateTime(28, wxDateTime::Feb, 1998, 5, 23, 30),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* the following format is sent by the Connect:Enterprise server by Sterling Commerce */
@@ -981,11 +983,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				true,
+				CDirentry::timestamp_time,
 				wxDateTime(13, wxDateTime::Jan, calcYear(1, 13), 7, 42),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* Microware OS-9
@@ -1000,11 +1002,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(29, wxDateTime::Mar, 2007),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	/* Localised Unix style listing. Month and day fields are swapped */
@@ -1018,11 +1020,11 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(1, wxDateTime::Oct, 2004),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -1035,11 +1037,11 @@ void CDirectoryListingParserTest::InitEntries()
 				false,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(11, wxDateTime::Nov, 2005),
 				false
-			}
+			},
+			DEFAULT
 		});
 
 	m_entries.push_back((t_entry){
@@ -1053,13 +1055,489 @@ void CDirectoryListingParserTest::InitEntries()
 				true,
 				false,
 				_T(""),
-				true,
-				false,
+				CDirentry::timestamp_date,
 				wxDateTime(6, wxDateTime::Jan, 2005),
 				false
-			}
+			},
+			DEFAULT
 		});
 
+	/* IBM AS/400 style listing with localized date*/
+	m_entries.push_back((t_entry){
+			"QPGMR           36864 18.09.06 14:21:26 *FILE      55-AS400.FILE",
+			{
+				_T("55-AS400.FILE"),
+				36864,
+				_T(""),
+				_T("QPGMR"),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_seconds,
+				wxDateTime(18, wxDateTime::Sep, 2006, 14, 21, 26),
+				false
+			},
+			DEFAULT
+		});
+
+	/* VMS style listing with complex size */
+	m_entries.push_back((t_entry){
+			"56-VMS-complex-size;1 2KB 23-SEP-2005 14:57:07.27",
+			{
+				_T("56-VMS-complex-size;1"),
+				2048,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_seconds,
+				wxDateTime(23, wxDateTime::Sep, 2005, 14, 57, 7),
+				false
+			},
+			DEFAULT
+		});
+
+	/* HP NonStop */
+	m_entries.push_back((t_entry){
+			"57-HP_NonStop 101 528 6-Apr-07 14:21:18 255, 0 \"oooo\"",
+			{
+				_T("57-HP_NonStop"),
+				528,
+				_T("\"oooo\""),
+				_T("255, 0"),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_seconds,
+				wxDateTime(6, wxDateTime::Apr, 2007, 14, 21, 18),
+				false
+			},
+			HPNONSTOP
+		});
+
+	// Only difference is in the owner/group field, no delimiting space.
+	m_entries.push_back((t_entry){
+			"58-HP_NonStop 101 528 6-Apr-07 14:21:18 255,255 \"oooo\"",
+			{
+				_T("58-HP_NonStop"),
+				528,
+				_T("\"oooo\""),
+				_T("255,255"),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_seconds,
+				wxDateTime(6, wxDateTime::Apr, 2007, 14, 21, 18),
+				false
+			},
+			HPNONSTOP
+		});
+
+
+	m_entries.push_back((t_entry){
+			"drwxr-xr-x 6 user sys 1024 30. Jan., 12:40 59-localized-date-dir",
+			{
+				_T("59-localized-date-dir"),
+				1024,
+				_T("drwxr-xr-x"),
+				_T("user sys"),
+				true,
+				false,
+				_T(""),
+				CDirentry::timestamp_time,
+				wxDateTime(30, wxDateTime::Jan, calcYear(1, 30), 12, 40),
+				false
+			},
+			DEFAULT
+		});
+
+	// MVS variants
+	//
+	// Note: I am not quite sure of these get parsed correctly, but so far
+	//       nobody did complain. Formats added here with what I think
+	//       is at least somewhat correct, so that there won't be any
+	//       regressions at least.
+
+	// The following 5 are loosely based on this format:
+	// Volume Unit    Referred Ext Used Recfm Lrecl BlkSz Dsorg Dsname
+	m_entries.push_back((t_entry){
+			"WYOSPT 3420   2003/05/21  1  200  FB      80  8053  PS  60-MVS.FILE",
+			{
+				_T("60-MVS.FILE"),
+				100,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_date,
+				wxDateTime(21, wxDateTime::May, 2003),
+				false
+		},
+		DEFAULT
+	});
+
+	m_entries.push_back((t_entry){
+			"WPTA01 3290   2004/03/04  1    3  FB      80  3125  PO  61-MVS.DATASET",
+			{
+				_T("61-MVS.DATASET"),
+				-1,
+				_T(""),
+				_T(""),
+				true,
+				false,
+				_T(""),
+				CDirentry::timestamp_date,
+				wxDateTime(04, wxDateTime::Mar, 2004),
+				false
+		},
+		DEFAULT
+	});
+
+	m_entries.push_back((t_entry){
+			"NRP004 3390   **NONE**    1   15  NONE     0     0  PO  62-MVS-NONEDATE.DATASET",
+			{
+				_T("62-MVS-NONEDATE.DATASET"),
+				-1,
+				_T(""),
+				_T(""),
+				true,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+		},
+		DEFAULT
+	});
+
+	m_entries.push_back((t_entry){
+			"TSO005 3390   2005/06/06 213000 U 0 27998 PO 63-MVS.DATASET",
+			{
+				_T("63-MVS.DATASET"),
+				-1,
+				_T(""),
+				_T(""),
+				true,
+				false,
+				_T(""),
+				CDirentry::timestamp_date,
+				wxDateTime(6, wxDateTime::Jun, 2005),
+				false
+		},
+		DEFAULT
+	});
+
+	m_entries.push_back((t_entry){
+			"TSO004 3390   VSAM 64-mvs-file",
+			{
+				_T("64-mvs-file"),
+				-1,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+		},
+		DEFAULT
+	});
+
+	// MVS Dataset members
+	// 
+	// As common with IBM misdesign, multiple styles exist.
+	
+	// Speciality: Some members have no attributes at all.
+	// Requires servertype to be MVS or it won't be parsed, as 
+	// it would conflict with lots of other servers.
+	m_entries.push_back((t_entry){
+			"65-MVS-PDS-MEMBER",
+			{
+				_T("65-MVS-PDS-MEMBER"),
+				-1,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+		},
+		MVS
+	});
+
+	// Name         VV.MM   Created      Changed       Size  Init  Mod Id
+	m_entries.push_back((t_entry){
+			"66-MVSPDSMEMBER 01.01 2004/06/22 2004/06/22 16:32   128   128    0 BOBY12",
+			{
+				_T("66-MVSPDSMEMBER"),
+				128,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_time,
+				wxDateTime(22, wxDateTime::Jun, 2004, 16, 32),
+				false
+		},
+		MVS
+	});
+
+	// Hexadecimal size
+	m_entries.push_back((t_entry){
+			"67-MVSPDSMEMBER2 00B308 000411  00 FO                31    ANY",
+			{
+				_T("67-MVSPDSMEMBER2"),
+				45832,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+		},
+		MVS
+	});
+
+	m_entries.push_back((t_entry){
+			"68-MVSPDSMEMBER3 00B308 000411  00 FO        RU      ANY    24",
+			{
+				_T("68-MVSPDSMEMBER3"),
+				45832,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+			},
+			MVS
+		});
+
+	// Migrated MVS file
+	m_entries.push_back((t_entry){
+			"Migrated				69-SOME.FILE",
+			{
+				_T("69-SOME.FILE"),
+				-1,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+			},
+			MVS
+		});
+
+	// z/VM, another IBM abomination. Description by Alexandre Charbey
+	// Requires type set to ZVM or it cannot be parsed.
+	// 
+	// 70-ZVMFILE
+	//   is a filename
+	// TRACE
+	//   is a filetype (extension, like exe or com or jpg...)
+	// V
+	//   is the file format. Designates how records are arranged in a file. F=Fixed and V=Variable. I don't think you care
+	// 65
+	//   is the logical record length.
+	// 107
+	//   is Number of records in a file.
+	// 2
+	//   (seems wrong) is the block size ( iirc 1 is 127, 2 is 381, 3 is 1028 and 4 is 4072 - not sure - the numbers are not the usual binary numbers)
+	// there is the date/time
+	// 060191
+	//   I think it is some internal stuff saying who the file belongs to.  191 is the "handle" of the user's disk. I don't know what 060 is. This 060191 is what FZ shows in its file list.
+	m_entries.push_back((t_entry){
+			"70-ZVMFILE  TRACE   V        65      107        2 2005-10-04 15:28:42 060191",
+			{
+				_T("70-ZVMFILE.TRACE"),
+				6955,
+				_T(""),
+				_T("060191"),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_seconds,
+				wxDateTime(4, wxDateTime::Oct, 2005, 15, 28, 42),
+				false
+			},
+			ZVM
+		});
+
+	m_entries.push_back((t_entry){
+			"drwxr-xr-x 3 slopri devlab 512 71-unix-dateless",
+			{
+				_T("71-unix-dateless"),
+				512,
+				_T("drwxr-xr-x"),
+				_T("slopri devlab"),
+				true,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+			},
+			DEFAULT
+		});
+
+	utc = wxDateTime(5, wxDateTime::Nov, 2008, 16, 52, 15);
+	utc.MakeFromTimezone(wxDateTime::UTC);
+	m_entries.push_back((t_entry){
+			"Type=file;mOdIfY=20081105165215;size=1234; 72-MLSD-file",
+			{
+				_T("72-MLSD-file"),
+				1234,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_seconds,
+				utc,
+				false
+			},
+			DEFAULT
+		});
+
+	// Yet another MVS format.
+	// Follows the below structure but with all but the first two and the last field empty.
+	// Furthermore, Unit is "Tape"
+	// Volume Unit    Referred Ext Used Recfm Lrecl BlkSz Dsorg Dsname
+	m_entries.push_back((t_entry) {
+			"V43525 Tape                                             73-MSV-TAPE.FILE",
+			{
+				_T("73-MSV-TAPE.FILE"),
+				-1,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+			},
+			MVS
+		});
+
+	m_entries.push_back((t_entry){
+			"Type=file; 74-MLSD-whitespace trailing\t ",
+			{
+				_T("74-MLSD-whitespace trailing\t "),
+				-1,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+			},
+			DEFAULT
+		});
+
+		m_entries.push_back((t_entry){
+			"Type=file; \t 75-MLSD-whitespace leading",
+			{
+				_T("\t 75-MLSD-whitespace leading"),
+				-1,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+			},
+			DEFAULT
+		});
+
+		utc = wxDateTime(26, wxDateTime::Apr, 2008, 13, 55, 01);
+		utc.MakeFromTimezone(wxDateTime::UTC);
+		m_entries.push_back((t_entry){
+			"modify=20080426135501;perm=;size=65718921;type=file;unique=802U1066013B;UNIX.group=1179;UNIX.mode=00;UNIX.owner=1179; 75 MLSD file with empty permissions",
+			{
+				_T("75 MLSD file with empty permissions"),
+				65718921,
+				_T("00"),
+				_T("1179 1179"),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_seconds,
+				utc,
+				false
+			},
+			DEFAULT
+		});	
+
+		m_entries.push_back((t_entry){
+			"type=OS.unix=slink:/foo; 76 MLSD symlink",
+			{
+				_T("76 MLSD symlink"),
+				-1,
+				_T(""),
+				_T(""),
+				true,
+				true,
+				_T("/foo"),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+			},
+			DEFAULT
+		});	
+
+		// Old ietf draft for MLST earlier than mlst-07 has no trailing semicolon after facts
+		m_entries.push_back((t_entry){
+			"type=file 77 MLSD file no trailing semicolon after facts < mlst-07",
+			{
+				_T("77 MLSD file no trailing semicolon after facts < mlst-07"),
+				-1,
+				_T(""),
+				_T(""),
+				false,
+				false,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+			},
+			DEFAULT
+		});
+
+		m_entries.push_back((t_entry){
+			"type=OS.unix=slink; 77 MLSD symlink notarget",
+			{
+				_T("77 MLSD symlink notarget"),
+				-1,
+				_T(""),
+				_T(""),
+				true,
+				true,
+				_T(""),
+				CDirentry::timestamp_none,
+				wxDateTime(),
+				false
+			},
+			DEFAULT
+		});	
+
 /*
 	wxString name;
 	wxLongLong size;
@@ -1090,7 +1568,8 @@ void CDirectoryListingParserTest::testIndividual()
 
 	m_sync.Leave();
 
-	const CServer server;
+	CServer server;
+	server.SetType(entry.serverType);
 
 	CDirectoryListingParser parser(0, server);
 
@@ -1115,10 +1594,12 @@ void CDirectoryListingParserTest::testIndividual()
 
 void CDirectoryListingParserTest::testAll()
 {
-	const CServer server;
+	CServer server;
 	CDirectoryListingParser parser(0, server);
 	for (std::vector<t_entry>::const_iterator iter = m_entries.begin(); iter != m_entries.end(); iter++)
 	{
+		server.SetType(iter->serverType);
+		parser.SetServer(server);
 		const char* str = iter->data.c_str();
 		const int len = strlen(str);
 		char* data = new char[len];
